/************************************************************************
 * This library is free software; you can redistribute it and/or	*
 * modify it under the terms of the GNU Library General Public		*
 * License as published by the Free Software Foundation; either		*
 * version 2 of the License, or (at your option) any later version.	*
 * 									*
 * This library is distributed in the hope that it will be useful,	*
 * but WITHOUT ANY WARRANTY; without even the implied warranty of	*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU	*
 * Library General Public License for more details.			*
 * 									*
 * You should have received a copy of the GNU Library General Public	*
 * License along with this library; if not, write to the		*
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,		*
 * Boston, MA  02111-1307, USA.						*
 * 									*
 * The GNU Library General Public License file is included in		*
 * Online-Documentation							*
 *									*
 * GNU's Homepage - http://www.gnu.ai.mit.edu/home.html			*
 *									*
 * Questions & comments referring to *FLIP++: please contact us:	*
 *									*
 * Snail-Mail:								*
 *	Dr. Wolfgang SLANY						*
 *	Database and Expert System Group (E184-2)     			*
 *	Institute of Information Systems               			*
 *	Vienna University of Technology					*
 *	Paniglgasse 16, A-1040 Vienna, AUSTRIA				*
 * Telephone:								*
 *	+43-1-58801-6123     (area-code 1 maybe changed to 201 in 1998) *
 * FAX:									*
 *	+43-1-5055304	     (area-code 1 maybe changed to 201 in 1998) *
 * eMail:								*
 *	sflip@dbai.tuwien.ac.at						*
 * WWW:									*
 *	http://www.dbai.tuwien.ac.at/proj/StarFLIP/			*
 *                                                                      *
 * $Id: domflip.cpp,v 1.3 1997/06/24 17:27:10 reichel Exp $ *
 ***********************************************************************/
// ----------------------------------------------------------------------------------------------------------
//	What   : Domain Knowledge Representation with *FLIP++
//	Project: DomFLIP++
//	Author : Andreas Raggl
//	Start  : 02.12.1996
//	Last   : 14.01.1997
// 	File   : domflip.cpp
// ----------------------------------------------------------------------------------------------------------

#include "domflip.h"

// global functions

bool operator<( class ConstraintEvaluation c1, class ConstraintEvaluation c2)
{
  return ( c1.GetValue() < c2.GetValue() );
};

ostream& operator<<		( ostream& os, CShift& c )
{
	c.Write( os );
	return os;
};

istream& operator>>		( istream& is, CShift& c )
{
	c.Read( is );
	return is;
};

ostream& operator<<		( ostream& os, CGroup& c )
{
	c.Write( os );
	return os;
};

istream& operator>>		( istream& is, CGroup& c )
{
	c.Read( is );
	return is;
};

ostream& operator<<		( ostream& os, COperating& c )
{
	c.Write( os );
	return os;
};

istream& operator>>		( istream& is, COperating& c )
{
	c.Read( is );
	return is;
};

ostream& operator<< ( ostream& os, CDomainInstance& c )
{
	c.Write( os );
	return os;
};

istream& operator>> ( istream& is, CDomainInstance& c )
{
	c.Read( is );
	return is;
};

ostream& operator<<		( ostream& os, CShiftSchedulingDomain& c )
{
	c.Write( os );
	return os;
};

istream& operator>>		( istream& is, CShiftSchedulingDomain& c )
{
	c.Read( is );
	return is;
};

// class definition of CShift

CShift::CShift			( char *s, int i, int j ) : CBase( "CShift", 1 ), length( i ), shiftsrequired( j )
{
	strcpy( name, s );
};

CShift::~CShift			()		
{};

void CShift::SetLength	( float f )	
{ 
	length = f; 
};

float CShift::GetLength	() const
{ 
	return length; 
};

void CShift::SetName	( char *s )	
{ 
	strcpy( name, s ); 
};

char *CShift::GetName	() const
{ 
	return (char *)name; 
};

void CShift::SetShiftsRequired	( int i )
{
	shiftsrequired = i;
};

int CShift::GetShiftsRequired	() const
{
	return shiftsrequired;
};

void CShift::Read		( istream& is )
{
	CBase::Read( is );
	switch(classversion)
	{
	case 1: 
		namestr str;
		float f;
		int i;
		is >> str >> f >> i;
		SetName(str);
		SetLength(f);
		SetShiftsRequired(i);
		break;
	default:
		;
	}
};
	
void CShift::Write		( ostream& os ) const
{
	CBase::Write( os );
	os << GetName() << space << GetLength() << space << GetShiftsRequired() << endl;
};

// class definition Group

CGroup::CGroup			( char *s, int i ) : CBase( "CGroup", 1 ), members( i )
{
	strcpy( name, s );
};

CGroup::~CGroup			()
{};

void CGroup::SetName	( char *s )
{
	strcpy( name, s ); 
};

char *CGroup::GetName	() const
{
	return (char *)name;
};

void CGroup::SetMembers	( int i )
{
	members = i;
};

int CGroup::GetMembers	() const
{
	return members;
};

void CGroup::Read		( istream& is )
{
	CBase::Read ( is );
	switch( classversion )
	{
	case 1: 
		namestr str;
		int i;
		is >> str >> i;
		SetName( str );
		SetMembers ( i );
		break;
	default:
		;
	}
};

void CGroup::Write		( ostream& os ) const
{
	CBase::Write( os );
	os << GetName() << space << GetMembers() << endl;
};

// class definition COperating

COperating::COperating		() : CBase( "COperating", 1 ) 
{};

COperating::~COperating()	
{};

void COperating::SetShift	( int i )	
{ 
	shift = i; 
};

int COperating::GetShift	() const	
{ 
	return shift; 
};

void COperating::SetLevel	( int i )	
{ 
	level = i; 
};

int COperating::GetLevel	() const	
{ 
	return level; 
};

void COperating::SetNumber	( int i )	
{ 
	number = i; 
};

int COperating::GetNumber	() const	
{ 
	return number; 
};	

void COperating::Read		( istream& is ) 
{ 
	CBase::Read( is ); 
	is >> shift >> level >> number; 
};

void COperating::Write		( ostream& os ) const 
{ 
	CBase::Write( os ); 
	os << shift << space << level << space << number << endl; 
};

// class definition CDomainInstance

CDomainInstance::CDomainInstance() : CBase( "CDomainInstance", 1 )
{ 
	pStaticData = new char[MAX_FILE_LENGTH];
};

CDomainInstance::~CDomainInstance() 
{ 
	delete pStaticData;
};

float CDomainInstance::Evaluate	()
{
	TVarlist* pVariables = new TVarlist();
	
	InitializeVariables( pVariables );	
	pVariables->write("test2.dat");	
	TEvalTree* pEvalTree = interpret( pStaticData, pVariables);
	if ( pEvalTree )
	{
		EvalData = pEvalTree->GetConstraintList();
		pEvalTree->clean_up();
	}
	if ( EvalData.size() )
	{
//		evaluation = EvalData[0].GetValue();
		sort( EvalData.begin(), EvalData.end() );
	        evaluation = 0;
                for (int i = 0; i < (int)EvalData.size(); i++)
                   evaluation += EvalData[i].GetValue();
		ComputeRepairStarts();
		return evaluation;
	}
	else
		return -1;
};

ConstraintEvaluations& CDomainInstance::GetEvalData()
{
	return EvalData;
};

void CDomainInstance::SetStaticData ( char *c ) 
{ 
	strcpy(pStaticData, c); 
};

char* CDomainInstance::GetStaticData() 
{ 
	return pStaticData;
};

void CDomainInstance::SetStaticDataFileName( char *s )
{
	strcpy( StaticDataFileName, s );
};

char *CDomainInstance::GetStaticDataFileName()
{
	return StaticDataFileName;
};

bool CDomainInstance::ReadStaticData()
{
	char *pHelp = pStaticData;
	char c;
	ifstream infile;

	infile.open( StaticDataFileName, ios::in );
	if ( infile.fail() )
	{
		cout << "CDomainInstance: couldn't find static data file" << endl;
		return false;
	}
	while ( infile.get(c) )
		*( pHelp++ ) = c;
	infile.close();
	*pHelp = '\0';
	return true;
};

float CDomainInstance::GetEvaluation() 
{ 
	return evaluation;
};

void CDomainInstance::Read		( istream& is )
{
	CBase::Read( is );
	switch( classversion )
	{
	case 1:
		is >> StaticDataFileName;
		if ( !ReadStaticData() )
		{
			cout << "Failed to read StaticDataFile" << endl;
		}
		break;
	default:
		;
	}
};

void CDomainInstance::Write		( ostream& os ) const
{
	CBase::Write( os );
	os << StaticDataFileName << endl;
};


// class definition CShiftSchedulingDomain

CShiftSchedulingDomain::CShiftSchedulingDomain	() : CDomainInstance()
{
	strcpy( classname, "CShiftSchedulingDomain" );
	classversion = 1;
	operating = new COperatingMatrix;
	shifts = new CShiftVector;
	groups = new CGroupVector;
};

CShiftSchedulingDomain::~CShiftSchedulingDomain	()
{
	delete groups;
	delete shifts;
	delete operating;
};

bool CShiftSchedulingDomain::Repair( int group, int day, int &start, int &start2 )
{
	bool bResult = false;
	switch ( day % 7 )
	{
	case 0:
	case 1:
	case 2:
			bResult = RepairStep3( group, day, start, start2 );
			break;
	case 3:
	case 4:
			bResult = RepairStep4( group, day, start, start2 );
			break;
	case 5:
			switch ( schedule[group][day] )
			{
			case 0:
				// decide whether RepairStep1 or RepairStep2
				bResult = RepairStep2( group, day, start);
				break;
			case 2:
				bResult = RepairStep1( group, day, start);
				break;
			case 3:
				bResult = RepairStep2( group, day, start);
				break;
			}
			break;	
	case 6:
			bResult = RepairStep1( group, day, start );
	}
	return bResult;
};

// repairstep for shift TDWE (Sat, Sun)->(Sat, Sun)
// day must be a Sat/Sun
// after day is reached again false is returned
bool CShiftSchedulingDomain::RepairStep1( int group, int day, int &start )
{
	// check whether shift is (0,2)
	if ( schedule[ group ][ day ] != 2 && schedule[ group ][ day ] != 0)
	{
		cout << "Shift to repair out of range!\n";
		return false;
	}

	// check whether day is (Sat - Sun)
	if ( day % 7 < 5 ) 
	{
		cout << "Day to repair out of range!\n";
		return false;
	}
	
	// check whether start is (Mon - Sun)
	if ( start % 7 < 0 || start % 7 > 7 ) 
	{
		cout << "Repair start out of range!\n";
		return false;
	}

	static int iLoopBegin = -1;
	int i;
	int iSourceShift = schedule[ group ][ day ];
	int iDestShift;
	if ( iSourceShift == 0 )
		iDestShift = 2;
	else
		iDestShift = 0;
	int iSwapDay = start;
	if ( iLoopBegin == -1 ) 
		iLoopBegin = start - 1;
	while ( 1 )
	{
		iSwapDay = iSwapDay % ( LengthOfCycle * 7 );
		// one cycle has been searched ?
		if ( iSwapDay == iLoopBegin ) break;
		if ( iSwapDay % 7 == 5 || iSwapDay % 7 == 6 )
		{
			if ( schedule[ group ][ iSwapDay ] == iDestShift )
			{
				for ( i = 0; i < (int)groups->size(); i++ )
				{
					if ( schedule[ i ][ iSwapDay ] == iSourceShift && 
						schedule[ i ][ day ] == iDestShift )
					{	
						Swap( group, day, group, iSwapDay );
						Swap( i, day, i, iSwapDay );
						start = iSwapDay + 1;
						return true;
					}
				}
			}
		}
		iSwapDay++;
	}
	iLoopBegin = -1;
	return false;
};

// repairstep for shift SWWE (Sat)->(Sat)
// day must be a Sat
// after day is reached again false is returned
bool CShiftSchedulingDomain::RepairStep2( int group, int day, int &start )
{
	// check whether shift is (0,3)
	if ( schedule[ group ][ day ] != 0 && schedule[ group ][ day ] != 3 )
	{
		cout << "Shift to repair out of range!\n";
		return false;
	}

	if ( schedule[ group ][ day ] == 0 )
	{
		int h = group / 2;
		if ( schedule[ h * 2 ][ day ] != schedule[ h * 2 + 1 ][ day ] )
		{
			cout << "No repair possible, violation must be a group!\n";
			return false;
		}
	}	
	// check whether day is (Sat)
	if ( day % 7 != 5 ) 
	{
		cout << "Day to repair out of range!\n";
		return false;
	}
	
	// check whether start is (Mon - Sun)
	if ( start % 7 < 0 || start % 7 > 7 ) 
	{
		cout << "Repair start out of range!\n";
		return false;
	}

	static int iLoopBegin = -1;
	int i, j;
	int iSourceShift = schedule[ group ][ day ];
	int iDestShift;
	if ( iSourceShift == 0 )
		iDestShift = 3;
	else
		iDestShift = 0;
	int iSwapDay = start;
	int iSwapGroup = group / 2;
	if ( iLoopBegin == -1 ) 
		iLoopBegin = start - 1;
	while ( 1 )
	{
		iSwapDay = iSwapDay % ( LengthOfCycle * 7 );
		// one cycle has been searched ?
		if ( iSwapDay == iLoopBegin ) break;
		if ( iSwapDay % 7 == 5 )
		{
			// is the iSwapDay a possible candidate
			bool bFree = true;
			for ( i = 0; i < 2; i++ )
				if ( schedule[ iSwapGroup * 2 + i ][ iSwapDay ] != iDestShift )
					bFree = false;	
			if ( bFree )
			{
				for ( i = 0; i < ( (int)groups->size() / 2 ); i++ )
				{
					bFree = true;
					for ( j = 0; j < 2; j++ )
						if ( !( schedule[ i * 2 + j ][ iSwapDay ] == iSourceShift &&
							schedule[ i * 2 + j ][ day ] == iDestShift) )
							bFree = false;
					if ( bFree )
					{	
						for ( j = 0; j < 2; j++ )
						{
							Swap( iSwapGroup * 2 + j, day, iSwapGroup * 2 + j, iSwapDay );
							Swap( i * 2 + j, day, i * 2 + j, iSwapDay );
						}
						start = iSwapDay + 1;
						return true; 
					}
				}
			} 
		}
		iSwapDay++;
	} // while
	iLoopBegin = - 1;
	return false;
};

// repairstep for shift TD (Mon - Wed)->(Mon - Fri)
// shift must be (0,1)
// day must be (Mon - Wed)
// start must be (Mon - Fri)
// start2 must be (Mon - Wed)
// after day is reached again false is returned
bool CShiftSchedulingDomain::RepairStep3( int group, int day, int &start, int &start2 )
{
	// check whether shift is (0,1)
	if ( schedule[ group ][ day ] < 0 || schedule[ group ][ day ] > 1 )
	{
		cout << "Shift to repair out of range!\n";
		return false;
	}

	// check whether day is (Mon - Wed)
	if ( day % 7 < 0 || day % 7 > 6 ) 
	{
		cout << "Day to repair out of range!\n";
		return false;
	}
	
	// check whether start is (Mon - Sun)
	if ( start % 7 < 0 || start % 7 > 7 ) 
	{
		cout << "Repair start out of range!\n";
		return false;
	}

	static int iLoopBegin = -1;
	static bool bFirstTry = true;
	static int iLastGroup = 0;
	int iSourceShift = schedule[ group ][ day ];
	int iDestShift;
	if ( iSourceShift == 0 )
		iDestShift = 1;
	else
		iDestShift = 0;
	bool bFound = false;
	int iSwapGroup;
	int iSwapDay = start;
	if ( iLoopBegin == -1 ) 
		iLoopBegin = start - 1;
	while ( 1 )
	{
		iSwapDay = iSwapDay % ( LengthOfCycle * 7 );
		// one cycle has been searched ?
		if ( iSwapDay == iLoopBegin ) break;
		// is the swapday a possible candidate
		if ( schedule[ group ][ iSwapDay ] == iDestShift )
		{
			int i;
			for ( i = iLastGroup; i < (int)groups->size(); i++ )
			{
				// reswap possible ?
				if ( ( schedule[ i ][ iSwapDay ] == iSourceShift ) && ( schedule[ i ][ day ] == iDestShift ) )
				{
					iSwapGroup = i;
					if ( iSwapDay % 7 >= 0 && iSwapDay % 7 <= 2 )
					{
						Swap( group, day, group, iSwapDay );
						Swap( iSwapGroup, day, iSwapGroup, iSwapDay );
						start = iSwapDay + 1;
						iLastGroup = 0;
						return true;
					}
					else if ( iSwapDay % 7 >= 3 && iSwapDay % 7 <= 4 )
					{
						int iSwapDay2;
						if ( bFirstTry )
						{
							iSwapDay2 = iSwapDay + 1;
							bFirstTry = false;
						}
						else
							iSwapDay2 = start2;
						int iSwapGroup2, iSwapGroup3;
						if ( group % 2 == 0 )
							iSwapGroup2 = group + 1;
						else
							iSwapGroup2 = group - 1;
						if ( iSwapGroup % 2 == 0 )
							iSwapGroup3 = iSwapGroup + 1;
						else
							iSwapGroup3 = iSwapGroup - 1;
						while (!bFound)
						{
							iSwapDay2 = iSwapDay2 % ( LengthOfCycle * 7 );
							if ( iSwapDay2 == iSwapDay ) 
							{
								bFirstTry = true;
								if ( iSwapGroup == (int)groups->size() ) 
									iLastGroup = 0;
								else
									iLastGroup = iSwapGroup + 1;
								break;
							}
							if ( iSwapDay2 % 7 >= 0 && iSwapDay2 % 7 <= 2 )
							{
								if ( ( schedule[ iSwapGroup2 ][ iSwapDay2 ] == iSourceShift ) &&
									 ( schedule[ iSwapGroup3 ][ iSwapDay2 ] == iDestShift ) )
								{
									Swap( group, day, group, iSwapDay );
									Swap( iSwapGroup, day, iSwapGroup, iSwapDay );
									Swap( iSwapGroup2, iSwapDay, iSwapGroup2, iSwapDay2 );
									Swap( iSwapGroup3, iSwapDay, iSwapGroup3, iSwapDay2 );
									start = iSwapDay;
									start2 = iSwapDay2 + 1;
									bFound = true;
									return bFound;
								}
							}
							iSwapDay2++;
						}
					} 
				} // if reswap possible ?
			} // for
		} // if swapday possible candidate ?
		iSwapDay++;
		iLastGroup = 0;
	} // while
	iLoopBegin = -1;
	return false;
};

// after day is reached again false is returned
// repairstep for shift TD (Thu, Fri)->(Mon, Fri)
// shift must be (0,1)
// day must be (Thu - Fri)
// start must be (Mon - Fri)
// after day is reached again false is returned
bool CShiftSchedulingDomain::RepairStep4( int group, int day, int &start, int &start2 )
{
	// check whether shift is (0,1)
	if ( schedule[ group ][ day ] < 0 || schedule[ group ][ day ] > 1 )
	{
		cout << "Shift to repair out of range!\n";
		return false;
	}

	// check whether day is (Thu - Fri)
	if ( day % 7 < 3 || day % 7 > 4 ) 
	{
		cout << "Day to repair out of range!\n";
		return false;
	}
	
	// check whether start is (Mon - Sun)
	if ( start % 7 < 0 || start % 7 > 6 ) 
	{
		cout << "Repair start out of range!\n";
		return false;
	}

	int i, j;
	static int iLoopBegin = -1;
	static bool bFirstTry = true;
	static int iLastGroup = 0;
	int iSourceShift = schedule[ group ][ day ];
	int iDestShift;
	if ( iSourceShift == 0 )
		iDestShift = 1;
	else
		iDestShift = 0;
	bool bFound = false;
	int iSwapGroup;
	int iSwapDay = start;
	int iGroup = group / 2;	
	if ( iLoopBegin == -1 ) 
		iLoopBegin = start - 1;
	while ( 1 )
	{
		iSwapDay = iSwapDay % ( LengthOfCycle * 7 );
		// one cycle has been searched ?
		if ( iSwapDay == iLoopBegin ) break;
		// swap with (Thu-Fri) ?
		if ( iSwapDay % 7 >= 3 && iSwapDay % 7 <= 4)
		{
			// is the swapday a possible candidate ?
			bool bFree = true;
			for ( i = 0; i < 2; i++ )
				if ( schedule[ iGroup * 2 + i ][ iSwapDay ] != iDestShift )
					bFree = false;	
			if ( bFree )
			{
				for ( i = iLastGroup; i < (int)groups->size() / 2; i++ )
				{
					// reswap possible ?
					if ( ( schedule[ i * 2 ][ iSwapDay ] == iSourceShift ) && 
						( schedule[ i * 2][ day ] == iDestShift ) &&
						( schedule[ i * 2 + 1][ iSwapDay ] == iSourceShift ) &&
						( schedule[ i * 2 + 1][ day ] == iDestShift ) )
					{
						for ( j = 0; j < 2; j++ )
						{
							Swap( iGroup * 2 + j, day, iGroup * 2 + j, iSwapDay );
							Swap( i * 2 + j, day, i * 2 + j, iSwapDay );
						}
						start = iSwapDay + 1;
						iLastGroup = 0;
						return true;
					} // reswap possible ?
				} // for
			} // is the swapday a possible candidate ?
		} // swap with (Thu - Fri)
		// swap with (Mon - Wed) ?
		if ( iSwapDay % 7 >= 0 && iSwapDay % 7 <= 3 )
		{
			// is the swapday a possible candidate
			if ( schedule[ group ][ iSwapDay ] == iDestShift )
			{
				int i;
				for ( i = iLastGroup; i < (int)groups->size(); i++ )
				{
					// reswap possible ?
					if ( ( schedule[ i ][ iSwapDay ] == iSourceShift ) && ( schedule[ i ][ day ] == iDestShift ) )
					{
						iSwapGroup = i;
						int iSwapDay2;
						if ( bFirstTry )
						{
							iSwapDay2 = iSwapDay + 1;
							bFirstTry = false;
						}
						else
							iSwapDay2 = start2;
						int iSwapGroup2, iSwapGroup3;
						if ( group % 2 == 0 )
							iSwapGroup2 = group + 1;
						else
							iSwapGroup2 = group - 1;
						if ( iSwapGroup % 2 == 0 )
							iSwapGroup3 = iSwapGroup + 1;
						else
							iSwapGroup3 = iSwapGroup - 1;
						while (!bFound)
						{
							iSwapDay2 = iSwapDay2 % ( LengthOfCycle * 7 );
							if ( iSwapDay2 == iSwapDay ) 
							{
								bFirstTry = true;
								if ( iSwapGroup == (int)groups->size() ) 
									iLastGroup = 0;
								else
									iLastGroup = iSwapGroup + 1;
								break;
							}
							if ( iSwapDay2 % 7 >= 0 && iSwapDay2 % 7 <= 2 )
							{
								if ( ( schedule[ iSwapGroup2 ][ iSwapDay2 ] == iDestShift ) &&
									 ( schedule[ iSwapGroup3 ][ iSwapDay2 ] == iSourceShift ) )
								{
									Swap( group, day, group, iSwapDay );
									Swap( iSwapGroup, day, iSwapGroup, iSwapDay );
									Swap( iSwapGroup2, day, iSwapGroup2, iSwapDay2 );
									Swap( iSwapGroup3, day, iSwapGroup3, iSwapDay2 );
									start = iSwapDay;
									start2 = iSwapDay2 + 1;
									return true;
								}
							}
							iSwapDay2++;
						}
					} // reswap possible ?
				} // for
			} // is the swapday a possible candidate
		} // swap with (Mon - Wed) ?
		iSwapDay++;
		iLastGroup = 0;
	} // while
	iLoopBegin = -1;
	return false;
};

void CShiftSchedulingDomain::Swap( int group1, int day1, int group2, int day2)
{
	int h;

	cout << "("<< group1 << "," << day1 << ") <-> (" << group2 << "," << day2 <<") " << schedule[group1][day1]<<space<< schedule[ group2 ][ day2 ] << endl;
	h = schedule[ group1 ][ day1 ];
	schedule[ group1 ][ day1 ] = schedule[ group2 ][ day2 ];
	schedule[ group2 ][ day2 ] = h;
}

bool CShiftSchedulingDomain::CheckPlan() const
{
	int i, j, k;
	for ( i = 0; i < (int)groups->size(); i++ )
	{
		CIntVector c( shifts->size(), 0 );
		for ( j = 0; j < LengthOfCycle * 7; j++ )
			c[ schedule[ i ][ j ] ]++;
		for ( j = 0; j < (int)shifts->size(); j++ )
			if ( c[ j ] != (*shifts)[j].GetShiftsRequired() )
				return false;			
	}

	for ( i = 0; i < LengthOfCycle * 7; i++ )
	{
		CIntVector c ( shifts->size(), 0 );
		for ( j = 0; j < (int)groups->size(); j++ )
			c[ schedule[ j ][ i ] ]++;

		if ( i % 7 >= 0 && i % 7 <= 2 )
		{
			if ( c[0] != 1 || c[1] != 5 )
				return false;
		}
		if ( i % 7 >= 3 && i % 7 <= 4 )
		{
			if ( c[0] != 2 || c[1] != 4 )
				return false;
			for ( k = 0; k < (int)groups->size(); k+=2 )
				if ( schedule[ k ][ i ] != schedule[ k + 1 ][ i ] )
					return false;
		}
	}
	return true;
};

void CShiftSchedulingDomain::Write	( char *filename ) const
{
	ofstream outfile;
	outfile.open(filename, ios::out|ios::trunc);
	if(outfile)
	{
		Write(outfile);
		outfile.close();
	}
};

void CShiftSchedulingDomain::Write	( ostream &output ) const
{
	CDomainInstance::Write( output );
	
	int i, n;
	
	// write domain variables
	output << "LengthOfCycle" << space << LengthOfCycle << "\n";
	
	// write groups
	n = groups->size();
	output << "/* #: " << n << " */ \n";
	for (i = 0; i < n; i++)
		output << (*groups)[i];
	output << endl;

	// write shifts
	n = shifts->size();
	output << "/* #: " << n << " */ \n";
	for (i = 0; i < n; i++)
		output << (*shifts)[i];
	output << endl;

	// write operation plan
	n = operating->size();
	output << "/* #: " << n << " */ \n";
	for (i = 0; i < n; i++)
	{
		int m = (*operating)[i].size();
		output << "/* #: " << m << " */ \n";
		for (int j = 0; j < m; j++)
			output << (*operating)[i][j];
		output << endl;
	}

	// write schedule
	n = schedule.size();
	output << "/* #: " << n << " */ \n";
	for (i = 0; i < n; i++)
	{
		int m = schedule[i].size();
		output << "/* #: " << m << " */ \n";
		for (int j = 0; j < m; j++)
			output << schedule[i][j] << space;
		output << endl;
	}
};

void CShiftSchedulingDomain::Read	( char *filename )
{
	ifstream infile;
	infile.open(filename, ios::in);
	if(infile)
	{
		Read(infile);
		infile.close();
	}
};

void CShiftSchedulingDomain::Read	( istream &input )
{
	CDomainInstance::Read( input );
	
	char instr[20];
	int i, j, m, n;

	// read domain variables
	input >> instr;	// LengthOfCycle
	input >> LengthOfCycle;
	
	// read groups
	groups->erase(groups->begin(), groups->end());
	input >> instr; //  /*
	input >> instr; //  #:
	input >> n;
	input >> instr; //  */
	for ( i = 0; i < n; i++ )
	{
		CGroup g;
		input >> g;
		groups->push_back(g);
	}

	// read shifts
	shifts->erase( shifts->begin(), shifts->end() );
	input >> instr; //  /*
	input >> instr; //  #:
	input >> n;
	input >> instr; //  */
	for ( i = 0; i < n; i++ )
	{
		CShift s;
		input >> s;
		shifts->push_back( s );
	}

	// read operation plan
	operating->erase( operating->begin(), operating->end() );
	input >> instr; //  /*
	input >> instr; //  #:
	input >> n;
	input >> instr; //  */
	n = n / 7;
	if ( (LengthOfCycle % n ) != 0 )
	{
		cout << "Operation plan cycle should be divisor of the shift schedule cycle" << "\n";
		return;
	}
	LengthOfOperatingCycle = n;
	for ( i = 0; i < n * 7; i++ )
	{
		COperatingVector c;
		
		input >> instr; //  /*
		input >> instr; //  #:
		input >> m;
		input >> instr; //  */
		for ( j = 0; j < m; j++)
		{
			COperating k;
			input >> k;
			c.push_back( k );
		}
		operating->push_back(c);
	}
	
	// read schedule
	schedule.erase( schedule.begin(), schedule.end() );
	input >> instr; //  /*
	input >> instr; //  #:
	input >> n;
	input >> instr; //  */
	if ( n != (int)groups->size() )
	{
		cout << "Shift schedule does not allocate for each group" << "\n";
		return;
	}
	for ( i = 0; i < n; i++ )
	{
		CIntVector v;
		input >> instr; //  /*
		input >> instr; //  #:
		input >> m;
		input >> instr; //  */
		if ( m != (LengthOfCycle * 7) )
		{
			cout << "Shift schedule does not allocate for entire cycle length in group #" << i << "\n";
			return;
		}
		for ( j = 0; j < m; j++ )
		{
			int c;
			input >> c;
			v.push_back(c);
		}
		schedule.push_back(v);

	}	
};

void CShiftSchedulingDomain::InitializeVariables ( TVarlist *vars ) 
{
	int i;
	char r_name[MAX_NAME_LEN];
	char r_formula[MAX_FORMULA_LEN];
	float r_value = 0.0;

	// compute constraint variables
	ComputeConstraintVars();

	// length of shiftplan cycle
	strcpy( r_name, "LengthOfCycle" );
	sprintf( r_formula, "%d", LengthOfCycle );
	r_value = LengthOfCycle; 
	vars->insert( r_name, r_formula, r_value );

	// groups
	int n = groups->size();
	for (i = 0; i < n; i++)
	{
		// group name
		sprintf( r_name, "groups[%d].name", i);
		sprintf( r_formula, "\"%s\"", (*groups)[i].GetName() );
		r_value = 0.0; 
		vars->insert( r_name, r_formula, r_value );
		// group members
		sprintf( r_name, "groups[%d].members", i);
		sprintf( r_formula, "%d", (*groups)[i].GetMembers() );
		r_value = (*groups)[i].GetMembers(); 
		vars->insert( r_name, r_formula, r_value );
	}

	// shifts
	n = shifts->size();
	for (i = 0; i < n; i++)
	{
		// shift name
		sprintf( r_name, "shifts[%d].name", i);
		sprintf( r_formula, "\"%s\"", (*shifts)[i].GetName() );
		r_value = 0.0; 
		vars->insert( r_name, r_formula, r_value );
		// shift length
		sprintf( r_name, "shifts[%d].length", i);
		sprintf( r_formula, "%f", (*shifts)[i].GetLength() );
		r_value = (*shifts)[i].GetLength(); 
		vars->insert( r_name, r_formula, r_value );
	}
		
	// shiftplan
	n = schedule.size();
	for (i = 0; i < n; i++)
	{
		int m = schedule[i].size();
		for (int j = 0; j < m; j++)
		{
			sprintf( r_name, "plan[%d][%d]", i, j);
			sprintf( r_formula, "%d", schedule[i][j] );
			r_value = schedule[i][j]; 
			vars->insert( r_name, r_formula, r_value );
		}
			
	}

	// Constraint variables
	// DIFFHOURS
	n = c_diff.size();
	for (i = 0; i < n; i++)
	{
		int m = c_diff[i].size();
		for (int j = 0; j < m; j++)
		{
			sprintf( r_name, "c_diff[%d][%d]", i, j);
			sprintf( r_formula, "%f", c_diff[i][j] );
			r_value = c_diff[i][j]; 
			vars->insert( r_name, r_formula, r_value );
		}			
	}

	// WORKHOURS
	n = c_work.size();
	for (i = 0; i < n; i++)
	{
		int m = c_work[i].size();
		for (int j = 0; j < m; j++)
		{
			sprintf( r_name, "c_work[%d][%d]", i, j);
			sprintf( r_formula, "%f", c_work[i][j] );
			r_value = c_work[i][j]; 
			vars->insert( r_name, r_formula, r_value );
		}			
	}

	// SHIFTBLOCKS
	n = c_shiftblocks.size();
	for (i = 0; i < n; i++)
	{
		int m = c_shiftblocks[i].size();
		sprintf( r_name, "c_numberofblocks[%d]", i);
		sprintf( r_formula, "%d", m );
		r_value = m; 
		vars->insert( r_name, r_formula, r_value );
		for (int j = 0; j < m; j++)
		{
			sprintf( r_name, "c_shiftblocks[%d][%d].start", i, j);
			sprintf( r_formula, "%d", c_shiftblocks[i][j].start );
			r_value = c_shiftblocks[i][j].start; 
			vars->insert( r_name, r_formula, r_value );
			sprintf( r_name, "c_shiftblocks[%d][%d].size", i, j);
			sprintf( r_formula, "%d", c_shiftblocks[i][j].size );
			r_value = c_shiftblocks[i][j].size; 
			vars->insert( r_name, r_formula, r_value );
		}			
	}

	// FREEWEEKENDS
	n = c_freeweekends.size();
	for (i = 0; i < n; i++)
	{
		sprintf( r_name, "c_freeweekends[%d]", i);
		sprintf( r_formula, "%f", c_freeweekends[i] );
		r_value = c_freeweekends[i]; 
		vars->insert( r_name, r_formula, r_value );
	}	

	// BADWEEKENDS
	n = c_badweekends.size();
	for (i = 0; i < n; i++)
	{
		int m = c_badweekends[i].size();
		for (int j = 0; j < m; j++)
		{
			sprintf( r_name, "c_badweekends[%d][%d]", i, j);
			sprintf( r_formula, "%d", c_badweekends[i][j] );
			r_value = c_badweekends[i][j]; 
			vars->insert( r_name, r_formula, r_value );
		}			
	
	}

	vars->substitute_all();
}

void CShiftSchedulingDomain::WriteVariables ( char *filename )
{
	ofstream outfile;
	outfile.open(filename, ios::out|ios::trunc);
	if(outfile)
	{
		WriteVariables(outfile);
		outfile.close();
	}
};

void CShiftSchedulingDomain::WriteVariables ( ostream &output )
{

	int i, n;
	
	// compute constraint variables
	ComputeConstraintVars();

	// number of variables
	n = 1 +										// LengthOfCycle
		groups->size() * 2 +					// # groups
		shifts->size() * 2 +					// # shifts
		LengthOfCycle * 7 * groups->size() +	// # schedule
		( LengthOfCycle - 1 ) * groups->size() +// # diffhours
		LengthOfCycle * groups->size() +		// # workhours
		groups->size() +						// # freeweekends
		( LengthOfCycle - 1 ) * groups->size() + 6;	// # badweekends
	for ( i = 0; i < (int)c_shiftblocks.size(); i++ )// # shiftblocks
		n += c_shiftblocks[i].size() * 2;

	output << "/ #: " << n << " / " << endl;

	// length of the shift cycle
	output << "LengthOfCycle" << space << LengthOfCycle << space << LengthOfCycle << endl;
	
	// groups
	n = groups->size();
	for (i = 0; i < n; i++)
	{
		output << "groups[" << i << "].name \"" << (*groups)[i].GetName() << "\" 0" << endl;
		output << "groups[" << i << "].members " << (*groups)[i].GetMembers() << space << (*groups)[i].GetMembers() << endl;
	}
	
	// shifts
	n = shifts->size();
	for (i = 0; i < n; i++)
	{
		output << "shifts[" << i << "].name \"" << (*shifts)[i].GetName() << "\" 0" << endl;
		output << "shifts[" << i << "].length " << (*shifts)[i].GetLength() << space << (*shifts)[i].GetLength() << endl;
	}
	
	// shiftplan
	n = schedule.size();
	for (i = 0; i < n; i++)
	{
		int m = schedule[i].size();
		for (int j = 0; j < m; j++)
			output << "plan[" << i << "][" << j << "] " << schedule[i][j] << space << schedule[i][j] << endl;
	}

	// constraint variables
	// DIFFHOURS
	n = c_diff.size();
	for (i = 0; i < n; i++)
	{
		int m = c_diff[i].size();
		for (int j = 0; j < m; j++)
			output << "c_diff[" << i << "][" << j << "] " << c_diff[i][j] << space << c_diff[i][j] << endl;
	}

	// WORKHOURS
	n = c_work.size();
	for (i = 0; i < n; i++)
	{
		int m = c_work[i].size();
		for (int j = 0; j < m; j++)
			output << "c_work[" << i << "][" << j << "] " << c_work[i][j] << space << c_work[i][j] << endl;
	}

	// SHIFTBLOCKS
	n = c_shiftblocks.size();
	for (i = 0; i < n; i++)
	{
		int m = c_shiftblocks[i].size();
		output << "c_numberofblocks[" << i << "] " << m << space << m << endl;
		for (int j = 0; j < m; j++)
		{
			output << "c_shiftblocks[" << i << "][" << j << "].start " << c_shiftblocks[i][j].start << space << c_shiftblocks[i][j].start << endl;
			output << "c_shiftblocks[" << i << "][" << j << "].size " << c_shiftblocks[i][j].size << space << c_shiftblocks[i][j].size << endl;
		}
	}
	
	// FREEWEEKENDS
	n = c_freeweekends.size();
	for (i = 0; i < n; i++)
		output << "c_freeweekends[" << i << "] " << c_freeweekends[i] << space << c_freeweekends[i] << endl;
	
	// BADWEEKENDS
	n = c_badweekends.size();
	for (i = 0; i < n; i++)
	{
		int m = c_badweekends[i].size();
		for (int j = 0; j < m; j++)
			output << "c_badweekends[" << i << "][" << j << "] " << c_badweekends[i][j] << space << c_badweekends[i][j] << endl;
	}	
};

void CShiftSchedulingDomain::ComputeRepairStarts()
{
	int iNumberOfViol = 10;
	int iCount = 0, i = 0;
 
        repairstarts.erase( repairstarts.begin(), repairstarts.end() );
	while ( iCount < iNumberOfViol )
	{
		if ( i == (int)EvalData.size() ) break;
		if ( strcmp( EvalData[i].GetBranch(), "branch" ) == 0 )
		{
			i++;
			continue;
		}
		else if ( strcmp( EvalData[i].GetBranch(), "DIFFHOURS") == 0 ) 
		{
			ComputeRepairStart_DIFFHOURS( EvalData[i].GetFormula() );
			iCount++; 
		        i++;
		}
		else if ( strcmp( EvalData[i].GetBranch(), "WORKINGHOURS") == 0 ) 
		{
			ComputeRepairStart_WORKINGHOURS( EvalData[i].GetFormula() );
			iCount++;
		        i++;
			}
		else if ( strcmp( EvalData[i].GetBranch(), "SHIFTBLOCKS") == 0 ) 
		{
			ComputeRepairStart_SHIFTBLOCKS( EvalData[i].GetFormula() );
			iCount++;
		        i++;
			}
		else if ( strcmp( EvalData[i].GetBranch(), "FREEWEEKENDS") == 0 ) 
		{
			ComputeRepairStart_FREEWEEKENDS( EvalData[i].GetFormula() );
			iCount++;
		        i++;
			}
		else if ( strcmp( EvalData[i].GetBranch(), "BADWEEKENDS") == 0 ) 
		{
			ComputeRepairStart_BADWEEKENDS( EvalData[i].GetFormula() );
			iCount++;
		        i++;
			}
		else
		{
			cout << "No repair positions function for constraint: " << EvalData[i].GetBranch() << endl;
		}
	}		
};

// computes repair starts for constraint DIFFHOURS
void CShiftSchedulingDomain::ComputeRepairStart_DIFFHOURS( char* s )
{
	char formula[MAX_FORMULA_LEN];
	strcpy( formula, s );
	char g[MAX_FORMULA_LEN];
	char w[MAX_FORMULA_LEN];
	char *h = formula, *h1 = g, *h2 = w;
	while ( ! (*h++ == '[' ) );
	while ( ! ( (*h1++ = *h++ ) == ']' ) );
	*(--h1) = '\0';
	while ( ! (*h++ == '[' ) );
	while ( ! ( (*h2++ = *h++ ) == ']' ) );
	*(--h2) = '\0';

	int group = atoi( g );
	int week = atoi( w );

	// which of the different weeks has more hours
	float s1 = 0;
	float s2 = 0;
	int j;
	for ( j = 0; j < 7; j++)
	{
		s1 = s1 + (*shifts)[ schedule[ group ][ week * 7 + j ] ].GetLength();
		s2 = s2 + (*shifts)[ schedule[ group ][ ( week + 1 ) * 7 + j ] ].GetLength();
	}
	if ( s1 < s2 ) week++;

	// fill vector with possible repair starts
	CPositionVector v;
	for ( j = 0; j < 7; j++ )
	{
	 if ( schedule[ group ][ ( week ) * 7 + j ] == 1 || schedule[ group ][ ( week ) * 7 + j ] == 3 )
	   {
		CPosition p;
		p.group = group;
		p.day = week * 7 + j;
		v.push_back( p );
//		cout << "Repair start: " << group << space << p.day << endl;
	      }
	}
        srand( (unsigned) time( NULL ) );
	int iRandomStart = rand() % v.size();
        v[0].day = v[iRandomStart].day;
	repairstarts.push_back( v );
};


// computes repair starts for constraint WORKINGHOURS
void CShiftSchedulingDomain::ComputeRepairStart_WORKINGHOURS( char* s )
{
	char formula[MAX_FORMULA_LEN];
	strcpy( formula, s );
	char g[MAX_FORMULA_LEN];
	char w[MAX_FORMULA_LEN];
	char *h = formula, *h1 = g, *h2 = w;
	while ( ! (*h++ == '[' ) );
	while ( ! ( (*h1++ = *h++ ) == ']' ) );
	*(--h1) = '\0';
	while ( ! (*h++ == '[' ) );
	while ( ! ( (*h2++ = *h++ ) == ']' ) );
	*(--h2) = '\0';

	int group = atoi( g );
	int week = atoi( w );

	// fill vector with possible repair starts
	CPositionVector v;
	for ( int i = 0; i < 7; i++ )
	{
		CPosition p;
		p.group = group;
		p.day = week * 7 + i;
		v.push_back( p );
	}
	repairstarts.push_back( v );
};

void CShiftSchedulingDomain::ComputeRepairStart_SHIFTBLOCKS( char* s )
{
	char formula[MAX_FORMULA_LEN];
	strcpy( formula, s );
	char g[MAX_FORMULA_LEN];
	char w[MAX_FORMULA_LEN];
	char *h = formula, *h1 = g, *h2 = w;
	while ( ! (*h++ == '[' ) );
	while ( ! ( (*h1++ = *h++ ) == ']' ) );
	*(--h1) = '\0';
	while ( ! (*h++ == '[' ) );
	while ( ! ( (*h2++ = *h++ ) == ']' ) );
	*(--h2) = '\0';

	int group = atoi( g );
	int blocknumber = atoi( w );

	// fill vector with possible repair starts
	CPositionVector v;
	for ( int i = 0; i < c_shiftblocks[group][blocknumber].size ; i++ )
	{
		CPosition p;
		p.group = group;
		p.day = c_shiftblocks[group][blocknumber].start + i;
		v.push_back( p );
	}
	repairstarts.push_back( v );
};

void CShiftSchedulingDomain::ComputeRepairStart_FREEWEEKENDS( char* s )
{
	CPosition p;
	p.group = 0;
	p.day = 0;
	CPositionVector v;
	v.push_back( p );
	repairstarts.push_back( v );
};

void CShiftSchedulingDomain::ComputeRepairStart_BADWEEKENDS( char* s )
{
	CPosition p;
	p.group = 0;
	p.day = 0;
	CPositionVector v;
	v.push_back( p );
	repairstarts.push_back( v );
};

// computes variables for constraints
void CShiftSchedulingDomain::ComputeConstraintVars()
{
	ComputeConstraintVars_DIFFHOURS();
	ComputeConstraintVars_WORKINGHOURS();
	ComputeConstraintVars_SHIFTBLOCKS();
	ComputeConstraintVars_FREEWEEKENDS();
	ComputeConstraintVars_BADWEEKENDS();
};

void CShiftSchedulingDomain::ComputeConstraintVars_DIFFHOURS()
{
	// clean data structure
	c_diff.erase( c_diff.begin(), c_diff.end() );
	
	float previous_hoursperweek = 0;
	for ( int i = 0; i < (int)groups->size(); i++ )
	{	
		CFloatVector v;
		for ( int j = 0; j < LengthOfCycle; j++ )
		{
			float hoursperweek =  0;
			for ( int k = 0; k < 7; k++ )
				hoursperweek += (*shifts)[ schedule[ i ][ j * 7 + k ] ].GetLength();
			if ( j > 0 )    
				v.push_back( hoursperweek - previous_hoursperweek );
			previous_hoursperweek = hoursperweek;
		}
		c_diff.push_back( v );
	}
};

void CShiftSchedulingDomain::ComputeConstraintVars_WORKINGHOURS()
{
	// clean data structure
	c_work.erase( c_work.begin(), c_work.end() );
	
	for ( int i = 0; i < (int)groups->size(); i++ )
	{	
		CFloatVector v;
		for ( int j = 0; j < LengthOfCycle; j++ )
		{
			float hoursperweek =  0;
			for ( int k = 0; k < 7; k++ )
				hoursperweek += (*shifts)[ schedule[ i ][ j * 7 + k ] ].GetLength();
			v.push_back( hoursperweek );
		}
		c_work.push_back( v );
	}
};

void CShiftSchedulingDomain::ComputeConstraintVars_SHIFTBLOCKS()
{
	// clean data structure
	c_shiftblocks.erase( c_shiftblocks.begin(), c_shiftblocks.end() );

	for ( int i = 0; i < (int)groups->size(); i++ )
	{
		int blocksize = 0, j = 0;
		CShiftBlockVector v;
		while (1)
		{
			while ( schedule[i][j] == 0 && j < LengthOfCycle * 7)
				j++;
			while ( schedule[i][j] != 0 && j < LengthOfCycle * 7)
			{
				blocksize++; 
				j++;
			}
			if ( j < LengthOfCycle * 7 )
			{
				CShiftBlock c;
				c.start = j - blocksize;
				c.size = blocksize;
				v.push_back( c );
				blocksize = 0;
			}
			else
				break;
		}
		c_shiftblocks.push_back( v );
	}
};

void CShiftSchedulingDomain::ComputeConstraintVars_FREEWEEKENDS()
{
	// clean data structure
	c_freeweekends.erase( c_freeweekends.begin(), c_freeweekends.end() );

	for ( int i = 0; i < (int)groups->size(); i++ )
	{	
		int count = 0;
		for ( int j = 5; j < LengthOfCycle; j+=7 )
		{
			if ( schedule[i][j] == 0 && schedule[i][j+1] == 0 )
				count++;
		}
		c_freeweekends.push_back( (float) count / (float) LengthOfCycle );
	}
};

void CShiftSchedulingDomain::ComputeConstraintVars_BADWEEKENDS()
{
	// clean data structure
	c_badweekends.erase( c_badweekends.begin(), c_badweekends.end() );

	for ( int i = 0; i < (int)groups->size(); i++ )
	{	
		CIntVector v;
		for ( int j = 5; j < ( LengthOfCycle - 1 ) * 7; j+=7 )
		{
			// count number of allocated days on consecutive weekends
			int count = 0;
			if ( schedule[ i ][ j ] != 0 ) count++;
			if ( schedule[ i ][ j + 1 ] != 0 ) count++;
			if ( schedule[ i ][ j + 7 ] != 0 ) count++;
			if ( schedule[ i ][ j + 8 ] != 0 ) count++;
			v.push_back( count );
		}
		c_badweekends.push_back( v );
	}
};
