/************************************************************************
 * This library is free software; you can redistribute it and/or	*
 * modify it under the terms of the GNU Library General Public		*
 * License as published by the Free Software Foundation; either		*
 * version 2 of the License, or (at your option) any later version.	*
 * 									*
 * This library is distributed in the hope that it will be useful,	*
 * but WITHOUT ANY WARRANTY; without even the implied warranty of	*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU	*
 * Library General Public License for more details.			*
 * 									*
 * You should have received a copy of the GNU Library General Public	*
 * License along with this library; if not, write to the		*
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,		*
 * Boston, MA  02111-1307, USA.						*
 * 									*
 * The GNU Library General Public License file is included in		*
 * Online-Documentation							*
 *									*
 * GNU's Homepage - http://www.gnu.ai.mit.edu/home.html			*
 *									*
 * Questions & comments referring to *FLIP++: please contact us:	*
 *									*
 * Snail-Mail:								*
 *	Dr. Wolfgang SLANY						*
 *	Database and Expert System Group (E184-2)     			*
 *	Institute of Information Systems               			*
 *	Vienna University of Technology					*
 *	Paniglgasse 16, A-1040 Vienna, AUSTRIA				*
 * Telephone:								*
 *	+43-1-58801-6123     (area-code 1 maybe changed to 201 in 1998) *
 * FAX:									*
 *	+43-1-5055304	     (area-code 1 maybe changed to 201 in 1998) *
 * eMail:								*
 *	sflip@dbai.tuwien.ac.at						*
 * WWW:									*
 *	http://www.dbai.tuwien.ac.at/proj/StarFLIP/			*
 *                                                                      *
 * $Id: domflip.h,v 1.3 1997/06/24 17:27:12 reichel Exp $ *
 ***********************************************************************/
// ----------------------------------------------------------------------------------------------------------
//	What   : Domain Knowledge Representation with *FLIP++
//	Project: DomFLIP++
//	Author : Andreas Raggl
//	Start  : 11.12.1996
//	Last   : 31.01.1997
//	File   : domflip.h
// ----------------------------------------------------------------------------------------------------------

#if !defined( DOMFLIP_H )
#define DOMFLIP_H

#if defined ( __GNUC__ ) 
#include <fstream.h>
#include <String.h>
#include <algo.h>
#endif
#if defined ( _MSC_VER )
#include <fstream>
#include <algorithm>
#endif 

#include <time.h>
#include "base.h"
#include "inter.h"


#define MAX_FILE_LENGTH	65535

class CShift : public CBase
{
public:
	CShift			( char *s = "NONAME", int i = 0, int j = 0 );
	virtual ~CShift	();
	
	void SetLength	( float f );
	float GetLength	() const;
	void SetName	( char *s );
	char *GetName	() const;
	void SetShiftsRequired ( int i );
	int GetShiftsRequired () const;

	void Read		( istream& is );
	void Write		( ostream& os ) const;
	friend ostream& operator<< ( ostream& os, CShift& c );
	friend istream& operator>> ( istream& is, CShift& c );
private:
	namestr		name;			// identifier
	float		length;			// length in hours
	int			shiftsrequired;	// number of shifts required throughout cycle
};


class CGroup : public CBase
{
public:
	CGroup			( char *s = "NONAME", int i = 0 );
	virtual ~CGroup	();

	void SetMembers	( int i );
	int GetMembers	() const;
	void SetName	( char *s );
	char *GetName	() const;
	
	void Read		( istream& is );
	void Write		( ostream& os ) const;
	friend ostream& operator<< ( ostream& os, CGroup& c );
	friend istream& operator>> ( istream& is, CGroup& c );
private:
	namestr			name;			// identifier
	int				members;		// number of group members
};


class COperating : public CBase
{
public:
	COperating		();
	virtual ~COperating();

	void SetShift	( int i );
	int GetShift	() const;
	void SetLevel	( int i );
	int GetLevel	() const;
	void SetNumber	( int i);
	int GetNumber	() const;	
	
	void Read		( istream& is );
	void Write		( ostream& os ) const;
	friend ostream& operator<< ( ostream& os, COperating& c );
	friend istream& operator>> ( istream& is, COperating& c );
private:
	int				shift;
	int				level;
	int				number;
};

typedef CIntMatrix CSchedule;

struct CShiftBlock
{
	int start;
	int size;
};

typedef vector <CShiftBlock
#if defined ( _MSC_VER )
	, allocator<CShiftBlock>
#endif
> CShiftBlockVector;

typedef vector <CShiftBlockVector 
#if defined ( _MSC_VER )
	, allocator<CShiftBlockVector>
#endif
> CShiftBlockMatrix;

struct CPosition
{
	int group;
	int day;
};

typedef vector <CPosition 
#if defined ( _MSC_VER )
	, allocator<CPosition>
#endif
> CPositionVector;

typedef vector <CPositionVector 
#if defined ( _MSC_VER )
	, allocator<CPositionVector>
#endif
> CPositionMatrix;

typedef vector<float
#if defined ( _MSC_VER )
	, allocator<float>
#endif
> CFloatVector;

typedef vector <CFloatVector 
#if defined ( _MSC_VER )
	, allocator<CFloatVector>
#endif
> CFloatMatrix;

typedef vector <COperating
#if defined ( _MSC_VER )
	, allocator<COperating>
#endif
> COperatingVector;

typedef vector <COperatingVector 
#if defined ( _MSC_VER )
	, allocator<COperatingVector>
#endif
> COperatingMatrix;

typedef vector <CGroup 
#if defined ( _MSC_VER )
	, allocator<CGroup>
#endif
> CGroupVector;

typedef vector <CShift 
#if defined ( _MSC_VER )
	, allocator<CShift>
#endif
> CShiftVector;

class CDomainInstance : public CBase
{
public:
	CDomainInstance();
	virtual ~CDomainInstance		();

	virtual float Evaluate			();
	virtual bool Repair( int group, int day, int &start, int &start2 ) = 0;
	
	bool Optimized					();

	ConstraintEvaluations& GetEvalData();
	float GetEvaluation				();
	CPositionMatrix& GetRepairPositions() { return repairstarts; };
	void SetSchedule( CSchedule cs ) { schedule = cs; };
	CSchedule& GetSchedule() { return schedule; };
	void SetStaticData				( char *c );
	char* GetStaticData				();
	void SetStaticDataFileName		( char *s );
	char *GetStaticDataFileName		();
	bool ReadStaticData				();
	
	void Read		( istream& is );
	void Write		( ostream& os ) const;
	friend ostream& operator<< ( ostream& os, CDomainInstance& c );
	friend istream& operator>> ( istream& is, CDomainInstance& c );

protected:	
	virtual void InitializeVariables( TVarlist *vars ) = 0;
	virtual void ComputeRepairStarts() = 0;
	
	
	CSchedule		schedule;
	CPositionMatrix	repairstarts;	
	ConstraintEvaluations EvalData;
	char			*pStaticData;
	namestr			StaticDataFileName;
	bool			optimized;
	float			evaluation;
};

class CShiftSchedulingDomain : public CDomainInstance
{
public:
	CShiftSchedulingDomain		();
	virtual ~CShiftSchedulingDomain();
	
	bool Repair( int group, int day, int &start, int &start2 );
	
	void SetLengthOfCycle( int i );
	int GetLengthOfCycle() const { return LengthOfCycle; };
		
	void WriteVariables	( char *filename );
	void WriteVariables	( ostream &output );
	void Write		( char *filename ) const;
	void Read		( char *filename );
	void Write		( ostream &output ) const;
	void Read		( istream &input );
	friend ostream& operator<< ( ostream& os, CShiftSchedulingDomain& c );
	friend istream& operator>> ( istream& is, CShiftSchedulingDomain& c );

protected:
	bool CheckPlan	() const; 
	bool RepairStep1( int group, int day, int &start );
	bool RepairStep2( int group, int day, int &start );
	bool RepairStep3( int group, int day, int &start, int &start2 );
	bool RepairStep4( int group, int day, int &start, int &start2 );
	void Swap		( int group1, int day1, int group2, int day2);
	
	void ComputeConstraintVars();
	void ComputeConstraintVars_DIFFHOURS();
	void ComputeConstraintVars_WORKINGHOURS();
	void ComputeConstraintVars_SHIFTBLOCKS();
	void ComputeConstraintVars_FREEWEEKENDS();
	void ComputeConstraintVars_BADWEEKENDS();

	void ComputeRepairStarts();
	void ComputeRepairStart_DIFFHOURS( char *s );
	void ComputeRepairStart_WORKINGHOURS( char *s );
	void ComputeRepairStart_SHIFTBLOCKS( char *s );
	void ComputeRepairStart_FREEWEEKENDS( char *s );
	void ComputeRepairStart_BADWEEKENDS( char *s );

	void InitializeVariables	( TVarlist *vars );

	COperatingMatrix *operating;
	CShiftVector	*shifts;
	CGroupVector	*groups;

	int				LengthOfCycle;
	int				LengthOfOperatingCycle;
	int				HoursPerWeek;
	
	CFloatMatrix	c_diff;
	CFloatMatrix	c_work;
	CShiftBlockMatrix	c_shiftblocks;
	CFloatVector	c_freeweekends;
	CIntMatrix		c_badweekends;
};

#endif // !defined( DOMFLIP_H )
