/************************************************************************
 * This library is free software; you can redistribute it and/or	*
 * modify it under the terms of the GNU Library General Public		*
 * License as published by the Free Software Foundation; either		*
 * version 2 of the License, or (at your option) any later version.	*
 * 									*
 * This library is distributed in the hope that it will be useful,	*
 * but WITHOUT ANY WARRANTY; without even the implied warranty of	*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU	*
 * Library General Public License for more details.			*
 * 									*
 * You should have received a copy of the GNU Library General Public	*
 * License along with this library; if not, write to the		*
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,		*
 * Boston, MA  02111-1307, USA.						*
 * 									*
 * The GNU Library General Public License file is included in		*
 * Online-Documentation							*
 *									*
 * GNU's Homepage - http://www.gnu.ai.mit.edu/home.html			*
 *									*
 * Questions & comments referring to *FLIP++: please contact us:	*
 *									*
 * Snail-Mail:								*
 *	Dr. Wolfgang SLANY						*
 *	Database and Expert System Group (E184-2)     			*
 *	Institute of Information Systems               			*
 *	Vienna University of Technology					*
 *	Paniglgasse 16, A-1040 Vienna, AUSTRIA				*
 * Telephone:								*
 *	+43-1-58801-6123     (area-code 1 maybe changed to 201 in 1998) *
 * FAX:									*
 *	+43-1-5055304	     (area-code 1 maybe changed to 201 in 1998) *
 * eMail:								*
 *	sflip@dbai.tuwien.ac.at						*
 * WWW:									*
 *	http://www.dbai.tuwien.ac.at/proj/StarFLIP/			*
 *                                                                      *
 * $Id: evaltree.cpp,v 1.3 1997/06/24 17:27:13 reichel Exp $ *
 ***********************************************************************/
// -----------------------------------------------------------------------
//	What   : evaluation tree for constraint aggregation
//	Project: DynaFLIP++
//	Author : Ulrich Santa
//	Start  : 13.12.1995
//	Last   : 20.07.1996
//  File   : evaltree.cpp
// -----------------------------------------------------------------------

#include "evaltree.h"

// -----------------------------------------------------------------------
// Constructor for default TEvalTree - object
// -----------------------------------------------------------------------
TEvalTree::TEvalTree(void)
{
 set_name("empty");  
 Rules=NULL;
 paras=NULL;
 result=NULL;
 scores=new TVarlist();
 next=NULL;
 entries = 0;
 c_op = AND;
 comp = FALSE;
}

// -----------------------------------------------------------------------
// Constructor for generation of root node of TEvalTree
// -----------------------------------------------------------------------
TEvalTree::TEvalTree(char *n,ParameterSet *ps, RuleSet *rs, TVarlist *bl)
{
 TEvalTree *run=this;
 set_name(n);
 Rules = rs;
 paras = ps;
 scores = new TVarlist();
 next = NULL;
 result = new CombinedVariable(ps,name);
 entries = 1;
 comp = FALSE;
 c_op = no_cop;

 while (bl!=NULL)
 {
  run->next=new TEvalTree();
  run->next->set_name(bl->get_name());
  run->next->set_rule_set(Rules);
  run->next->set_parameter_set(paras);
  run->next->result = new CombinedVariable(paras,run->next->name);
  if (strcmp(bl->get_formula(),"AND")==0)
		run->next->c_op = c_and;
  else
      run->next->c_op = c_or;
  entries++;
  run=run->next;
  bl = bl->get_next();
 }
}

// -----------------------------------------------------------------------
// Create help CombinedVariables and appropriate rules for AND/OR rules
// -----------------------------------------------------------------------
void TEvalTree::insert_help_rules(char *tablename)
{
 CombinedVariable *eval_in_1 = new CombinedVariable(paras,"eval_in_1");
 CombinedVariable *eval_in_2 = new CombinedVariable(paras,"eval_in_2");
 CombinedVariable *eval_out_and = new CombinedVariable(paras,"eval_out_and");
 CombinedVariable *eval_out_or = new CombinedVariable(paras,"eval_out_or");
 char c_operator[MAX_NAME_LEN];
 char in1[MAX_NAME_LEN];
 char in2[MAX_NAME_LEN];
 char out[MAX_NAME_LEN];
 char token[MAX_NAME_LEN];

 concat_operator conc_op;
 ifstream tablefile;			// stream for table
 unsigned int found;			// flag for success of search in table

 if (strcmp(tablename,"empty")==0)
   {
    cout<<endl<<"Error: no TABLE defined for generation of rules!"<<endl;
    return;
   }
 
   found = FALSE;					// initialize flag for search in table
   tablefile.open(tablename,ios::in);	 		// open table
   while((!found)&&(tablefile.eof()==0))
         {
          tablefile >> token;
          if (strcmp(token,"CONCAT")==0)		// search for CONCAT block
              {
 	       tablefile>>token;
	       if (strcmp(token,paras->get_ps_name())==0)	
		  {
		   tablefile>>token;
		   if (strcmp(token,paras->get_ps_name())==0) 	
			{
			 tablefile>>token;
		         if (strcmp(token,paras->get_ps_name())==0)  
			    {
				found = TRUE;
				Rules->get_cv_list()->add(eval_in_1);
				Rules->get_cv_list()->add(eval_in_2);
				Rules->get_cv_list()->add(eval_out_and);
				Rules->get_cv_list()->add(eval_out_or);
			    }
			}
		  }
	      }
	  }
   if (!found)	// no entry for this type of concatenation
        {
         cout<<endl<<"Error: no entry in TABLE for generation of rules!"<<endl;
         return;
        }
   
     // Add all rules to ruleset
     tablefile>>c_operator;		// skip AND/OR - TAG
     conc_op = no_cop;
     if (strcmp(c_operator,"OR")==0) 				// set right operator for concatenation
	    conc_op = OR;
     else
	    conc_op = AND;
   while (conc_op!=no_cop)
   {
     while(tablefile.eof()==0)
        {
	 tablefile>>in1;							// get first input linguistic term
	 if ((strcmp(in1,"CONCAT")==0)||(strcmp(in1,"COMPARE")==0)||	// end of block in table reached?
	     (strcmp(in1,"AND")==0)||(strcmp(in1,"OR")==0)||
	     (strcmp(in1,"=")==0)||(strcmp(in1,"#")==0)||
	     (strcmp(in1,"<")==0)||(strcmp(in1,">")==0)||(strcmp(in1,"")==0))
	     {
	      break;
	     }
         else
	     {
		tablefile>>in2;							// get second input linguistic term
		tablefile>>out;							// get output linguistic term
		if (conc_op==AND)
		    Rules->add_ComplexRule(eval_in_1, in1, eval_in_2, in2, eval_out_and, out,  conc_op, 1);
		else
		    Rules->add_ComplexRule(eval_in_1, in1, eval_in_2, in2, eval_out_or, out,  conc_op, 1);
	     }
         }
         conc_op = no_cop;
         if (strcmp(in1,"AND")==0) conc_op = AND;
         if (strcmp(in1,"OR")==0) conc_op = OR;
    }
   tablefile.close();								// close table
   return;
}


// -----------------------------------------------------------------------
// Insert a new CombinedVariable (and its score) in a branch of TEvalTree 
// -----------------------------------------------------------------------
void TEvalTree::insert(CombinedVariable *cv, char *branch, char *place, char *oper)
{
 CombinedVariable *cv_in_1 = Rules->get_cv_list()->find_cv("eval_in_1");
 CombinedVariable *cv_in_2 = Rules->get_cv_list()->find_cv("eval_in_2");
 CombinedVariable *cv_out;

 TEvalTree *run=this->next;
 TEvalTree *last=this;
 if (strcmp(oper,"AND")==0)
    cv_out =  Rules->get_cv_list()->find_cv("eval_out_and");
 else
    cv_out =  Rules->get_cv_list()->find_cv("eval_out_or");

// Aggregate cv to single branches 
 while (run!=NULL)
    if (strcmp(run->get_name(),branch)==0)
        if (run->comp==TRUE)
          {
	   run->scores->insert(cv->get_cv_name(),place,cv->get_crisp_value());
 	   cv_in_1->copy(cv);
    	   cv_in_2->copy(run->result);
   	   cv_in_1->set_cv_name("eval_in_1");
   	   cv_in_2->set_cv_name("eval_in_2");
   	   Rules->evaluate(cv_out);
    	   run->result->copy(cv_out);
    	   run->result->set_cv_name(run->name);
 	   break;
         }
       else
         {
	  run->scores->insert(cv->get_cv_name(),place,cv->get_crisp_value());
    	  run->result->copy(cv);
    	  run->result->set_cv_name(run->name);
	  run->comp = TRUE;
	  break;
	 }
     else
	 {
	  if (run->next==NULL) last = run;
          run=run->next;
	 }
 if (run==NULL) 
   {
    cout<<endl<<"Warning: "<<branch<<" does not exist in EVALTREE "<<this->name;
    cout<<". A default branch is generated ..."<<endl;
    entries++;
    run = new TEvalTree(branch,paras, Rules,NULL);
    run->set_name(branch);
    run->scores->insert(cv->get_cv_name(),place,cv->get_crisp_value());
    run->result->copy(cv);
    run->result->set_cv_name(run->name);
    run->comp = TRUE;
    last->next = run;
   }
// Aggregate all branches to overall value
run = this->next;
if (this->comp==FALSE) 
{
  while((run!=NULL)&&(run->comp==FALSE)) run=run->next;
  if (run!=NULL) 
  {
   this->comp = TRUE;
   this->result->copy(run->result);
   this->result->set_cv_name(this->name);
  }
}
else
  while(run!=NULL)
   {
    if (run->comp == TRUE)
      {
	if (run->c_op == AND)
	   cv_out =  Rules->get_cv_list()->find_cv("eval_out_and");
 	else
 	   cv_out =  Rules->get_cv_list()->find_cv("eval_out_or");
       cv_in_1->copy(this->result);
       cv_in_2->copy(run->result);
       cv_in_1->set_cv_name("eval_in_1");
       cv_in_2->set_cv_name("eval_in_2");
       Rules->evaluate(cv_out);
       this->result->copy(cv_out);
       this->result->set_cv_name(this->name);
      }
    run = run->next;
   }

} 

//RAGGL lots of things to tidy up here (oh my godness!)
//eventually the nasty clean_up should be replaced by
//recursive destructor calls
// -----------------------------------------------------------------------
// Delete the structure TEvalTree
// -----------------------------------------------------------------------
void TEvalTree::clean_up(void)
{
	TEvalTree *run = next;
	TEvalTree *kill;
	while ( ( run != NULL ) && ( entries > 0 ) )
    {
		kill = run;
		run = run->next;
		if (kill->scores) kill->scores->clean_up();
		delete kill->result;
		kill->next = NULL;
		delete kill;
		entries--;
    }
	// RAGGL
	if (this->scores) this->scores->clean_up();
	// RAGGL
	delete this->result;
	this->next = NULL;
	delete this;
}

// -----------------------------------------------------------------------
// Write the structure TEvalTree to a file
// -----------------------------------------------------------------------
void TEvalTree::write(char *file)
{
 TEvalTree *run=this;
 if (entries>0)
 {
  ofstream outfile;
  outfile.open(file);
  outfile<<"/* #: ";
  outfile<<entries<<" */";
  while (run!=NULL)
      {
	outfile<<endl<<"Branch: "<<run->name;
  	outfile<<"  Crisp value: "<<run->result->get_crisp_value()<<endl;
	//run->result->write(&outfile);
	run->scores->write(outfile);
        outfile<<endl<<"----------------------<End of branch: "<<run->name<<">";
	run=run->next;
      }
 outfile.close();
 }
}

// RAGGL
// -----------------------------------------------------------------------
// Write the structure TEvalTree to a vector
// -----------------------------------------------------------------------
ConstraintEvaluations TEvalTree::GetConstraintList()
{
	ConstraintEvaluations ce; 
	TEvalTree *ptr1 = this;
	if (entries > 0)
	{
		// create a vector of all evaluations
		while ( ptr1 != NULL)							// copy items of all Varlists of all branches of EvalTree
		{
			// add branch constraint evaluation
			ConstraintEvaluation c1;					// create ConstraintEvaluation object
			c1.SetName( ptr1->name );					// set values
			c1.SetBranch( "branch" );					// mark element as branch
  			c1.SetFormula( "" );					
			c1.SetValue( ptr1->result->get_crisp_value() );
//			ce.push_back( c1 );
			TVarlist *ptr2 = ptr1->scores;
			while (ptr2 != NULL)						// copy all Varlist items into ConstraintEvaluations vector
			{
				if ( strcmp( ptr2->get_name(), "") != 0)
				{
					// add constraint evaluation
					ConstraintEvaluation c2;				// create ConstraintEvaluation object	
					c2.SetName( ptr2->get_name() );			// set values 
					c2.SetBranch( ptr1->name );
					c2.SetFormula( ptr2->get_formula() );
					c2.SetValue( ptr2->get_value() );
					ce.push_back( c2 );						// store ConstraintEvaluation in vector
				}
				ptr2 = ptr2->get_next();
			}
			ptr1 = ptr1->next;		
		}
	}
	return ce;
}






