/************************************************************************
 * This library is free software; you can redistribute it and/or	*
 * modify it under the terms of the GNU Library General Public		*
 * License as published by the Free Software Foundation; either		*
 * version 2 of the License, or (at your option) any later version.	*
 * 									*
 * This library is distributed in the hope that it will be useful,	*
 * but WITHOUT ANY WARRANTY; without even the implied warranty of	*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU	*
 * Library General Public License for more details.			*
 * 									*
 * You should have received a copy of the GNU Library General Public	*
 * License along with this library; if not, write to the		*
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,		*
 * Boston, MA  02111-1307, USA.						*
 * 									*
 * The GNU Library General Public License file is included in		*
 * Online-Documentation							*
 *									*
 * GNU's Homepage - http://www.gnu.ai.mit.edu/home.html			*
 *									*
 * Questions & comments referring to *FLIP++: please contact us:	*
 *									*
 * Snail-Mail:								*
 *	Dr. Wolfgang SLANY						*
 *	Database and Expert System Group (E184-2)     			*
 *	Institute of Information Systems               			*
 *	Vienna University of Technology					*
 *	Paniglgasse 16, A-1040 Vienna, AUSTRIA				*
 * Telephone:								*
 *	+43-1-58801-6123     (area-code 1 maybe changed to 201 in 1998) *
 * FAX:									*
 *	+43-1-5055304	     (area-code 1 maybe changed to 201 in 1998) *
 * eMail:								*
 *	sflip@dbai.tuwien.ac.at						*
 * WWW:									*
 *	http://www.dbai.tuwien.ac.at/proj/StarFLIP/			*
 *                                                                      *
 * $Id: evaltree.h,v 1.3 1997/06/24 17:27:14 reichel Exp $ *
 ***********************************************************************/
// -----------------------------------------------------------------------
//	What   : evaluation tree for constraint aggregation
//	Project: DynaFLIP++
//	Author : Ulrich Santa
//	Start  : 13.12.1995
//	Last   : 20.07.1996
//  File   : evaltree.h
// -----------------------------------------------------------------------

#if !defined ( EVALTREE_H )
#define EVALTREE_H

#include "flip.h"
#include "expint.h"

class ConstraintEvaluation
{
public:
	char* GetName()			{ return name; };
	void SetName(char *c)	{ strcpy(name, c); };
	char* GetBranch()		{ return branch; };
	void SetBranch(char *c) { strcpy(branch, c); };
	char* GetFormula()		{ return formula; };
	void SetFormula(char *c){ strcpy(formula, c); };
	float GetValue() const	{ return value; };
	void SetValue(float f)	{ value = f; };
	bool operator<(const ConstraintEvaluation& c) { return ( value < c.GetValue() ); };
private:
	char		name[MAX_NAME_LEN];
	char		branch[MAX_NAME_LEN];
	char		formula[MAX_FORMULA_LEN];
	float		value;
};

typedef vector
<
	ConstraintEvaluation
#if defined ( _MSC_VER )
,
allocator<ConstraintEvaluation>
#endif
> ConstraintEvaluations;

class TEvalTree
{
 private:
   char name[MAX_NAME_LEN];
   CombinedVariable *result;
   RuleSet *Rules;
   ParameterSet *paras;
   TVarlist *scores;
   TEvalTree *next;
   unsigned int entries;
   concat_operator c_op;
   unsigned int comp;
 public:
   TEvalTree(void);
   TEvalTree(char *,ParameterSet *, RuleSet *, TVarlist *);
   ~TEvalTree(void) {};
   void set_name(char *n) { strcpy(name,n);};
   void set_rule_set(RuleSet *rs) { Rules = rs;};
   void set_parameter_set(ParameterSet *ps) {paras = ps;};
   RuleSet *get_rule_set(void) {return Rules;};
   char *get_name(void) { return name;};
   float get_score(void) {return result->get_crisp_value();};
   void insert(CombinedVariable *, char *,char *,char*);
   void insert_help_rules(char *);
   void write(char *);

   // RAGGL
   ConstraintEvaluations GetConstraintList();
   // RAGGL

   void clean_up();
};

#endif // !defined ( EVALTREE_H )
