/************************************************************************
 * This library is free software; you can redistribute it and/or	*
 * modify it under the terms of the GNU Library General Public		*
 * License as published by the Free Software Foundation; either		*
 * version 2 of the License, or (at your option) any later version.	*
 * 									*
 * This library is distributed in the hope that it will be useful,	*
 * but WITHOUT ANY WARRANTY; without even the implied warranty of	*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU	*
 * Library General Public License for more details.			*
 * 									*
 * You should have received a copy of the GNU Library General Public	*
 * License along with this library; if not, write to the		*
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,		*
 * Boston, MA  02111-1307, USA.						*
 * 									*
 * The GNU Library General Public License file is included in		*
 * Online-Documentation							*
 *									*
 * GNU's Homepage - http://www.gnu.ai.mit.edu/home.html			*
 *									*
 * Questions & comments referring to *FLIP++: please contact us:	*
 *									*
 * Snail-Mail:								*
 *	Dr. Wolfgang SLANY						*
 *	Database and Expert System Group (E184-2)     			*
 *	Institute of Information Systems               			*
 *	Vienna University of Technology					*
 *	Paniglgasse 16, A-1040 Vienna, AUSTRIA				*
 * Telephone:								*
 *	+43-1-58801-6123     (area-code 1 maybe changed to 201 in 1998) *
 * FAX:									*
 *	+43-1-5055304	     (area-code 1 maybe changed to 201 in 1998) *
 * eMail:								*
 *	sflip@dbai.tuwien.ac.at						*
 * WWW:									*
 *	http://www.dbai.tuwien.ac.at/proj/StarFLIP/			*
 *                                                                      *
 * $Id: expint.cpp,v 1.3 1997/06/24 17:27:15 reichel Exp $ *
 ***********************************************************************/
// -----------------------------------------------------------------------
//	What   : an expression interpreter
//	Project: DynaFLIP++:
//	Author : Ulrich Santa
//	Start  : 13.12.1995
//	Last   : 28.06.1996
//  File   : expint.cpp  
// -----------------------------------------------------------------------

// Operators:		+,-,*,/
// Compares:		=,<,>,!
// Logical:			&,|
// Conditions:		?(expr Compares expr) (expr) (expr)
// Functions:  		SIN(expr),COS(),TAN(),SQR(),SRT(),ABS(),LOG(),MIN(),MAX()
// Aggregators:		SUM{expr[,expr]},MUL{expr[,expr]}
// 					SUM{Varname[i]}/n -> Varname and n will be substituted
// Signs have to be put into brackets (-1), (-(SIN(2))
// -----------------------------------------------------------------------

#include "expint.h"


/*-----------------------------------------------------------------------*/
/* write a list of variables and formulas to a file                      */
/*-----------------------------------------------------------------------*/
int is_delimiter(char c)
{
if((c=='+')||(c=='-')||
	(c=='*')||(c=='/')||
	(c==')')||(c=='(')||
	(c=='{')||(c=='}')||
	(c==',')||(c=='?')||
	(c=='=')||(c=='<')||
	(c=='>')||(c=='!')||
	(c=='&')||(c=='|')||
	(c=='#')||(c=='\0')
  )
  return 1;
else
  return 0;
}


/*-----------------------------------------------------------------------*/
/* write a list of variables and formulas to a file                      */
/*-----------------------------------------------------------------------*/
void TVarlist::write(ofstream &output)
{
	TVarlist *run = this;
	if (entries==0) return;
	output << "/* #: " << entries << " */ ";
	while (run!=NULL)
		{
		 output<<endl<<run->name<<" ";
		 output<<run->formula<<" ";
		 output <<run->value;
		 run=run->next;
		}
}


/*-----------------------------------------------------------------------*/
/* write a list of variables and formulas to a file                      */
/*-----------------------------------------------------------------------*/
void TVarlist::write(char *file_name)
{
ofstream outfile;
outfile.open(file_name,ios::out|ios::trunc);
if(outfile)
  {
	write(outfile);
	outfile.close();
  }
}


/*-----------------------------------------------------------------------*/
/* reads a list of variables and formulas from a file                    */
/*-----------------------------------------------------------------------*/
void TVarlist::read(char *file_name)
{
ifstream infile;
infile.open(file_name,ios::in);
if(infile)
  {
	 read(infile);
	 infile.close();
  }
}


/*-----------------------------------------------------------------------*/
/* read a list of variables and formulas from a file                     */
/*-----------------------------------------------------------------------*/
void TVarlist::read(ifstream &input)
{
	char instr[10];
	char r_name[MAX_NAME_LEN];
	char r_formula[MAX_FORMULA_LEN];
	float r_value = 0;
	int i,n;
	input >> instr; //  /*
	input >> instr; //  #:
	input >> n;
	input >> instr; //  */
	for (i=0; i<n; i++)
	{
		input>>r_name;
		input>>r_formula;
		input>>r_value;
		insert(r_name,r_formula,r_value);
		//set_comp(r_name,TRUE);
	}
}


/*-----------------------------------------------------------------------*/
/* substitue short sum description for all formulas in list of variables */
/*-----------------------------------------------------------------------*/
void TVarlist::substitute_all(void)
{
TVarlist *run = this;
if (entries == 0) return;
while(run!=NULL)
 {
  run->set_formula(substitute(run->get_formula()));
  run=run->next;
 }
}


/*-----------------------------------------------------------------------*/
/* substitue short sum description with real number of elements          */
/*-----------------------------------------------------------------------*/
char *TVarlist::substitute(char *expr)
{

 Tindex indices[10];					// a maximum of 10 indices can be used
 unsigned int active_index = 0;           // next free index field in indices
 unsigned int index_i = 0;                // position pointer in index (i=start:end)
 unsigned int index_start = 0;            // help pointer for lower and upper boundaries
 unsigned int index_name = 0;             // index name such as i,j,x,y,...
 char index_value[10];           // help string for index convertion

 char help[MAX_FORMULA_LEN];     // copy of expression
 char final[MAX_FORMULA_LEN];    // final formula string that will be returned
 char last[MAX_FORMULA_LEN];
 char insert[MAX_FORMULA_LEN];   // for Var[i] substituions
 char subst_name[MAX_NAME_LEN];  // names of real variables
 char index[10];                 // help string for index convertion
 char function[10];              // help string for function detection

 unsigned int i,j,lb,rb;
 unsigned int start, end, real_index;

 for (i=0;i<10;i++)					// reset index array
	 {
	  strcpy(indices[i].name,"");
	  indices[i].start = 0;
	  indices[i].end = 32767;
	  indices[i].real_found = 0;
	 }

strcpy(help,expr);					// we work on help, a copy of expr
strcpy(final,"");                // resulting string after substituition
i = 0;                           // position pointer of char in expr
j = 0;
start = 0;                       // start of substituted variable
end = 0;                         // end of substituted variable
real_index = 0;                  // 0...real index   1...abstract index
while ((help[i]!='\0')&&(i<strlen(help))) // still characters left
{
lb = 0; 			  						// index of '[' in help
rb = 0;                          // index of ']' in help
// Search start of Variable or function name
while ((is_delimiter(help[i]))&&(help[i]!='\0')) // skip delimiters
	strncat(final,&help[i++],1);
if ((strlen(help)-i)>3)				// i points to function name?
	{
	 strncpy(function,&help[i],3);
	 for (j=0;j<3;j++)            // lowercase to uppercase
		  if ((function[j]>='a') && (function[j]<='z'))	function[j]+='A' - 'a';
	 function[3]='\0';
	}
else
	strcpy(function,""); 			// no valid function name detected
if ((strcmp(function,"SQR")==0) ||
	(strcmp(function,"SIN")==0)  ||
	(strcmp(function,"COS")==0)  ||
	(strcmp(function,"TAN")==0)  ||
	(strcmp(function,"ABS")==0)  ||
	(strcmp(function,"SUM")==0)  ||
	(strcmp(function,"MUL")==0)  ||
	(strcmp(function,"LOG")==0)  ||
	(strcmp(function,"MIN")==0)  ||
	(strcmp(function,"MAX")==0)  ||
	(strcmp(function,"SRT")==0))
	{
	 strcat(final,function); 		// append function name
	 i+=3;  								// skip function name
	 index_i = i;
	 if ((strcmp(function,"SUM")==0)  ||
		  (strcmp(function,"MUL")==0)  ||
		  (strcmp(function,"MAX")==0)  ||
		  (strcmp(function,"MIN")==0))
			// does the formula include index restrictions?
		  if ((help[index_i]=='(')&& (help[index_i+2]=='='))
			  {
				index_name = index_i+1;
				index_start= index_i+3;
				while ((help[index_i]!=':')&& (help[index_i]!=')')) index_i++;
				if (help[index_i]==':')	// lower and upper bounds given
				 {
				  strncpy(index_value,&help[index_start],index_i-index_start);
				  index_value[index_i-index_start]='\0';
				  strncpy(indices[active_index].name,&help[index_name],1);
				  indices[active_index].name[1]='\0';
				  indices[active_index].start=atoi(index_value);
				  index_start=index_i+1;
				  while (help[index_i]!=')') index_i++;
				  strncpy(index_value,&help[index_start],index_i-index_start);
				  index_value[index_i-index_start]='\0';
				  indices[active_index].end=atoi(index_value);
				  active_index++;
				  i=index_i+1; // i points to '('
				 }
				else						// only lower bound given
				  {
				  strncpy(index_value,&help[index_start],index_i-index_start);
				  index_value[index_i-index_start]='\0';
				  strncpy(indices[active_index].name,&help[index_name],1);
				  indices[active_index].name[1]='\0';
				  indices[active_index].start=atoi(index_value);
				  active_index++;
				  i=index_i+1; 		// i points to '('
				  }
			  }
	 strncat(final,&help[i],1); 	// append '(' or '{'
	 i++;   								// skip functional bracket
	}
 else // non function
  {
	start = i;  // set start of variable or value
	// find next functional delimiter
	while (!is_delimiter(help[i])) i++;

	if ((help[i]!='\0')||(start!=i))      	// reached end of string?
	 {
	 end = i;                  // set position as end of variable
	 for (j=start;j<end;j++)   // check variable name
		 {
		  if (help[j] == '[') lb = j;
		  if (help[j] == ']') rb = j;
		 }
	 real_index = 1;           // assume a real index is given in Varname[i]
	 // When i in Varname[i] is not a number, set real_index to zero
	 if ((lb!=0)&&(rb!=0))
		{
		 for (j=lb+1;j<rb;j++)
				if (!isdigit(help[j])) real_index*= 0;
		}
	 if (!real_index)          // not a real index given?
		 {
		  strcpy(insert,"");    // this string will hold the substitution
		  // find all variables that match to Varname[i]
		  for(j=0;j<=entries;j++)
			  {
				strncpy(subst_name,&(help[start]),lb-start+1);
				subst_name[lb-start+1] = '\0';
				//itoa (j,index,10);
				sprintf(index,"%d",j);
				strcat(subst_name,index);
				strncat(subst_name,&(help[rb]),end-rb);
				if (in_list(subst_name)!=0) // RAGGL
					 {
					  if (active_index == 0) // no index restrictions
						 {
						  strcat(insert,subst_name); // insert varname[i]
						  strcat(insert,",");
						 }
					  else
					  if ((indices[active_index-1].start<=j)&&
							(indices[active_index-1].end>=j))
							 {
							  indices[active_index-1].real_found++;
							  strcat(insert,subst_name); // insert varname[i]
							  strcat(insert,",");
							 }
					 }
				}
		  // remove last ','
		  if (insert[strlen(insert)-1] == ',') insert[strlen(insert)-1] = '\0';
		  strcat(final,insert);
		 }
	 else
		  strncat(final,&help[start],end-start); // normal Variable name
	 }
  }
}
// substitute #i with the real number of arguments
for(j=0;j<strlen(final);j++)
	if (final[j]=='#')
		 for(i=0;i<10;i++)
			  if (final[j+1]==indices[i].name[0])
				  {
					//itoa(indices[i].real_found,last,10);
					sprintf(last,"%d",indices[i].real_found);
					strcpy(help,&final[j+2]);
					final[j] = '\0';
					strcat(final,last);
					strcat(final,help);
					break;
				  }

j = strlen(final);
for (i=0;i<j;i++)
	expr[i]=final[i];
expr[j]='\0';
return expr;
}


/*-----------------------------------------------------------------------*/
/* add an item to the linear list for variable handling                  */
/*-----------------------------------------------------------------------*/
void TVarlist::insert(char *n, char *f, float v=0)
{
static TVarlist *new_item;
TVarlist *bridge;
if (entries++ == 0) 						// First item
	 {
	  set_name(n);
	  set_formula(f);
	  set_value(v);
	  if (v==0)
		  set_comp(FALSE);
	  else
		  set_comp(TRUE);
	 }
else
	 {
	  bridge = this;
	  while (bridge->next!=NULL) bridge=bridge->next;
	  new_item = new TVarlist(n,f,v);	// create new item
	  if (v==0)
		  new_item->set_comp(FALSE);
	  else
		  new_item->set_comp(TRUE);
	  bridge->next = new_item;
	 }
}


/*-----------------------------------------------------------------------*/
/* display the items with their respective values                        */
/*-----------------------------------------------------------------------*/
void TVarlist::show(void)
{
TVarlist *run = this;
cout<<endl<<endl<<"List of Variables:"<<endl;
if (entries == 0)
	 cout << "empty list ..."<<endl;
else
  while(run!=NULL)
	{
	 cout<<endl<<"Name   : "<<run->name<<endl<<"Formula  : "<<run->formula;
	 cout<<endl<<"Value  : "<<run->value;
	 if (run->computed)
		 cout<<endl<<"Computed : TRUE";
	 else
		 cout<<endl<<"Computed : FALSE";
	 cout<<endl<<"----";
	 //getch();
	 run = run->next;
	}
return;
}

/*-----------------------------------------------------------------------*/
/* display an item with its respective values                            */
/*-----------------------------------------------------------------------*/
void TVarlist::show(char *varname)
{

if (entries == 0)
  cout << "empty list ..."<<endl;
else
  if (in_list(varname))
	{             
	 cout<<endl<<"Name     : "<<varname;
	 cout<<endl<<"Formula  : "<<this->get_formula(varname);
	 cout<<endl<<"Value    : "<<this->get_value(varname);
	 if (this->get_comp(varname))
		 cout<<endl<<"Computed : TRUE";
	 else
		 cout<<endl<<"Computed : FALSE";
	}
return;
}


//RAGGL lots of things to tidy up here (oh my godness!)
//eventually the nasty clean_up should be replaced by
//recursive destructor calls
/*-----------------------------------------------------------------------*/
/* remove all items from list (could be written recursively)             */
/*-----------------------------------------------------------------------*/
TVarlist *TVarlist::clean_up(void)
{
	TVarlist *run = next;
	TVarlist *killer;
	while ( ( run != NULL ) && ( entries > 0 ) )
	{
		killer = run;
		run = run->next;
		delete killer;
		entries--;
	}
	delete this;
	return NULL;
}


/*-----------------------------------------------------------------------*/
/* remove an item from the linear list for variable handling             */
/*-----------------------------------------------------------------------*/
void TVarlist::remove(char *n)
{
TVarlist *run = this;
TVarlist *bridge;
if (strcmp(name,n)==0)
	if (next==NULL)
	 entries = 0;
	else
		 {
			run = next;
			strcpy(name,run->name);
	 strcpy(formula,run->formula);
	 value = run->value;
	 computed = run->computed;
	 next = run->next;
	 if (entries>0) entries--;
	 delete run;
		 }

else
	{
	bridge = run->next;
	while (bridge != NULL)
		{
		 if (strcmp(bridge->name,n)==0)
			 {
			  run->next = bridge->next;
			  delete bridge;
			  if (entries>0) entries--;
			  return;
			 }
		 run = run->next;
		 bridge = run->next;
		}
	}
}


/*-----------------------------------------------------------------------*/
/* is variable stored in list ?                                          */
/*-----------------------------------------------------------------------*/
int TVarlist::in_list(char *n)
{
TVarlist *run = this;
while (run != NULL)
	if (strcmp(run->name,n)==0)
		return 1;
	else
		run = run->next;
return  0;
}


/*-----------------------------------------------------------------------*/
/* get variable stored in list                                           */
/*-----------------------------------------------------------------------*/
TVarlist *TVarlist::get_var(char *n)
{
TVarlist *run = this;
while (run != NULL)
	if (strcmp(run->name,n)==0)
		return run;
	else
		run = run->next;
return  NULL;
}


/*-----------------------------------------------------------------------*/
/* get the formula of a specified item                                   */
/*-----------------------------------------------------------------------*/
char *TVarlist::get_formula(char *n)
{
TVarlist *run = this;
if (entries == 0 ) return NULL;
while (run != NULL)
  if (strcmp(run->name,n)==0)
	  return run->formula;
  else
	 run = run->next;
return NULL;
}


/*-----------------------------------------------------------------------*/
/* get the floating point value of a specified value                     */
/*-----------------------------------------------------------------------*/
float TVarlist::get_value(char *n)
{
TVarlist *run = this;
if (entries == 0 ) return 0;
while (run != NULL)
  if (strcmp(run->name,n)==0)
	  return run->value;
  else
	 run = run->next;
return 0;
}


/*-----------------------------------------------------------------------*/
/* get the status of variable: evaluated - not evaluated                 */
/*-----------------------------------------------------------------------*/
int TVarlist::get_comp(char *n)
{
TVarlist *run = this;
if (entries == 0 ) return 0;
while (run != NULL)
  if (strcmp(run->name,n)==0)
	  return run->computed;
  else
	 run = run->next;
return 0;
}


/*-----------------------------------------------------------------------*/
/* set the status of variable: computed or not computed                  */
/*-----------------------------------------------------------------------*/
void TVarlist::set_comp(char *n, int c)
{
TVarlist *run = this;
if (entries == 0 ) return;
while (run != NULL)
  if (strcmp(run->name,n)==0)
	  {
		run->set_comp(c);
		return;
	  }
  else
	 run = run->next;
}



/*-----------------------------------------------------------------------*/
/* set the formula of a specified item                                   */
/*-----------------------------------------------------------------------*/
void TVarlist::set_formula(char *n, char *f)
{
TVarlist *run = this;
if (entries == 0 ) return;
//real_name(n);
//check_expr(f);
//real_name(f);
while (run != NULL)
  if (strcmp(run->name,n)==0)
	  {
	  run->set_formula(f);
	  return;
	  }
  else
	 run = run->next;
}


/*-----------------------------------------------------------------------*/
/* set the floating point value of a specified value                     */
/*-----------------------------------------------------------------------*/
void TVarlist::set_value(char *n, float v)
{
TVarlist *run = this;
if (entries == 0 ) return;
while (run != NULL)
  if (strcmp(run->name,n)==0)
	  {
		run->set_value(v);
		return;
	  }
  else
	 run = run->next;
}


/*-----------------------------------------------------------------------*/
/* eliminate blanks and check brackets                                   */
/*-----------------------------------------------------------------------*/
char *TVarlist::check_expr(char *expr)
{
unsigned int j,i = 0;
unsigned int lb = 0, rb = 0;
while (i<strlen(expr))
 {
	if (expr[i] == ' ')
		for (j=i;j<strlen(expr);j++) expr[j] = expr[j+1];
	else
	  {
		if (expr[i] == ')') rb++;
		if (expr[i] == '(') lb++;
		i++;
	  }
 }
if (rb!=lb) cout<<endl<<"ERROR: Klammerfehler!"<<endl;
return expr;
}

 

/*-----------------------------------------------------------------------*/
/* substitute indices with real values : NEUE VERSION beliebige []      */
/*-----------------------------------------------------------------------*/
char *TVarlist::real_name(char *expr)
{
unsigned int pos_in_expr=0;
unsigned int pos_in_result = 0;
char result[MAX_FORMULA_LEN];
char real[MAX_FORMULA_LEN];
char real_index[MAX_NAME_LEN];
unsigned int i = 0;
unsigned int lb=0, rb=0;
unsigned int index;
float f_value = 0;
TExptree *Tree;

while (pos_in_expr<strlen(expr))
{
 while ((rb==0)&&(pos_in_expr<strlen(expr))) 
      {
	if (lb==0) result[pos_in_result++] = expr[pos_in_expr];
	if (expr[pos_in_expr] == '[')  lb = pos_in_expr+1;
	if (expr[pos_in_expr] == ']')  rb = pos_in_expr;
	pos_in_expr++;
      }
 if ((lb!=0)&&(rb!=0))
   {
    result[pos_in_result]='\0';
    i = 0;
    while (lb<rb) real[i++] = expr[lb++];
    real[i]='\0';
    if (in_list(real))
      {
       index = (int)get_value(real);
       sprintf(real_index,"%d",index);
       strcat(result,real_index);
       strcat(result,"]");
       pos_in_result++;
       pos_in_result+=strlen(real_index);	
      }
     else
       {
	// Compute expression between [ and ] 
        //real_name(real);				   // further indexing in index?
        Tree = new TExptree(NODE,"","","",0.0,NULL,NULL);  // build up evaluation tree
	Tree->insert_expr(real,0);  			   // insert expression
	f_value=Tree->calculate(this,SMART);		   // evaluate expression
	Tree=Tree->clean_up(Tree);			   // remove evaluation tree
	sprintf(real,"%d",(int)f_value);
        strcat(result,real);
	strcat(result,"]");
	pos_in_result+=strlen(real);
        pos_in_result++;
       }
   }
 lb=0;rb=0;
}
result[pos_in_result]='\0';
strcpy(expr,result);
return expr;
}




/*-----------------------------------------------------------------------*/
/* constructor for a tree-node                                           */
/*-----------------------------------------------------------------------*/
TExptree::TExptree(void)
{
set_type(NODE);
set_value(0.0);
set_variable("");
set_oper("");
set_sign(1);
set_function("");
set_left(NULL);
set_right(NULL);
}


//RAGGL lots of things to tidy up here (oh my godness!)
//eventually the nasty clean_up should be replaced by
//recursive destructor calls
/*-----------------------------------------------------------------------*/
/* clean up the constructed parse tree                                   */
/*-----------------------------------------------------------------------*/
TExptree *TExptree::clean_up(TExptree *node)
{
	if (node == NULL) return NULL;
	if (left!=NULL) left = left->clean_up(left);
	if (right!=NULL) right = right->clean_up(right);
	delete this;
	return NULL;
}


/*-----------------------------------------------------------------------*/
/* print expression from parse tree                                      */
/*-----------------------------------------------------------------------*/
void TExptree::show_tree(int first = 1, int what=LEAF_VALUE)
{
TExptree *help;
if (type != NODE)
	 if ((what == LEAF_VARIABLE)&&( type == LEAF_VARIABLE))
		  if (strcmp(variable,"")!=0)
				cout<<variable;
		  else
				cout<<value;
	 else
		 cout<<value;
else
  {
	if ((left != NULL)&&(strcmp(func,"")==0))
	 {
	  if (first==1) cout<<"(";
	  if (sign == -1) cout<<"-";
	  left->show_tree(1,what);
	 }
	if ((strcmp(func,"SQR")==0)||(strcmp(func,"SRT")==0)||
		(strcmp(func,"SIN")==0)||(strcmp(func,"COS")==0)||
		(strcmp(func,"ABS")==0)||(strcmp(func,"LOG")==0)||
		(strcmp(func,"TAN")==0))
		{
		if (first == 1) cout<<"(";
		cout<<func<<"(";
		if (sign == -1) cout<<"-";
		if (left != NULL) left->show_tree(1,what);
		cout<<oper;
		if (right != NULL) right->show_tree(1,what);
		cout<<")";
		if (first == 1) cout<<")";
		return;
		}
	else
	  if ((strcmp(func,"SUM")==0)||(strcmp(func,"MUL")==0)||
			(strcmp(func,"MIN")==0)||(strcmp(func,"MAX")==0)||
			(strcmp(func,"?")==0))
		  {
			if (first == 1) cout <<"(";
			if (sign == -1) cout<<"-";
			cout<<func<<"{";
			help = this;
			while (help!=NULL)
				{
				 help->left->show_tree(1,what);
				 help = help->right;
				 cout<<",";
				}
		  cout<<"}";
		  if (first == 1) cout<<")";
		  return;
		  }
	  else
		cout<<oper;
  if (right!=NULL)
	 {
	  right->show_tree(1,what);
	  if (first == 1) cout<<")";
	 }
  else
	  if(first == 1) cout<<")";
  }
}


/*-----------------------------------------------------------------------*/
/* get next character from expression string                             */
/*-----------------------------------------------------------------------*/
char *TExptree::get_next(char *expr)
{
int i, l=strlen(expr);
for(i=0;i<l;i++)
 *(expr+i) =*(expr+i+1);
return expr;
}


/*-----------------------------------------------------------------------*/
/* insert mathematical expression into the parse-tree                    */
/*-----------------------------------------------------------------------*/
void TExptree::avoid_recursion(char *var, TVarlist* variables)
{
if (type==NODE)
	{
	 if (left!=NULL) left->avoid_recursion(var,variables);
	 if (right!=NULL) right->avoid_recursion(var,variables);
	}
else
	{
	 if ((type== LEAF_VARIABLE)&&(strcmp(var,variable)==0))
			variables->set_comp(var,TRUE); // same variable on left and right side
	 return;
   }
}


/*-----------------------------------------------------------------------*/
/* insert mathematical expression into the parse-tree                    */
/*-----------------------------------------------------------------------*/
void TExptree::insert_expr(char *expr,int first=1)
{
static TExptree *Nodel,*Noder;
TExptree  *SumNode,*IfNode;
char help[MAX_FORMULA_LEN];
char function[10];
int i = 0, j = 0;
if (strcmp(expr,"") == 0) return; 	// nothing left to parse
// First function call
if (first == 0)                   	// normally at least term op term
	{
	Nodel = new TExptree();				// create left branch
	set_left(Nodel);             		// link this branch to tree
	Nodel->insert_expr(expr);    		// follow left branch
	if (expr[0]==')')
		{
		 expr = get_next(expr);			// skip ')'
		 return;                      // level complete
		}
	if (strcmp(expr,"")==0) return; 	// nothing left for right branch
	set_type(NODE);                 	// set type = operator node
	strcpy(help,expr);
	help[1] = '\0';
	set_oper(help);              	// set operator
	Noder = new TExptree();				// create right branch
	set_right(Noder);               	// link this branch to tree
	expr = get_next(expr);				// skip to next char
	Noder->insert_expr(expr);       	// follow right branch
	if (expr[0]==')') expr = get_next(expr);
	return;
	}

// Function ?
if (strlen(expr)>3)
	{
	 strncpy(function,expr,3);
	 // lowercase characters to uppercase characters
	 for (i=0;i<3;i++)
		if ((function[i]>='a') && (function[i]<='z'))	function[i]+='A' - 'a';
	 function[3]='\0';
	}
else
	strcpy(function,"");
// Condition: ?{(cond_expr),(then_expr),(else_expr)}
if (expr[0]=='?')
	 {
	  set_oper("?");
     set_function("?");
	  expr = get_next(expr); // skip '?'
	  expr = get_next(expr); // skip '{'
	  set_type(NODE);
	  IfNode = this;
	  while(expr[0]!= '}')
		  {
			Nodel = new TExptree();
			IfNode->left = Nodel;
			Nodel->insert_expr(expr); // insert condition, then and else
			if (expr[0]==',') expr = get_next(expr); //skip ','
			if (expr[0] != '}')
				{
				 Noder = new TExptree();
				 IfNode->right = Noder;
				 IfNode = Noder;
				}
		  }
	  expr = get_next(expr);  // skip '}'
	  return;
	 }

// Sum or Product Formula?
if ((strcmp(function,"SUM")==0)||(strcmp(function,"MUL")==0)||
	 (strcmp(function,"MIN")==0)||(strcmp(function,"MAX")==0))
	 {
	  set_function(function);
	  for (i=0;i<=3;i++) expr = get_next(expr); // skip 'SUM{'
	  set_type(NODE);
	  SumNode = this;
	  while(expr[0]!= '}')
		  {
			Nodel = new TExptree();
			SumNode->left = Nodel;
			Nodel->insert_expr(expr);
			if (expr[0]==',') expr = get_next(expr); //skip ','
			if (expr[0] != '}')
				{
				 Noder = new TExptree();
				 SumNode->right = Noder;
				 SumNode = Noder;
				}
		  }
	  expr = get_next(expr);  // skip '}'
	  return;
	 }

if ((expr[0] == '(')||
	(strcmp(function,"SQR")==0) ||
	(strcmp(function,"SIN")==0)  ||
	(strcmp(function,"COS")==0)  ||
	(strcmp(function,"TAN")==0)  ||
	(strcmp(function,"ABS")==0)  ||
	(strcmp(function,"LOG")==0)  ||
	(strcmp(function,"SRT")==0))
	{
	if (expr[0]!= '(')
	 {
	  set_function(function);
	  for(i=0;i<=3;i++) expr = get_next(expr); // skip function and '('
	  set_type(NODE);
	 }
	else
	  {
		expr = get_next(expr); 			// skip to next char
		if (expr[0]=='-')
			{
			 sign = -1;  // sign
			 expr = get_next(expr);
			}
	  }
	Nodel = new TExptree();				// create left branch
	set_left(Nodel);              	// link this branch to tree
	Nodel->insert_expr(expr);    		// follow left branch
	if (expr[0]==')')
		{
		 expr = get_next(expr);       // skip ')'
		 return;								// level comlete
		}
	if (strcmp(expr,"")==0) return; 	// nothing left for right branch
	set_type(NODE);                 	// set type = operator node
	strcpy(help,expr);
	help[1] = '\0';
	set_oper(help);              		// set operator
	Noder = new TExptree();				// create right branch
	set_right(Noder);               	// link this branch to tree
	expr = get_next(expr);				// skip to next char
	Noder->insert_expr(expr);       	// follow right branch
	if (expr[0]==')') expr = get_next(expr); // level complete -> skip ')'
	return;
	}
// We reached a leaf
strcpy(help,expr);					// find value/variable for leaf node
i=0;
while (!is_delimiter(help[i])) i++;
help[i] = '\0';						// end of value/variable in expr
if (isdigit(help[0]))
	{
	 set_value(atof(help));			// set value of leaf
	 set_type(LEAF_VALUE); 			// set type = leaf|value
	}
else
	{
	 set_variable(help);				// set variable of leat
	 set_type(LEAF_VARIABLE);     // set type = leaf|variable
	}
for (j=0;j<i;j++) expr = get_next(expr); // skip to next item
left = NULL;							// no left successor
right = NULL;                    // no right successor
return;
}


/*-----------------------------------------------------------------------*/
/* convert string to numerical equivalent of string                      */
/*-----------------------------------------------------------------------*/
float  TExptree::string_to_float(char *exp_s)
{
float s_value = 0;
char s[MAX_FORMULA_LEN];
strcpy(s,exp_s);
int i,l=strlen(s);
if (s[l-1]=='\"') s[l-1]='\0';				// skip last "
strcpy(s,&s[1]);											// skip first "
l=strlen(s);
for(i=0;i<l;i++)
	s_value=s_value*256+s[i];
return s_value;
}


/*-----------------------------------------------------------------------*/
/* evaluate expression stored in the parse-tree                          */
/*-----------------------------------------------------------------------*/
float  TExptree::calculate(TVarlist *var_list,int base_calc)
{
TExptree *help;
TExptree *var_tree;
int min_1,max_1;
float min,max;
char var_expr[MAX_FORMULA_LEN];
float c_value;
float comp_value1, comp_value2;
// Leaf reached -> return value
if (type == LEAF_VALUE)	return value;
if (type == LEAF_VARIABLE)
	 {
	  if (variable[0]=='"')					// direct string compare
		  {
			set_type(LEAF_VALUE);
			set_value(string_to_float(variable));
			return value;
		  }
	  if (!var_list->in_list(variable))				// variable does not exist
		  {
		    cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
                    return 0;
                  }
	  if ((var_list->get_comp(variable))&&(base_calc==SMART))
		  {
			set_value(var_list->get_value(variable));  // Var already evaluated
			return value;
		  }
	  strcpy(var_expr,var_list->get_formula(variable));
	  if (var_expr[0]=='"')				  // variable holds a string
		  {
			set_type(LEAF_VALUE);
			set_value(string_to_float(var_expr));
			var_list->set_value(variable,value);
			var_list->set_comp(variable,TRUE);
			return value;
		  }
	  else
		  {
			var_tree = new TExptree(NODE,"","","",0.0,NULL,NULL);
			var_tree->insert_expr(var_expr,0);
			set_type(LEAF_VALUE);
			set_value(var_tree->calculate(var_list,base_calc));
			var_tree = var_tree->clean_up(var_tree);
			var_list->set_value(variable,value);
			var_list->set_comp(variable,TRUE);
			return value;
		  }
	 }
// redundant node -> proceed on left branch
if ((strcmp(oper,"")==0)&&(strcmp(func,"")==0))
	if (left!=NULL) 
          { 
           c_value = sign*(left->calculate(var_list,base_calc));
           return c_value;
          }
        else  
           {
            cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
            return 0;
	   }

// Operator Nodes
if (strcmp(oper,"&")==0)
	{
        comp_value1=0;
	comp_value2=0;
	if (left!=NULL) 
           comp_value1=left->calculate(var_list,base_calc);
        else  
           {
            cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
            return 0;
	   }

	if (right!=NULL)
           comp_value2= right->calculate(var_list,base_calc);
        else  
           {
            cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
            return 0;
	   }
	c_value = comp_value1&&comp_value2;
	return c_value;
	}
if (strcmp(oper,"|")==0)
	{
	comp_value1=0;
	comp_value2=0;
	if (left!=NULL) 
	    comp_value1=left->calculate(var_list,base_calc);
	else  
           {
            cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
            return 0;
	   }
	if (right!=NULL) 
	    comp_value2= right->calculate(var_list,base_calc);
	else  
           {
            cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
            return 0;
	   }
	c_value = comp_value1||comp_value2;
	return c_value;
	}
if (strcmp(oper,"?")==0)
	{
	 help = this;
	 c_value = 0;
	 comp_value1=0;
	 comp_value2=0;
	 if (help->left!=NULL) 
            comp_value1 = help->left->calculate(var_list,base_calc);
	 else  
           {
            cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
            return 0;
	   }
	 if  (comp_value1)  
		  if (help->right->left!=NULL)
		      c_value= help->right->left->calculate(var_list,base_calc); // then
                  else  
           	    {
      		      cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
     		      return 0;
		    }
	 else
		  if (help->right->right!=NULL)
		      c_value= help->right->right->left->calculate(var_list,base_calc);//else
		  else  
           	    {
      		      cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
     		      return 0;
		    }
	  return c_value;
	}
if (strcmp(oper,"=")==0)
	{
	comp_value1=0;
	comp_value2=0;
	if (left!=NULL) 
           comp_value1=left->calculate(var_list,base_calc);
        else  
           {
            cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
            return 0;
	   }
	if (right!=NULL) 
           comp_value2= right->calculate(var_list,base_calc);
	else  
           {
            cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
            return 0;
	   }
	if (comp_value1 == comp_value2)
		c_value=1;
	else
		c_value=0;
	}
if (strcmp(oper,"<")==0)
	{
	comp_value1=0;
	comp_value2=0;
	if (left!=NULL) 
	   comp_value1=left->calculate(var_list,base_calc);
	else  
           {
            cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
            return 0;
	   }
	if (right!=NULL) 
	   comp_value2= right->calculate(var_list,base_calc);
	else  
           {
            cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
            return 0;
	   }
	if (comp_value1 < comp_value2)
		c_value=1;
	else
		c_value=0;
	}
if (strcmp(oper,">")==0)
	{
	comp_value1=0;
	comp_value2=0;
	if (left!=NULL) 
	   comp_value1=left->calculate(var_list,base_calc);
	else  
           {
            cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
            return 0;
	   }
	if (right!=NULL) 
	   comp_value2= right->calculate(var_list,base_calc);
	else  
           {
            cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
            return 0;
	   }
	if (comp_value1 > comp_value2)
		c_value=1;
	else
		c_value=0;
	}
if (strcmp(oper,"!")==0)
	{
	comp_value1=0;
	comp_value2=0;
	if (left!=NULL) 
	   comp_value1=left->calculate(var_list,base_calc);
	else  
           {
            cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
            return 0;
	   }
	if (right!=NULL) 
	   comp_value2= right->calculate(var_list,base_calc);
	else  
           {
            cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
            return 0;
	   }
	if (comp_value1 != comp_value2)
		c_value=1;
	else
		c_value=0;
	}
if (strcmp(oper,"+")==0)
	{
        c_value = 0;
	if (left!=NULL) 
	   c_value=left->calculate(var_list,base_calc);
	else  
           {
            cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
            return 0;
	   }
	if (right!=NULL) 
	   c_value+= right->calculate(var_list,base_calc);
	else  
           {
            cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
            return 0;
	   }
	}

if (strcmp(oper,"-")==0)
	{
	c_value = 0;
	if (left!=NULL) 
	   c_value=left->calculate(var_list,base_calc);
	else  
           {
            cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
            return 0;
	   }
	if (right!=NULL) 
	   c_value-= right->calculate(var_list,base_calc);
	else  
           {
            cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
            return 0;
	   }
	}
if (strcmp(oper,"*")==0)
	{
	c_value = 0;
	if (left!=NULL) 
	   c_value=left->calculate(var_list,base_calc);
	else  
           {
            cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
            return 0;
	   }
	if (right!=NULL) 
	   c_value*= right->calculate(var_list,base_calc);
	else  
           {
            cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
            return 0;
	   }
	}
if (strcmp(oper,"/")==0)
	{
	c_value = 0;
	comp_value2 = 0;
	if (left!=NULL) 
	   c_value=left->calculate(var_list,base_calc);
	else  
           {
            cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
            return 0;
	   }
	if (right!=NULL) 
	   comp_value2 = right->calculate(var_list,base_calc);
	else  
           {
            cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
            return 0;
	   }
	if (comp_value2!=0)
		c_value/=comp_value2;
	else
	  {
		cout<<endl<<"ERROR: DIVISION BY ZERO! Program terminated."<<endl;
		exit(0);
	  }
	return c_value;
	}
// Function Nodes
if (strcmp(func,"SIN")==0)
	if (strcmp(oper,"")==0)
	    if (left!=NULL) 
		{c_value=sin(left->calculate(var_list,base_calc));return c_value;}
	    else  
                {
                 cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
                 return 0;
	        }
	else
	  {c_value = sin(c_value);return c_value;};
if (strcmp(func,"COS")==0)
	if (strcmp(oper,"")==0)
	   if (left!=NULL) 
		{c_value=cos(left->calculate(var_list,base_calc));return c_value;}
	    else  
                {
                 cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
                 return 0;
	        }
	else
	  {c_value = cos(c_value);return c_value;};
if (strcmp(func,"TAN")==0)
	if (strcmp(oper,"")==0)
	    if (left!=NULL) {
		c_value=tan(left->calculate(var_list,base_calc));return c_value;}
	    else  
                {
                 cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
                 return 0;
	        }
	else
	  {c_value = tan(c_value);return c_value;};
if (strcmp(func,"SQR")==0)
	if (strcmp(oper,"")==0)
	    if (left!=NULL) 
		{c_value=(left->calculate(var_list,base_calc));return (c_value*=c_value);}
	    else  
                {
                 cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
                 return 0;
	        }
	else
	  {return c_value*=c_value;};
if (strcmp(func,"SRT")==0)
	if (strcmp(oper,"")==0)
	   if (left!=NULL) 
		{c_value=sqrt(left->calculate(var_list,base_calc));return c_value;}
	    else  
                {
                 cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
                 return 0;
	        }
	else
	  {c_value = sqrt(c_value);return c_value;};
if (strcmp(func,"ABS")==0)
	if (strcmp(oper,"")==0)
	   if (left!=NULL) 
		{c_value=fabs(left->calculate(var_list,base_calc));return c_value;}
	    else  
                {
                 cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
                 return 0;
	        }
	else
	  {c_value = fabs(c_value);return c_value;};
if (strcmp(func,"LOG")==0)
	if (strcmp(oper,"")==0)
	   if (left!=NULL) 
		{c_value=log(left->calculate(var_list,base_calc));return c_value;}
	    else  
                {
                 cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
                 return 0;
	        }
	else
	  {c_value = log(c_value);return c_value;};

// Aggregation Nodes
if (strcmp(func,"SUM")==0)
	{
	 help = this;
	 c_value = 0;
	 while (help != NULL)
		 {
		  if (help->left!=NULL) 
			c_value+=help->left->calculate(var_list,base_calc);
	          else  
                     {
                      cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
                      return 0;
	             }
		  help = help->right;
		 }
	  return c_value;
	}
if (strcmp(func,"MUL")==0)
	{
	 help = this;
	 c_value = 1;
	 while (help != NULL)
		 {
		  if (help->left!=NULL) 
			c_value*=help->left->calculate(var_list,base_calc);
	          else  
                     {
                      cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
                      return 0;
	             }

		  help = help->right;
		 }
	 return c_value;
	}
if (strcmp(func,"MIN")==0)
	{
	 help = this;
	 c_value = 0;
	 min_1 = 1;						// first element
	 while (help != NULL)
		 {
		  if (min_1)
			{
			 if (help->left!=NULL) 
			    min = help->left->calculate(var_list,base_calc);
          		 else  
              		   {
               		    cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
               		    return 0;
            		   }
			 min_1 = 0;
			}
		  else
		  {
			 if (help->left!=NULL) 
			    c_value=help->left->calculate(var_list,base_calc);
          		 else  
              		   {
               		    cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
               		    return 0;
            		   }
			 if (min>c_value) min = c_value;
		  }
		  help = help->right;
		 }
	  c_value = min;
	  return c_value;
	}
if (strcmp(func,"MAX")==0)
	{
	 help = this;
	 c_value = 0;
	 max_1 = 1;						// first element
	 while (help != NULL)
	 	{
		  if (max_1)
			{
			 if (help->left!=NULL) 
			   max = help->left->calculate(var_list,base_calc);
          		 else  
              		   {
               		    cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
               		    return 0;
            		   }
			 max_1 = 0;
			}
		  else
		  {
			 if (help->left!=NULL) 
			    c_value=help->left->calculate(var_list,base_calc);
          		 else  
              		   {
               		    cout<<endl<<"DynaFLIP++: error in expression!"<<endl;
               		    return 0;
            		   }
			 if (max<c_value) max = c_value;
		  }
		  help = help->right;
		 }
	  c_value = max;
	  return c_value;
	}

// end of calculate
return c_value;
}







