/************************************************************************
 * This library is free software; you can redistribute it and/or	*
 * modify it under the terms of the GNU Library General Public		*
 * License as published by the Free Software Foundation; either		*
 * version 2 of the License, or (at your option) any later version.	*
 * 									*
 * This library is distributed in the hope that it will be useful,	*
 * but WITHOUT ANY WARRANTY; without even the implied warranty of	*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU	*
 * Library General Public License for more details.			*
 * 									*
 * You should have received a copy of the GNU Library General Public	*
 * License along with this library; if not, write to the		*
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,		*
 * Boston, MA  02111-1307, USA.						*
 * 									*
 * The GNU Library General Public License file is included in		*
 * Online-Documentation							*
 *									*
 * GNU's Homepage - http://www.gnu.ai.mit.edu/home.html			*
 *									*
 * Questions & comments referring to *FLIP++: please contact us:	*
 *									*
 * Snail-Mail:								*
 *	Dr. Wolfgang SLANY						*
 *	Database and Expert System Group (E184-2)     			*
 *	Institute of Information Systems               			*
 *	Vienna University of Technology					*
 *	Paniglgasse 16, A-1040 Vienna, AUSTRIA				*
 * Telephone:								*
 *	+43-1-58801-6123     (area-code 1 maybe changed to 201 in 1998) *
 * FAX:									*
 *	+43-1-5055304	     (area-code 1 maybe changed to 201 in 1998) *
 * eMail:								*
 *	sflip@dbai.tuwien.ac.at						*
 * WWW:									*
 *	http://www.dbai.tuwien.ac.at/proj/StarFLIP/			*
 *                                                                      *
 * $Id: flipcvar.h,v 1.3 1997/06/24 17:27:20 reichel Exp $ *
 ***********************************************************************/
// -----------------------------------------------------------------------
//	What   : Fuzzy Logic Inference Processor 
//	Project: FLIP++
//	Author : Markus Bonner
//  Start  :
//  Last   :
//  File   : flipcvar.h
// -----------------------------------------------------------------------

#if !defined ( FLIPCVAR_H )
#define FLIPCVAR_H

#include "flippara.h"

//
//
//	* * * * * * * * * *  F l o a t V a l u e   * * *
//
//

class FloatValue {

public:
// * * constructors:
	FloatValue(float init_value=0)
		{ value=init_value; next=NULL; };

// * * simple methods to access data elements:
	void set_value(float new_v)         { value=new_v; };
	void set_next(FloatValue *new_next) { next=new_next; };

	float      get_value(void) { return value; };
	FloatValue *get_next(void) { return next; };

protected:
// * * data elements:
	float      value;
	FloatValue *next; };

//
//
//	* * * * * * * * * *  F L i s t  * * *
//
//

class FList {

public:
// * * constructors:
	FList(void) { first=NULL; last=NULL; number=0; };

// * * destructor:
	~FList(void);

// * * complex methods:
	void       add(float new_v=0);
	FloatValue *find(int index);
	float      find_v(int index);
	float      operator [] (int index) { return find_v(index); };
	int        remove(int index);
	void 	   remove_all(void);
	float	   mean(void);

// * * simple methods to access data elements:
	FloatValue *get_first(void) { return first; };
	FloatValue *get_last(void)  { return last; };
	int        get_number(void) { return number; };

protected:
// * * data elements:
	FloatValue *first;
	FloatValue *last;
	int        number; };

//
//
//	* * * * * * * * * *  P o i n t  * * *
//
//

class Point {

friend class PList;

public:
// * * constructors:
	Point(float init_x, float init_y)
		{ x_value=init_x; y_value=init_y; next=NULL; };
	Point() { x_value = 0.0; y_value = -1.0; next=NULL; };

// * * complex methods:
	void show(void);
	int write(ofstream *output);
	int read(ifstream *input);

// * * simple methods to access data elements:
	void set_x(float new_x)        { x_value=new_x; };
	void set_y(float new_y)        { y_value=new_y; };
	void set_next(Point *new_next) { next = new_next; };

	float get_x(void)     { return x_value; };
	float get_y(void)     { return y_value; };
	Point *get_next(void) { return next; };

protected:
// * * data elements:
	float x_value;
	float y_value;
	Point *next;
	};

//
//
//	* * * * * * * * * *  P L i s t  * * *
//
//

class PList {

public:
// * * constructors:
	PList(void) { first=NULL; last=NULL; number=0; };

// * * destructor:
	~PList(void);

// * * complex methods:
	void  show(void);
	int   write(ofstream *output);
	int   read(ifstream *input);
	void  add      (float new_x=0, float new_y=-1);
	void  add_PList(float new_x=0, float new_y=-1)
		{ add(new_x,new_y);};
	Point *find(int index);
	float find_x(int index) { return find(index)->get_x(); };
	float find_y(int index) { return find(index)->get_y(); };
	int   find(float x,float y);
	int   find(Point *p) { return find(p->get_x(),p->get_y()); };
	Point *operator [] (int index) { return find(index); };
	int   insert(float new_x=0, float new_y=-1);
	void  insert(Point *p) { insert(p->get_x(),p->get_y()); };
	void  remove(int index);
	void  remove_all(void);
	void  copy      (PList *p_list);
	void  copy_PList(PList *p_list)     { copy(p_list); };
	float get_max_y(void);
	float get_first_x(float y);
	void  get_x(FList *xl, float y);
	void  clear_horizontal_line(void);
	void  clear_vertical_line(void);
	void  clear_lines(void) { clear_horizontal_line();
				  clear_vertical_line(); };

// * * simple methods to access data elements:
	Point *get_first(void) { return first; };
	Point *get_last(void)  { return last; };
	int   get_number(void) { return number; };

protected:
// * * data elements:
	Point *first;
	Point *last;
	int   number;
	};

//
//
//	* * * * * * * * * *  F u z z y S e t B a s e  * * *
//
//

class FuzzySetBase {
public:
// * * constructors:
	FuzzySetBase(name init_linguistic_term, float init_lower_support_end,
		float init_upper_support_end, colourtype init_fs_colour);
	FuzzySetBase();

// * * complex methods:
	void show(void);
	void show_FSB(void) 		 { show(); };
	int write(ofstream *output);
	int write_FSB(ofstream *output)  { return write(output); };
	int read(ifstream *input);
	int read_FSB(ifstream *input)    { return read(input); };
	void copy(FuzzySetBase *fsb);
	void copy_FSB(FuzzySetBase *fsb) { copy(fsb); };

// * * simple methods to access data elements:
	void set_linguistic_term(char *new_lt)
		{ strcpy(linguistic_term,new_lt); };
	void set_lower_support_end(float new_value)
		{ lower_support_end=new_value; };
	void set_upper_support_end(float new_value)
		{ upper_support_end=new_value; };
	void set_fs_colour(colourtype new_c)
		{ fs_colour.set_colour(new_c); };
	void set_fs_colour(Colour new_c)
		{ fs_colour.set_colour(*(new_c.get_colour())); };

	char *get_linguistic_term(void)   { return linguistic_term; };
	float get_lower_support_end(void) { return lower_support_end; };
	float get_upper_support_end(void) { return upper_support_end; };
	colourtype *get_fs_colour(void)   { return fs_colour.get_colour(); };

protected:
	name   linguistic_term;
	float  lower_support_end;
	float  upper_support_end;
	Colour fs_colour;	};

//
//
//	* * * * * * * * * *   P l i m F u z z y S e t   * * *
//
//
//	P_oint L_ist I_nstanciated M_atched   FuzzySet

class PlimFuzzySet : public FuzzySetBase {
friend class PlimList;

public:
// * * constructors:
	PlimFuzzySet(char *init_linguistic_term, float init_lower_support_end,
		 float init_upper_support_end, colourtype init_fs_colour,
		 name init_function);
	PlimFuzzySet();
	// RAGGL
	virtual ~PlimFuzzySet	( void );
// * * complex methods:
	void show(void);
	int  write(ofstream *output);
	int  read(ifstream *input);
	void calculate_match_value(float crisp_input);
	void build_imsf(void);
	void add_msf_point(float new_x, float new_y);
	int  delete_msf_point(int index);
	int  get_msf_point_number(void)   { return msf.get_number(); };
	int  get_imsf_point_number(void)  { return imsf.get_number(); };
	void copy    (PlimFuzzySet *pfs);
	void copy_PFS(PlimFuzzySet *pfs)  { copy(pfs); };
	FuzzySetBase *get_FSB(void);
	void set_msf_point_x(int index, float new_x);
	void set_msf_point_y(int index, float new_y);

// * * simple methods to access lower class data elements:
	float get_msf_point_x(int index)
		{ return msf[index]->get_x(); };
	float get_msf_point_y(int index)
		{ return msf[index]->get_y(); };
	float get_imsf_point_x(int index)
		{ return imsf[index]->get_x(); };
	float get_imsf_point_y(int index)
		{ return imsf[index]->get_y(); };

// * * simple methods to access data elements:
	void set_match_value(float new_mv) { match_value=new_mv; };
	void set_function(name new_f)      { function.set_name(new_f); };

	float get_match_value(void)        { return match_value; };
	PlimFuzzySet *get_next(void)       { return next; };
	PList *get_msf(void)		   { return &msf; };
	PList *get_imsf(void)		   { return &imsf; };
	name  *get_function(void)	   { return function.get_name(); };

private:
	PList msf;   	// membership function
	PList imsf;   	// instanciated membership function
	float match_value;
	Name  function; // name of msf
	PlimFuzzySet *next; };

//
//
//	* * * * * * * * * *  P l i m L i s t  * * *
//
//

class PlimList {
public:
// * * constructors:
	PlimList(void) { first=NULL; last=NULL; number=0; };

// * * destructor:
	virtual ~PlimList(void);

// * * complex methods:
	void show(void);
	void show_PL(void) 		{ show(); };
	int  write(ofstream *output);
	int  write_PL(ofstream *output) 	{ return write(output); };
	int  read(ifstream *input);
	int  read_PL(ifstream *input) 	{ return read(input); };
	void add(void);
	void add_PFS(void) { add(); };
	void add(name init_name,
	       float lower_support_end,       float upper_support_end,
	       float lower_dilatation_length, float upper_dilatation_length,
	       colourtype colour);
	void add_PFS(name init_name,
	       float lower_support_end,       float upper_support_end,
	       float lower_dilatation_length, float upper_dilatation_length,
	       colourtype colour)
		{ add(init_name, lower_support_end, upper_support_end,
		      lower_dilatation_length, upper_dilatation_length,
		      colour); };
	PlimFuzzySet *find(int index);
	PlimFuzzySet *operator [] (int index) { return find(index); };
	int   get_fs_index(char *linguistic_term);
	float get_membership(float x_value,int fs_index);
	NList *get_names_of_vars(void);
	int   check_match(void);
	void  set_all_mvs(float mv);
	int   add_msf_point(int fs_index, float new_x, float new_y)
		{ find(fs_index)->add_msf_point(new_x, new_y); return 1; };
	int   delete_msf_point(int fs_index, int point_index);
	void  copy         (PlimList *plim_list);
	void  copy_PlimList(PlimList *plim_list) { copy(plim_list); };
	int   put_imsf_into_cover(PList *cover, int imsf_index);

	int   above(float x,float y,float ax,float ay,float bx,float by);
	int   above(Point *p, PList *l);
	int   picture(float *pic,float x,float ax,float ay,float bx,float by);
	int   crosspoint(float *cx,float *cy,
			 float ax1,float ay1,float ax2,float ay2,
			 float bx1,float by1,float bx2,float by2);
	void  set_fs_lower_support_end(int fs_index,float new_value);
	void  set_fs_upper_support_end(int fs_index, float new_value);
	void  remove_all(void);

// * * simple methods to access lower class data elements:
	void set_msf_x(int fs_index, int point_index, float new_x)
		{ find(fs_index)->set_msf_point_x(point_index,new_x); };
	void set_msf_y(int fs_index, int point_index, float new_y)
		{ find(fs_index)->set_msf_point_y(point_index,new_y); };
	void set_fs_match_value(int fs_index, float init_mv)
		{ find(fs_index)->set_match_value(init_mv); };
	void set_fs_colour(int fs_index, Colour new_c)
		{ find(fs_index)->set_fs_colour(*(new_c.get_colour())); };

	int   get_msf_point_number(int fs_index)
		{ return find(fs_index)->msf.get_number(); };
	int   get_imsf_point_number(int fs_index)
		{ return find(fs_index)->imsf.get_number(); };
	float get_msf_x(int fs_index, int point_index)
		{ return find(fs_index)->get_msf_point_x(point_index); };
	float get_msf_y(int fs_index, int point_index)
		{ return find(fs_index)->get_msf_point_y(point_index); };
	float get_imsf_x(int fs_index, int point_index)
		{ return find(fs_index)->get_imsf_point_x(point_index); };
	float get_imsf_y(int fs_index, int point_index)
		{ return find(fs_index)->get_imsf_point_y(point_index); };
	float get_fs_match_value( int fs_index)
		{ return find(fs_index)->get_match_value(); };
	float get_fs_lower_support_end(int fs_index)
		{ return find(fs_index)->get_lower_support_end(); };
	float get_fs_upper_support_end(int fs_index)
		{ return find(fs_index)->get_upper_support_end(); };
	colourtype *get_fs_colour(int fs_index)
		{ return find(fs_index)->get_fs_colour(); };

// * * simple methods to access data elements:
	void set_first(PlimFuzzySet *pfs) { first=pfs; };
	void set_last(PlimFuzzySet *pfs)  { last=pfs; };
	void set_number(int nn)           { number=nn; };

	PlimFuzzySet *get_first(void) { return first; };
	PlimFuzzySet *get_last(void)  { return last; };
	int          get_number(void) { return number; };

protected:
// * * data elements:
	PlimFuzzySet *first;
	PlimFuzzySet *last;
	int number;
	};

//
//
//	* * * * * * * * * *   C o m b i n e d V a r i a b l e    * * *
//
//

class CombinedVariable : public ParameterSetBase, public PlimList {
friend class CVList;

public:
// * * constructors:
	CombinedVariable(void);
	CombinedVariable(ParameterSet *ps, name ncv_name="default_cv");
	virtual ~CombinedVariable(void) { 
	while (first != NULL) {
		PlimFuzzySet *e=first;
		first = e->get_next();
		delete e; }; };
// * * complex methods:
	void show(void);
	int  write(ofstream *output);
	int  read(ifstream *input);
	int  set_value(float crisp_input);
	void copy(CombinedVariable *cv);
	ParameterSetBase *get_PSB(void);
	void copy_PlimList(CombinedVariable *cv);

// * * simple methodes to access data elements
	void  set_cv_name(char *new_cv_name) {strcpy(cv_name,new_cv_name);};
	void  set_crisp_value(float ncv) {crisp_value=ncv; }; // dangerous

	char  *get_cv_name(void)    { return cv_name; };
	float get_crisp_value(void) { return crisp_value; };
	CombinedVariable *get_next(void) { return next; };

protected:
	name 		 cv_name;
	float        	 crisp_value;
	CombinedVariable *next; };

//
//
//	* * * * * * * * * *   C V L i s t   * * *
//
//

class CVList {
public:
// * * constructor:
	CVList(void) { first=NULL; last=NULL; number=0; };

// * * destructor:
	~CVList(void);

// * * complex methods:
	void show(void);
	int  write(ofstream *output);
	int  read(ifstream *input);
	void add(void);
	void add(CombinedVariable *cv);
	void add(CVList *cvl);
	CombinedVariable *find(int index);
	CombinedVariable *find_cv(name cv_name);
	int		 find_cv_by_pointer(CombinedVariable *cv);
	int		 find_cv_by_name   (CombinedVariable *cv);
	CombinedVariable *operator [] (int index) { return find(index); };
	void remove(int index);
	void remove_all(void);

// * * simple methods to access data elements:
	CombinedVariable *get_first(void)       { return first; };
	CombinedVariable *get_last(void)        { return last; };
	int   		 get_number(void)       { return number; };

private:
// * * data elements:
	CombinedVariable *first;
	CombinedVariable *last;
	int number;
	};


//
//
//	* * * * * * * * * *   o t h e r e  f u n c t i o n s   * * *
//
//

float min(float ,float );
float max(float ,float );
void  wait_for_key_press(void);

#endif // !defined ( FLIPCVAR_H )
