/************************************************************************
 * This library is free software; you can redistribute it and/or	*
 * modify it under the terms of the GNU Library General Public		*
 * License as published by the Free Software Foundation; either		*
 * version 2 of the License, or (at your option) any later version.	*
 * 									*
 * This library is distributed in the hope that it will be useful,	*
 * but WITHOUT ANY WARRANTY; without even the implied warranty of	*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU	*
 * Library General Public License for more details.			*
 * 									*
 * You should have received a copy of the GNU Library General Public	*
 * License along with this library; if not, write to the		*
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,		*
 * Boston, MA  02111-1307, USA.						*
 * 									*
 * The GNU Library General Public License file is included in		*
 * Online-Documentation							*
 *									*
 * GNU's Homepage - http://www.gnu.ai.mit.edu/home.html			*
 *									*
 * Questions & comments referring to *FLIP++: please contact us:	*
 *									*
 * Snail-Mail:								*
 *	Dr. Wolfgang SLANY						*
 *	Database and Expert System Group (E184-2)     			*
 *	Institute of Information Systems               			*
 *	Vienna University of Technology					*
 *	Paniglgasse 16, A-1040 Vienna, AUSTRIA				*
 * Telephone:								*
 *	+43-1-58801-6123     (area-code 1 maybe changed to 201 in 1998) *
 * FAX:									*
 *	+43-1-5055304	     (area-code 1 maybe changed to 201 in 1998) *
 * eMail:								*
 *	sflip@dbai.tuwien.ac.at						*
 * WWW:									*
 *	http://www.dbai.tuwien.ac.at/proj/StarFLIP/			*
 *                                                                      *
 * $Id: flippara.cpp,v 1.3 1997/06/24 17:27:23 reichel Exp $ *
 ***********************************************************************/
// -----------------------------------------------------------------------
//	What   : Fuzzy Logic Inference Processor 
//	Project: FLIP++
//	Author : Markus Bonner
//  Start  :
//  Last   :
//  File   : flippara.cpp
// -----------------------------------------------------------------------

#include "flippara.h"

//
//
//	* * * * * * * * * *  N L i s t  * * *
//
//

// * * constructors:

NList::NList(int n_number) {
	first=NULL;
	last=NULL;
	number=0;

	switch (n_number) {
	   case	2: add("not_satisfied"); add("satisfied"); break;
	   case	3: add("bad"); add("neutral");
		   add("good"); break;
	   case	4: add("negative_big"); add("negative");
		   add("positive"); add("positive_big"); break;
	   case	5: add("negative_big"); add("negative");
		   add("zero"); add("positive");
		   add("positive_big"); break;
	   case	6: add("negative_huge"); add("negative_big");
		   add("negative"); add("positive");
		   add("positive_big"); add("positive_huge"); break;
	   case	7: add("negative_huge"); add("negative_big");
		   add("negative"); add("zero");
		   add("positive"); add("positive_big");
		   add("positive_huge"); break;
	   case	8: add("negative_huge"); add("negative_big");
		   add("negative"); add("negative_small");
		   add("positive_small"); add("positive");
		   add("positive_big"); add("positive_huge"); break;
	   case	9: add("negative_huge"); add("negative_big");
		   add("negative"); add("negative_small");
		   add("zero"); add("positive_small");
		   add("positive"); add("positive_big");
		   add("positive_huge"); break;
	   default: for (int i=0; i<n_number; i++) add();
	   }; };

// * * destructor:

 NList::~NList(void) {
	while (first != NULL) {
		Name *n=first;
		first = n->get_next();
		delete n; }; };

// * * complex methods:

void NList::show(void) {
	Name *element = first;
	while (element != NULL) {
		element->show();
		element = element->next; };
	cout << endl; };

int NList::write(ofstream *output) {
	Name *element = first;
	*output << "/* #: " << number << " */ ";
	while (element != NULL) {
		element->write(output);
		element = element->next; };
	*output << endl;
	return 1; };

int NList::read(ifstream *input) {
	char instr[10];
	int n_number;
	*input >> instr; //  /*
	*input >> instr; //  #:
	*input >> n_number;
	*input >> instr; //  */
	for (int i=0; i<n_number; i++) {
		add();
		last->read(input); };
	return 1; };

void NList::add(name new_name) {
	Name *element=new Name;
	strcpy(element->nam,new_name);
	if (first == NULL ) { first=element;      last=element; }
	else 		    { last->next=element; last=element; };
	number++; };

Name *NList::find(int index) {
	Name *element=first;
	for (int i=0; i<index; i++) element=element->next;
	return element; };

void NList::copy(NList *n_list) {
	remove_all();
	Name *element=n_list->first;
	for (int i=0; i<n_list->number; i++) {
		add(element->nam);
		element=element->next; }; };

void NList::remove_all(void) {
	while (first != NULL) {
		Name *element=first;
		first = element->get_next();
		delete element; };
	first=NULL;
	last=NULL;
	number=0; };

//
//
//	* * * * * * * * * *   C o l o u r   * * *
//
//

// * * complex methods:

void Colour::show(void) {
	char colourstr[20];
	switch (colour) {
		case violet:	  strcpy(colourstr,"violet"); break;
		case blue:	  strcpy(colourstr,"blue"); break;
		case light_blue:  strcpy(colourstr,"light_blue"); break;
		case turquoise:	  strcpy(colourstr,"turquoise"); break;
		case green:	  strcpy(colourstr,"green"); break;
		case yellow:	  strcpy(colourstr,"yellow"); break;
		case orange:	  strcpy(colourstr,"orange"); break;
		case red:	  strcpy(colourstr,"red"); break;
		case pink:	  strcpy(colourstr,"pink"); break;
		case white:	  strcpy(colourstr,"white"); break;
		case black:	  strcpy(colourstr,"black"); break;
		case grey:	  strcpy(colourstr,"grey"); break;
		case light_grey:  strcpy(colourstr,"light_grey"); break;
		case dark_grey:	  strcpy(colourstr,"dark_grey"); break;
		case dark_blue:	  strcpy(colourstr,"dark_blue"); break;
		case light_green: strcpy(colourstr,"light_green"); break;
		case dark_green:  strcpy(colourstr,"dark_green"); break;
		case light_red:	  strcpy(colourstr,"light_red"); break;
		case dark_red:	  strcpy(colourstr,"dark_red"); break;
		case brown:	  strcpy(colourstr,"brown");
		default:          strcpy(colourstr,"no_known_colour"); };
	cout << colourstr << space; };

int Colour::write(ofstream *output) {
	char colourstr[20];
	switch (colour) {
		case violet:	strcpy(colourstr,"violet"); break;
		case blue:	strcpy(colourstr,"blue"); break;
		case light_blue: strcpy(colourstr,"light_blue"); break;
		case turquoise:	strcpy(colourstr,"turquoise"); break;
		case green:	strcpy(colourstr,"green"); break;
		case yellow:	strcpy(colourstr,"yellow"); break;
		case orange:	strcpy(colourstr,"orange"); break;
		case red:	strcpy(colourstr,"red"); break;
		case pink:	strcpy(colourstr,"pink"); break;
		case white:	strcpy(colourstr,"white"); break;
		case black:	strcpy(colourstr,"black"); break;
		case grey:	strcpy(colourstr,"grey"); break;
		case light_grey:  strcpy(colourstr,"light_grey"); break;
		case dark_grey:	strcpy(colourstr,"dark_grey"); break;
		case dark_blue:	strcpy(colourstr,"dark_blue"); break;
		case light_green: strcpy(colourstr,"light_green"); break;
		case dark_green:  strcpy(colourstr,"dark_green"); break;
		case light_red:	strcpy(colourstr,"light_red"); break;
		case dark_red:	strcpy(colourstr,"dark_red"); break;
		case brown:	strcpy(colourstr,"brown"); };
	*output << colourstr << space;
	return 1; };

int Colour::read(ifstream *input) {
	char colourstr[20];
	*input >> colourstr;
	if (strcmp(colourstr,"violet")==0)	colour=violet;
	if (strcmp(colourstr,"blue")==0) 	colour=blue;
	if (strcmp(colourstr,"light_blue")==0)  colour=light_blue;
	if (strcmp(colourstr,"turquoise")==0)   colour=turquoise;
	if (strcmp(colourstr,"green")==0) 	colour=green;
	if (strcmp(colourstr,"yellow")==0) 	colour=yellow;
	if (strcmp(colourstr,"orange")==0) 	colour=orange;
	if (strcmp(colourstr,"red")==0) 	colour=red;
	if (strcmp(colourstr,"pink")==0) 	colour=pink;
	if (strcmp(colourstr,"white")==0) 	colour=white;
	if (strcmp(colourstr,"black")==0) 	colour=black;
	if (strcmp(colourstr,"grey")==0) 	colour=grey;
	if (strcmp(colourstr,"light_grey")==0) 	colour=light_grey;
	if (strcmp(colourstr,"dark_grey")==0) 	colour=dark_grey;
	if (strcmp(colourstr,"dark_blue")==0) 	colour=dark_blue;
	if (strcmp(colourstr,"light_green")==0) colour=light_green;
	if (strcmp(colourstr,"dark_green")==0) 	colour=dark_green;
	if (strcmp(colourstr,"light_red")==0) 	colour=light_red;
	if (strcmp(colourstr,"dark_red")==0) 	colour=dark_red;
	if (strcmp(colourstr,"brown")==0) 	colour=brown;
	return 1; };

//
//
//	* * * * * * * * * *   C L i s t   * * *
//
//

// * * constructors:

CList::CList(int c_number) {
	first=NULL;
	last=NULL;
	number=0;

	switch (c_number) {
	   case	2: add(blue); add(red); break;
	   case	3: add(blue); add(green);
		   add(red); break;
	   case	4: add(blue); add(green);
		   add(yellow); add(red); break;
	   case	5: add(blue); add(turquoise);
		   add(green); add(yellow);
		   add(red); break;
	   case	6: add(blue); add(turquoise);
		   add(green); add(yellow);
		   add(orange); add(red); break;
	   case	7: add(blue); add(light_blue);
		   add(turquoise); add(green);
		   add(yellow); add(orange);
		   add(red); break;
	   case	8: add(blue); add(light_blue);
		   add(turquoise); add(green);
		   add(yellow); add(orange);
		   add(red); add(pink); break;
	   case	9: add(violet); add(blue);
		   add(light_blue); add(green);
		   add(turquoise); add(yellow);
		   add(orange); add(red);
		   add(pink); break;
	   default: for (int i=0; i<c_number; i++) add();
	   }; };

// * * destructor:

 CList::~CList(void) {
	while (first != NULL) {
		Colour *n=first;
		first = n->get_next();
		delete n; }; };


// * * complex methods:

void CList::show(void) {
	Colour *element = first;
	while (element != NULL) {
		element->show();
		element = element->next; };
	cout << endl; };

int CList::write(ofstream *output) {
	Colour *element = first;
	*output << "/* #: " << number << " */ ";
	while (element != NULL) {
		element->write(output);
		element = element->next; };
	*output << endl;
	return 1; };

int CList::read(ifstream *input) {
	char instr[10];
	int c_number;
	*input >> instr; //  /*
	*input >> instr; //  #:
	*input >> c_number;
	*input >> instr; //  */
	for (int i=0; i<c_number; i++) {
		add();
		last->read(input); };
	return 1; };

void CList::add(colourtype new_colour) {
	Colour *element=new Colour;
	element->colour=new_colour;
	if (first == NULL ) { first=element;      last=element; }
	else 		    { last->next=element; last=element; };
	number++; };

void CList::copy(CList *c_list) {
	remove_all();
	Colour *element=c_list->get_first();
	for (int i=0; i<c_list->get_number(); i++) {
		add(*(element->get_colour()));
		element=element->get_next(); }; };

Colour *CList::find(int index) {
	Colour *element = first;
	if (index < number) {
		for (int i=0; i<index; i++)
			element=element->next;
		return element; }
	else 	return NULL; };

void CList::remove_all(void) {
	while (first != NULL) {
		Colour *element=first;
		first = element->get_next();
		delete element; };
	first=NULL;
	last=NULL;
	number=0; };

//
//
//	* * * * * * * * * *   P a r a m e t e r S e t B a s e   * * *
//
//

// * * constructor:

ParameterSetBase::ParameterSetBase(float new_range_lower_end,
		float new_range_upper_end, float new_median,
		float new_distribution,    float new_dilatation,
		float new_importance ) {
	range_lower_end=new_range_lower_end;
	range_upper_end=new_range_upper_end;
	if (range_lower_end>range_upper_end) {
		float rle=range_lower_end;
		range_lower_end=range_upper_end;
		range_upper_end=rle;
		cerr <<
		"Warning: Range was in wrong order in ParameterSet!" 
		<< endl; };
	median=new_median;
	if (median<range_lower_end || median>range_upper_end) {
		median=range_lower_end+(range_upper_end-range_lower_end)/2;
		cerr <<
		"Warning: Median was out of range in ParameterSet!"
		<< endl; };
	distribution=new_distribution;
	if (distribution<0.1) {
		distribution=0.1;
		cerr <<
		"Warning: Distribution was below [0.1,1] in ParameterSet!"
		<< endl; };
	if (distribution>1) {
		distribution=1;
		cerr <<
		"Warning: Distribution was above [0.1,1] in ParameterSet!"
		<< endl; };
	dilatation=new_dilatation;
	if (dilatation<0) {
		dilatation=0;
		cerr <<
		"Warning: Dilatation was below [0,1] in ParameterSet!"
		<< endl; };
	if (dilatation>1) {
		dilatation=1;
		cerr <<
		"Warning: Dilatation was above [0,1] in ParameterSet!"
		<< endl; };
	importance=new_importance;
	if (importance<0) {
		importance=0;
		cerr <<
		"Warning: Importance was below [0,1] in ParameterSet!"
		<< endl; };
	if (importance>1) {
		importance=1;
		cerr <<
		"Warning: Importance was above [0,1] in ParameterSet!"
		<< endl; }; };

// * * complex methods:

void ParameterSetBase::show(void) {
	cout << "range_lower_end: "  << range_lower_end << tab
	     << "range_upper_end: " << range_upper_end << endl
	     << "median: " << median << tab
	     << "distribution: " << distribution << endl
	     << "dilatation: " << dilatation << tab
	     << "importance: " << importance << endl; };

int ParameterSetBase::write(ofstream *output) {
	*output << range_lower_end << space << range_upper_end << space
		<< median << space << distribution << space
		<< dilatation << space << importance << endl;
	return 1; };

int ParameterSetBase::read(ifstream *input) {
	*input >> range_lower_end >> range_upper_end
		>> median >> distribution >> dilatation >> importance;
	return 1; };

void ParameterSetBase::copy(ParameterSetBase *psb) {
	range_lower_end = psb->get_range_lower_end();
	range_upper_end = psb->get_range_upper_end();
	median          = psb->get_median();
	distribution    = psb->get_distribution();
	dilatation      = psb->get_dilatation();
	importance      = psb->get_importance(); };

// * * simple methods to access data elements:

void ParameterSetBase::set_distribution(float nd) {
	distribution=nd;
	if (distribution<0.1) {
		distribution=0.1;
		cerr <<
		"Warning: Distribution was below [0.1,1] in ParameterSet!"
		<< endl; };
	if (distribution>1) {
		distribution=1;
		cerr <<
	        "Warning: Distribution was above [0.1,1] in ParameterSet!"
		<< endl; }; };


void ParameterSetBase::set_dilatation(float nd) {
	dilatation=nd;
	if (dilatation<0) {
		dilatation=0;
		cerr <<
		"Warning: Dilatation was below [0,1] in ParameterSet!"
		<< endl; };
	if (dilatation>1) {
		dilatation=1;
		cerr <<
		"Warning: Dilatation was above [0,1] in ParameterSet!"
		<< endl; }; };

void ParameterSetBase::set_importance(float ni) {
	importance=ni;
	if (importance<0) {
		importance=0;
		cerr <<
		"Warning: Importance was below [0,1] in ParameterSet!"
		<< endl; };
	if (importance>1) {
		importance=1;
		cerr <<
		"Warning: Importance was above [0,1] in ParameterSet!"
		<< endl; }; };


//
//
//	* * * * * * * * * *   P a r a m e t e r S e t   * * *
//
//
// * * constructor:

ParameterSet::ParameterSet() : ParameterSetBase(), CList(), NList() {
	strcpy(ps_name,"default_ParameterSet");
	next=NULL; };

ParameterSet::ParameterSet(name new_ps_name, float new_range_lower_end,
		 float new_range_upper_end, float new_median,
		 float new_distribution, float new_dilatation,
		 float new_importance,
		 CList *new_c_list, NList *new_n_list) :
		 ParameterSetBase(new_range_lower_end, new_range_upper_end,
			     new_median, new_distribution, new_dilatation,
			     new_importance),
		 CList(), NList()  {
	strcpy(ps_name,new_ps_name);
	copy_CList(new_c_list);
	copy_NList(new_n_list);
	next=NULL; };

// * * complex methods:

void ParameterSet::show(void) {
	cout << "ParameterSet_name: " << ps_name << endl;
	show_PSB();
	show_CList();
	show_NList();
	cout << endl;
	};

int ParameterSet::write(ofstream *output) {
	*output << ps_name << endl;
	write_PSB(output);
	write_CList(output);
	write_NList(output);
	*output << endl;
	return 1; };

ParameterSet *ParameterSet::read(ifstream *input) {
	*input >> ps_name;
	read_PSB(input);
	read_CList(input);
	read_NList(input);
	return this; };

void ParameterSet::copy(ParameterSet *ps) {
	strcpy(ps_name,ps->get_ps_name());
	ParameterSetBase *ppsb = ps->get_PSB();
	copy_PSB( ppsb );
	free( ppsb );
	CList *pcl = ps->get_CList();
	copy_CList( pcl );
	free( pcl );
	NList *pnl = ps->get_NList();
	copy_NList( pnl );
	free( pnl );
	next=ps->get_next(); };

// * * simple methods to access data elements:

ParameterSetBase *ParameterSet::get_PSB(void) {
	ParameterSetBase *psb = new ParameterSetBase;
	psb->set_range_lower_end(range_lower_end);
	psb->set_range_upper_end(range_upper_end);
	psb->set_median(median);
	psb->set_distribution(distribution);
	psb->set_dilatation(dilatation);
	psb->set_importance(importance);
	return psb; };

CList* ParameterSet::get_CList(void) {
	CList* cl = new CList();;
	cl->set_first(get_first_CList());
	cl->set_last (get_last_CList());
	cl->set_number(get_number_CList());
	return cl; };

NList* ParameterSet::get_NList(void) {
	NList* nl = new NList();
	nl->set_first(get_first_NList());
	nl->set_last (get_last_NList());
	nl->set_number(get_number_NList());
	return nl; };

//
//
//	* * * * * * * * * *  P S L i s t  * * *
//
//

// * * destructor:

 PSList::~PSList(void) {
	while (first != NULL) {
		ParameterSet *ps=first;
		first = ps->get_next();
		delete ps; }; };

// * * complex methods:

void PSList::show(void) {
	ParameterSet *element=first;
	while (element != NULL) {
		element->show();
		element=element->next; };
	cout << endl; };

int PSList::write(ofstream *output) {
	ParameterSet *element=first;
	*output << "/* #: " << number << " */ " << endl << endl;
	while (element != NULL) {
		element->write(output);
		element=element->next; };
	*output << endl;
	return 1; };

int PSList::read(ifstream *input) {
	char instr[10];
	int n;
	*input >> instr; //  /*
	*input >> instr; //  #:
	*input >> n;
	*input >> instr; //  */
	for (int i=0; i<n; i++) {
		add();
		last->read(input); };
	return 1; };

void PSList::add(void) {
	ParameterSet *element=new ParameterSet;
	if (first == NULL ) { first=element;      last=element; }
	else 		    { last->next=element; last=element; };
	number++; };

void PSList::add(ParameterSet *ps) {
	ParameterSet *element=new ParameterSet;
	element->copy(ps);
	if (first == NULL ) { first=element;      last=element; }
	else 		    { last->next=element; last=element; };
	number++; };

void PSList::add(PSList *psl) {
	ParameterSet *element=psl->get_first();
	while (element != NULL) {
		add(element);
		element=element->next; }; };

ParameterSet *PSList::find(int index) {
	ParameterSet *element=first;
	if (index < number) {
		for (int i=0; i<index; i++)
			element=element->next;
		return element; }
	else 	return NULL; };

void PSList::remove(int index) {
	ParameterSet *element;
	ParameterSet *to_delete;
	if (index==0) { to_delete=first;
			first=first->next; }
	else { element=find(index-1);
	       to_delete=element->next;
	       element->next=element->next->next; };
	number--;
	delete to_delete; };





