/************************************************************************
 * This library is free software; you can redistribute it and/or	*
 * modify it under the terms of the GNU Library General Public		*
 * License as published by the Free Software Foundation; either		*
 * version 2 of the License, or (at your option) any later version.	*
 * 									*
 * This library is distributed in the hope that it will be useful,	*
 * but WITHOUT ANY WARRANTY; without even the implied warranty of	*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU	*
 * Library General Public License for more details.			*
 * 									*
 * You should have received a copy of the GNU Library General Public	*
 * License along with this library; if not, write to the		*
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,		*
 * Boston, MA  02111-1307, USA.						*
 * 									*
 * The GNU Library General Public License file is included in		*
 * Online-Documentation							*
 *									*
 * GNU's Homepage - http://www.gnu.ai.mit.edu/home.html			*
 *									*
 * Questions & comments referring to *FLIP++: please contact us:	*
 *									*
 * Snail-Mail:								*
 *	Dr. Wolfgang SLANY						*
 *	Database and Expert System Group (E184-2)     			*
 *	Institute of Information Systems               			*
 *	Vienna University of Technology					*
 *	Paniglgasse 16, A-1040 Vienna, AUSTRIA				*
 * Telephone:								*
 *	+43-1-58801-6123     (area-code 1 maybe changed to 201 in 1998) *
 * FAX:									*
 *	+43-1-5055304	     (area-code 1 maybe changed to 201 in 1998) *
 * eMail:								*
 *	sflip@dbai.tuwien.ac.at						*
 * WWW:									*
 *	http://www.dbai.tuwien.ac.at/proj/StarFLIP/			*
 *                                                                      *
 * $Id: inter.cpp,v 1.3 1997/06/24 17:27:29 reichel Exp $ *
 ***********************************************************************/
 // ----------------------------------------------------------------------------------------------------------
//	What   : an evaluation script interpreter
//	Project: DynaFLIP++
//	Author : Ulrich Santa, Andreas Raggl
//	Start  : 11.03.1995
//	Last   : 14.01.1997
// 	File   : inter.cpp
//	Comment: complete functionaliy implemented 
// ----------------------------------------------------------------------------------------------------------

#include "inter.h"


// ----------------------------------------------------------------------------------------------------------
// name:	interpret
// parameter:	file_name
// what: 	script interpreter
// return:	none so far
// ----------------------------------------------------------------------------------------------------------

TEvalTree* interpret(char *pText, TVarlist *importvars)
{
	TEvalTree *Evals = NULL;			// Evaluation Tree to store single satisfaction scores
	// RAGGL
	TVarlist *Vars = importvars;		// list for data structures
	// RAGGL
	TVarlist *Controls = new TVarlist();// list for flow control variables
	char token[MAX_FORMULA_LEN];		// string for next token read from ifstream
	char *tok;							// pointer for string handling
	char varname[MAX_NAME_LEN];			// name of variables on LET varname = ... etc.
	char ifname[MAX_NAME_LEN];			// name of actual IF:if_cascade for control
	char forname[MAX_NAME_LEN];			// name of actual FOR:for_cascade for control
	char formula[MAX_FORMULA_LEN];		// string for formulas on LET varname = formula etc.
	char loopname[MAX_NAME_LEN];		// name of actual LOOP:loop_cascade for control
	char labelname[MAX_NAME_LEN];		// name of existing label for GOTO label statements
	TExptree *Tree=NULL;				// parse and evaluation tree for expressions
	float if_cond;						// result of IF (if_cond) THEN ... ELSE ... ENDIF
	float for_cond=0;					// result of FOR var = start TO end STEP step
	float f_value;						// float value for evaluated expressions
	int if_cascade = 0;					// number of actual cascaded IF structures
	int for_cascade = 0;				// number of actual cascaded FOR structures
	int loop_cascade = 0;				// number of actual cascaded LOOP structures
	int proc_cascade = 0;				// number of actual cascaded LOOP structures
	int number_of_com;					// number inner comment brackets
	int branch = 0;						// control variable holing B_THEN or B_ELSE on IF structures
	int number_of_ifs;					// number of completed IF - statements
	int number_of_skipped_loops = 0;	// number of inner LOOPs to skip when EXITLOOP encountered
	int label_found;					// result of label searchh
	int i;								// variable for loops etc.
	int l;								// variable for strlen() etc.
	int number_of_tokens=0;				// number of tokens already read
	float step_value = 0;				// float value for STEP in FOR - loops
										// ParameterSet variables:
	char ps_name[MAX_NAME_LEN];			// name of ParameterSet
	float lower;						// lower support end
	float upper;						// upper support end
	float median;						// median of ParameterSet
	float distribution;					// distribution of ParameterSet
	float dilatation;					// dilatation of ParameterSet
	float importance;					// importance of ParameterSet
	CList *cl;							// Color list pointer
	NList *nl;							// Name list pointer
	ParameterSet *ps;					// ParameterSet pointer
	RuleSet *Rules;						// RuleSet itself
	CombinedVariable *cv;				// pointer to a CombinedVariable
	char cv_name[MAX_NAME_LEN];			// name of a CombindedVariable
	char tablename[MAX_NAME_LEN];		// name of table used to generate rules for RuleSet


// ----------------------------------------------------------------------------------------------------------
// Find all LABEL and PROCEDURE tags and save position in 
// ----------------------------------------------------------------------------------------------------------

	istrstream infile(pText, 0);
	while(infile.eof()==0)		// Search all LABEL and PROCEDURE statements
	{
		infile>>token;
		if (strcmp(token,"/*")==0)  	// reached the begin of a comment -> skip to the end of comment
		{
			number_of_com = 1;
			while (number_of_com!=0)       // skip all inner comment brackets  
			{
				infile>>token;
				if (strcmp("/*",token)==0) number_of_com++;
				if (strcmp("*/",token)==0) number_of_com--;
			}
		}
		if (strcmp(token,"LABEL")==0)	// LABEL statement reached
		{
			infile>>token;			// get label name
			strcpy(formula,"1");
			strcpy(labelname,token);
			if (!Controls->in_list(labelname))
				Controls->insert(labelname,formula,0);
			else
				cout<<endl<<"Warning: Label "<<labelname<<" does exitst more than one time ..."<<endl;
		}
		if (strcmp(token,"PROCEDURE")==0)  	// PROCEDURE statement reached
		{
			infile>>token;			// get procedure name
			strcpy(formula,"1");
			strcpy(labelname,token);
			if (!Controls->in_list(labelname))
				Controls->insert(labelname,formula,0);
			else
				cout<<endl<<"Warning: Procedure "<<labelname<<" does exitst more than one time ..."<<endl;
		}
	}
	
// ----------------------------------------------------------------------------------------------------------
// Start Programm
// ----------------------------------------------------------------------------------------------------------

	strcpy(tablename,"empty");
	Rules = new RuleSet();
	Rules->set_rs_name("RULESET");

	infile.clear();					// reset flags
	infile.seekg(0);				// return to begin of script
	while(infile.eof()==0)			// not all tokens read from file
	{
		infile >> token;			// get next token
		number_of_tokens++;

// ----------------------------------------------------------------------------------------------------------
// Comment brackets /* ... */
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"/*")==0)  	// reached the begin of a comment -> skip to the end of comment
		{
			number_of_com = 1;
			while (number_of_com!=0)       // skip all inner comment brackets  
			{
				infile>>token;
				number_of_tokens++;
				if (strcmp("/*",token)==0) number_of_com++;
				if (strcmp("*/",token)==0) number_of_com--;
			}
			strcpy(token,"");
			continue;
		}

// ----------------------------------------------------------------------------------------------------------
// LET var = expression
// ----------------------------------------------------------------------------------------------------------
		if  (strcmp(token,"LET")==0) 	// reached an assignement statement LET var = ...
		{
			infile>>varname; 		// read name of variable
			number_of_tokens++;
			infile>>token;	   		// skip = or :
         
			number_of_tokens++;
			if ((strcmp(token,"=")!=0)&&(strcmp(token,":")!=0))
			{
				cout<<endl<<"Error: '= or :' missing in LET "<<varname<<" =|:";
				cout<<token<<" at token #"<<number_of_tokens<<endl;
				break;
			}
	 
			infile>>formula;  		// read formula
			number_of_tokens++;
	
			Vars->real_name(varname);	// indexed name as e.g. Var[i], i==10, -> Var[10]
			if (strcmp(token,"=")==0)	 
				Vars->real_name(formula);	// indexed names in formula as e.g. Var[i]+12, i==10, -> Var[10]+12

			if(!Vars->in_list(varname))
				Vars->insert(varname,formula,0);
			else
			{
				Vars->set_formula(varname,formula);
				Vars->set_comp(varname,TRUE);
			}

			if (strcmp(token,":")==0)	// LET var : SUM{...} -> substitute open indices 
				strcpy(formula,Vars->substitute(Vars->get_formula(varname)));

			Tree = new TExptree(NODE,"","","",0.0,NULL,NULL); // build up evaluation tree
			Tree->insert_expr(formula,0);  		// insert expression
			f_value=Tree->calculate(Vars,SMART);		// evaluate expression
			Tree=Tree->clean_up(Tree);			// remove evaluation tree
	 
			Vars->set_value(varname,f_value);		// set varname to calculated value of formula
			Vars->set_comp(varname,TRUE);			// computed flag = TRUE
			strcpy(token,"");
			continue;
		}

// ----------------------------------------------------------------------------------------------------------
// PRINT expression
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"PRINT")==0)	// PRINT statement reached
		{
			infile>>token;			// get argument of PRINT
			number_of_tokens++;
			if (token[0]=='"')
			{
				l = strlen(token);
				if (l==1) 
				{
					cout<<" ";
					infile>>token;
					number_of_tokens++;
					tok = token;
					l = strlen(token);
				}
				else
					tok = &token[1]; 	// "string" reached
				while  (token[l-1]!='"')  
				{
					cout<<tok<<" ";
					infile>>token;
					l = strlen(token);
					number_of_tokens++;
					tok = token;
				}
				if (l!=1)
				{
	 				if (token[0]=='"') tok = &token[1];
					token[l-1]='\0';
					cout<<tok;		// display string without ""
				}
			}
			else
			if (strcmp(token,"CR")==0) // newline argument reached
				cout<<endl;
			else
			{
				Vars->real_name(token); // EXAMPLE: n=10; PRINT Var[n] -> PRINT Var[10]
				if(Vars->in_list(token))		  // does the variable exist?
			    {
					strcpy(formula,Vars->get_formula(token));
					if (formula[0]=='"')
					{
						formula[strlen(formula)-1]='\0';
						strcpy(formula,&formula[1]);
			            //cout<<token<<" : "<<formula;
						cout<<formula;
					}
					else
						//cout<<token<<" : "<<Vars->get_value(token);
						cout<<Vars->get_value(token);
				}
				else
					cout<<endl<<"DynaFLIP++: Warning: variable "<<token<<" does not exist.";
			}
			strcpy(token,"");
			continue;
		}


// ----------------------------------------------------------------------------------------------------------
// INPUT expression
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"INPUT")==0)	// INPUT value statement reached
		{
			infile >> varname;		// fetch name of variable that has to get a new value
			number_of_tokens++;
			cin>>formula;			// get new value from cin
			if (formula[0]=='"') 
				f_value = 0;
			else
			{
				int is_ok=1;
				l=strlen(formula);
				for (i=0;i<l;i++)
					if (isalpha(formula[i])) is_ok*=0;
				if (is_ok) 
					f_value = atof(formula);
				else
				{
					f_value = 0;
					strcpy(formula,"0");
				}
            }
			Vars->real_name(formula);
			if (!Vars->in_list(varname))	// is varname not an existing variable?
				Vars->insert(varname,formula, f_value);  // insert new variable
			else
			{
				Vars->set_formula(varname,formula);	   // set new formula
				Vars->set_value(varname,f_value);	   // set new value
				Vars->set_comp(varname,TRUE);		   // computed flag = TRUE
			}
			strcpy(token,"");
			continue;
		}

// ----------------------------------------------------------------------------------------------------------
// IF statement
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"IF")==0)				// IF statement reached
		{
			if_cascade++;        				// increment number of cascaded IF - structures
			infile>>token;				// Get if expression
			number_of_tokens++;
			sprintf(ifname,"IF:%d",if_cascade);
			Controls->insert(ifname,token,0);		// insert actual IF:if_cascade Control 
			Vars->real_name(token);
			Tree = new TExptree(NODE,"","","",0.0,NULL,NULL);
			Tree->insert_expr(token,0);
			if_cond=Tree->calculate(Vars,SMART);		// evaluate IF - condition
			Tree = Tree->clean_up(Tree);
			Controls->set_value(ifname,if_cond);		// insert result to IF:if_cascade
			if (if_cond)		// which branch should be executed?
				branch = B_THEN;
			else
				branch = B_ELSE;
			strcpy(token,"");
			continue;
		}

// ----------------------------------------------------------------------------------------------------------
// THEN statement
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"THEN")==0)  	    // THEN statement reached
		if (branch==B_ELSE)
		{
			number_of_ifs = 1;          // skip all inner IF-THEN-ELSE-ENIDF structures
			while (number_of_ifs!=0)
			{
				infile>>token;
				number_of_tokens++;
				if (strcmp("IF",token)==0) number_of_ifs++;
				if ((strcmp("ELSE",token)==0)&&(number_of_ifs==1)) number_of_ifs--;
				if (strcmp("ENDIF",token)==0) number_of_ifs--;
			}
		}
		else
		{
			strcpy(token,"");
			continue;
		}


// ----------------------------------------------------------------------------------------------------------
// ELSE statement
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"ELSE")==0)	// ELSE statement reached
		if (branch ==B_THEN)
		{
			number_of_ifs = 1;
			while (number_of_ifs!=0)   // skip all inner IF-THEN-ELSE-ENIDF structures
			{
				infile>>token;
				number_of_tokens++;
				if (strcmp("IF",token)==0) number_of_ifs++;
				if ((strcmp("ELSE",token)==0)&&(number_of_ifs==1)) number_of_ifs--;
				if (strcmp("ENDIF",token)==0) number_of_ifs--;
			}
		}
		else
		{
			strcpy(token,"");
			continue;
		}


// ----------------------------------------------------------------------------------------------------------
// ENDIF statement
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"ENDIF")==0)  // ENDIF statement reached
		{
			sprintf(ifname,"IF:%d",if_cascade);
			Controls->remove(ifname); // remove actual IF:if_cascade constructs
			if_cascade--;		   // decrement number of cascaded if - structures
			if (if_cascade>0)	   // get branch value (THEN-ELSE) of outer IF
			{
				sprintf(ifname,"IF:%d",if_cascade);
				branch = (int)Controls->get_value(ifname);
			}
			else
				branch = 0;	  // no more branches left to execute
			strcpy(token,"");
			continue;
		}


// ----------------------------------------------------------------------------------------------------------
// FOR statement
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"FOR")==0)	  // FOR statement reached
		{
			for_cascade++;		  // increment number of cascaded FOR - statements encounterd
			sprintf(forname,"FOR:%d",for_cascade);
			sprintf(formula,"%d",for_cascade);
	
			Controls->insert(forname,formula,0);// put new FOR:for_cascade control variable to Controls
			infile>>varname;  			// read name of run variable
			number_of_tokens++;
			infile>> token;   			// skip =
			number_of_tokens++;
			if (strcmp(token,"=")!=0)
			{
				cout<<endl<<"Error: '=' missing in FOR "<<varname<<" = "<<token<<" at token #"<<number_of_tokens<<endl;
				break;
			}
			infile>>token;  			// get start value/variable
			Vars->real_name(token);
			number_of_tokens++;
	
			if(Vars->in_list(token))		// is start a value or a variable?
				f_value = Vars->get_value(token);// get value of variable
			else
				f_value = atof(token);		// get immendiate value
			sprintf(formula,"(%.3f)",f_value);
	
			sprintf(token,"%s+",varname);	// construct increment/decrement control like i=i+1 etc.
			strcat(token,formula);
			strcpy(formula,token);
			strcpy(token,"");
			if (!Vars->in_list(varname)) 	// does variable already exist?
				Vars->insert(varname,formula,f_value);	// insert as new variable
			else
			{
				Vars->set_value(varname,f_value);	// set new start value for variable
				Vars->set_formula(varname,formula);	// set increment/decrement formula
			}
			Vars->set_comp(varname,TRUE);		// computed flag = TRUE

		    // insert default values when TO .. STEP .. not given

		    sprintf(forname,"STEP:%d",for_cascade);	// put new STEP:for_cascade control variable to Controls
			sprintf(formula,"%d",for_cascade);
			Controls->insert(forname,formula,1);	// default STEP value is 1
			sprintf(forname,"TO:%d",for_cascade);	// put new TO:for_cascade control variable to Controls
			sprintf(formula,"%d",INT_MAX);
			Controls->insert(forname,formula,INT_MAX);	// default TO value is INT_MAX
			sprintf(forname,"VAR:%d",for_cascade);	// put new VAR:for_cascade control variable to Controls
			strcpy(formula,varname);			// remember loop variable on this level
			Controls->insert(forname,formula,f_value);	// insert VAR:for_cascade -> remember varname
			strcpy(token,"");
			continue;
		}

// ----------------------------------------------------------------------------------------------------------
// TO statement
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"TO")==0)			// TO statement reached
		{
			sprintf(forname,"TO:%d",for_cascade);	// update TO:for_cascade control variable to Controls
			infile>>formula;				// fetch TO value/variable
			Vars->real_name(formula);
			number_of_tokens++;
			if(Vars->in_list(formula))			// value or variable
				f_value = Vars->get_value(formula);	// if variable get its value
			else
				f_value = atof(formula);			// if value get the immediate value
			Controls->set_value(forname,f_value);	// update (default-) value
			strcpy(token,"");
			continue;
		}

// ----------------------------------------------------------------------------------------------------------
// STEP statement
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"STEP")==0)		// STEP statement reached
		{
			sprintf(forname,"STEP:%d",for_cascade);	// update TO:for_cascade control variable to Controls
			infile>>formula;
			Vars->real_name(formula);			// fetch STEP value/variable
			number_of_tokens++;
			if(Vars->in_list(formula))			// value or variable
				f_value = Vars->get_value(formula);	// if variable get its value
			else
				f_value = atof(formula);			// if value get the immediate value
			sprintf(token,"VAR:%d",for_cascade);	// get actual loop variable
			strcpy(varname,Controls->get_formula(token));
			sprintf(formula,"%s+",varname);
			sprintf(token,"(%.3f)",f_value);		// update FOR - variables increment/decrement formula
			strcat(formula,token);
			strcpy(token,"");
			Vars->set_formula(varname,formula);		// update FOR - variable with new formula
			Controls->set_value(forname,f_value);	// update STEP:value
		
			sprintf(forname,"FORPOS:%d",for_cascade);	// ifstreamposition of this FOR statement
			long l = (long)infile.tellg();
			f_value = (float)l;
			Controls->insert(forname,formula,f_value);	// insert FOR:for_cascade
			// cout<<"DEBUG :"<<token<<endl;
			strcpy(token,"");
			continue;
		}

// ----------------------------------------------------------------------------------------------------------
// NEXT statement
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"NEXT")==0)		// NEXT statement reached
		{
			infile>>varname;				// fetch name of variable
			Vars->real_name(varname);
			number_of_tokens++;
			f_value=Vars->get_value(varname);		// get actual value of variable
			sprintf(forname,"STEP:%d",for_cascade);	// do we increment or decrement the variable?
			if ((step_value=Controls->get_value(forname))>0)
				i = 1;					// we increment the variable of size STEP ..
			else
				i = -1;					// we decrement the variable of size STEP ..
			sprintf(forname,"TO:%d",for_cascade);
			for_cond = Controls->get_value(forname);	// get the TO:for_cascade value as terminate condition
			switch(i)            			// do we heve an incremental or decremental loop?
			{
			case 1: 
				if ((step_value+f_value )<=for_cond)
				{
					f_value+=step_value;		   	// calculate new value for variable (var +/- step)
					Vars->set_value(varname,f_value);	// set variabel on new value
				}
				else
					f_value+=step_value;
				break;
			case -1: 
				if ((f_value+step_value)>=for_cond)
				{
					f_value+=step_value;		   	// calculate new value for variable (var +/- step)
					Vars->set_value(varname,f_value);	// set variabel on new value
				}
				else
					f_value+=step_value;
				break;
			default:
				;
			}
			if (((f_value>for_cond)&&(i>0))||((f_value<for_cond)&&(i<0)))  // End of FOR condition reached?
			{
				sprintf(forname,"FOR:%d",for_cascade);
				Controls->remove(forname);		//  remove FOR:for_cascade from Controls
				sprintf(forname,"TO:%d",for_cascade);
				Controls->remove(forname);		//  remove TO:for_cascade from Controls
				sprintf(forname,"STEP:%d",for_cascade);
				Controls->remove(forname);		//  remove STEP:for_cascade from Controls
				sprintf(forname,"VAR:%d",for_cascade);
				Controls->remove(forname);		//  remove VAR:for_cascade from Controls
				sprintf(forname,"FORPOS:%d",for_cascade);
				Controls->remove(forname);		//  remove FORPOS:for_cascade from Controls
				for_cascade--;
			}
			else
			{                                        // return to begin of actual FOR statement
				sprintf(forname,"FORPOS:%d",for_cascade);
				f_value = Controls->get_value(forname);
				infile.seekg(long(f_value));		
			}
			strcpy(token,"");
			continue;
		}


// ----------------------------------------------------------------------------------------------------------
// LOOP statement
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"LOOP")==0)			// LOOP statement reached
		{
			loop_cascade++;   				// increment number of cascaded LOOP - structures
			sprintf(loopname,"LOOP:%d",loop_cascade);
			sprintf(formula,"%d",loop_cascade);
			Controls->insert(loopname,formula,0);
			sprintf(loopname,"LOOPPOS:%d",loop_cascade);	// ifstreamposition of this LOOP statement
			long l = (long)infile.tellg();
			f_value = (float)l;
			Controls->insert(loopname,formula,f_value);	// insert LOOP:for_loop
			strcpy(token,"");
			continue;
		}


// ----------------------------------------------------------------------------------------------------------
// ENDLOOP statement
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"ENDLOOP")==0)			// ENDLOOP statement reached
		{
			sprintf(loopname,"LOOPPOS:%d",loop_cascade);
			f_value = Controls->get_value(loopname);
			infile.seekg(long(f_value));
			strcpy(token,"");
			continue;				// reset contens of token
		}


// ----------------------------------------------------------------------------------------------------------
// EXITLOOP statement
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"EXITLOOP")==0)		// EXITLOOP statement reached
		{
			infile>>token;			// Get exit  expression
			Vars->real_name(token);
			number_of_tokens++;
			sprintf(loopname,"EXITLOOP:%d",loop_cascade);
			sprintf(formula,token);
			if(!Controls->in_list(loopname))
				Controls->insert(loopname,formula,0);
			else
				strcpy(formula,Controls->get_formula(loopname));

			Tree = new TExptree(NODE,"","","",0.0,NULL,NULL);
			Tree->insert_expr(formula,0);
			if_cond=Tree->calculate(Vars,SMART);
			Tree = Tree->clean_up(Tree);
	
			 Controls->remove(loopname);

			if (if_cond)		// exit condition ?
			{
				number_of_skipped_loops = 1;          // skip all inner LOOP structures
				while (number_of_skipped_loops!=0)
				{
					infile>>token;
					number_of_tokens++;
					if (strcmp("LOOP",token)==0) number_of_skipped_loops++;
					if (strcmp("ENDLOOP",token)==0) number_of_skipped_loops--;
				}
				sprintf(loopname,"LOOP:%d",loop_cascade);
				Controls->remove(loopname);
				sprintf(loopname,"LOOPPOS:%d",loop_cascade);
				Controls->remove(loopname);
				loop_cascade--;
			}
			strcpy(token,"");
			continue;
		}


// ----------------------------------------------------------------------------------------------------------
// GOTO statement
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"GOTO")==0)		// GOTO statement reached
		{
			infile>>token;				// get label name
			number_of_tokens++;
			strcpy(labelname,token);
			if (Controls->in_list(labelname))
			{
				infile.clear();			// reset flags
				infile.seekg(0);		// return to begin of script
				number_of_tokens=0;
				strcpy(token,"");
				label_found = 0;
				while(!label_found)
				{
					while ((strcmp(token,"LABEL")!=0)&&(infile.eof()==0)) 
					{ 
						infile>>token;
						number_of_tokens++;
					}
					if(infile.eof()!=0)
						label_found = 1; 	// label not found, exit
					else	
					{
						infile>>token;number_of_tokens++;
					}
					if (strcmp(token,labelname)==0) label_found = 1; // label found, continue
				}
			}
			strcpy(token,"");
			continue;
		}

// ----------------------------------------------------------------------------------------------------------
// EXIT statement
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"EXIT")==0)		// EXIT statement reached, terminate
		{
			break;
		}

// ----------------------------------------------------------------------------------------------------------
// CALL statement
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"CALL")==0)		// CALL statement reached, terminate
		{
			infile>>token;				// get label name
			number_of_tokens++;
			i = number_of_tokens;			// remember number_of_tokens before procedure call
			proc_cascade++;				// increment number of procedure calls
			strcpy(labelname,token);
			Vars->real_name(labelname);
			if (Controls->in_list(labelname))
			{
				infile.clear();			// reset flags
				infile.seekg(0);		// return to begin of file
				number_of_tokens=0;
				strcpy(token,"");
				label_found = 0;
				while(!label_found)
				{
					while ((strcmp(token,"PROCEDURE")!=0)&&(infile.eof()==0)) 
					{ 
						infile>>token;
						number_of_tokens++;
					}
					if(infile.eof()!=0)
						label_found = 1; 	// label not found, exit
					else
						{infile>>token;number_of_tokens++;}
					if (strcmp(token,labelname)==0) label_found = 1; // label found, continue
				}
				sprintf(labelname,"CALL:BRANCH:%d",proc_cascade);
				Controls->insert(labelname,"0",(float)branch);
				sprintf(labelname,"CALL:IF_CASCADE:%d",proc_cascade);
				Controls->insert(labelname,"0",(float)if_cascade);
				sprintf(labelname,"CALL:FOR_CASCADE:%d",proc_cascade);
				Controls->insert(labelname,"0",(float)for_cascade);
				sprintf(labelname,"CALL:LOOP_CASCADE:%d",proc_cascade);
				Controls->insert(labelname,"0",(float)loop_cascade);
				sprintf(labelname,"CALL:TOKENS:%d",proc_cascade);
				Controls->insert(labelname,"0",(float)i);
			}
			else
			{
				cout<<endl<<"Error: no PROCEDURE "<<labelname<<" defined!"<<endl;
			}  
			strcpy(token,"");
			continue;
		}

// ----------------------------------------------------------------------------------------------------------
// RETURN statement
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"RETURN")==0)
		{
			sprintf(labelname,"CALL:BRANCH:%d",proc_cascade);
			branch = (int)Controls->get_value(labelname);
			Controls->remove(labelname);
			sprintf(labelname,"CALL:IF_CASCADE:%d",proc_cascade);
			if_cascade = (int)Controls->get_value(labelname);
			Controls->remove(labelname);
			sprintf(labelname,"CALL:FOR_CASCADE:%d",proc_cascade);
			for_cascade = (int)Controls->get_value(labelname);
			Controls->remove(labelname);
			sprintf(labelname,"CALL:LOOP_CASCADE:%d",proc_cascade);
			loop_cascade = (int)Controls->get_value(labelname);
			Controls->remove(labelname);
			sprintf(labelname,"CALL:TOKENS:%d",proc_cascade);
			number_of_tokens = (int)Controls->get_value(labelname);
			Controls->remove(labelname);
			proc_cascade--;
			infile.clear();					// reset flags
			infile.seekg(0);				// return to begin of script
			for(i=0;i<number_of_tokens;i++) infile>>token;  // return to point of procedure call
			strcpy(token,"");				// reset contens of token
			continue;
		}

// ----------------------------------------------------------------------------------------------------------
// DELETE statement
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"DELETE")==0)    		// removes Variable(s) from Vars
		{
			infile>>varname;
			number_of_tokens++;
			Vars->real_name(varname);
			if (strcmp(varname,"RULESET")==0) 
			if (Rules!=NULL)
			{
				delete Rules;
				Rules = new RuleSet();
				strcpy(token,"");
				continue;
			}
			if (strcmp(varname,"EVALTREE")==0) 
			if (Evals!=NULL)
			{
				Vars->remove(Evals->get_name());
				// RAGGL delete Evals;
				Evals->clean_up();
				strcpy(token,"");
				continue;
			}	
			if (strcmp(varname,"ALL")==0)
			{
				Vars->clean_up();             	// remove all Variables
				Vars = new TVarlist();			// create new Memory Object	
			}
			else
			if (Vars->in_list(varname))
			{
				Vars->remove(varname);	       // remove Variable named varname (token)
				if ((cv=Rules->get_cv_list()->find_cv(varname))!=NULL)
					Rules->remove_cv_and_its_approaches(cv); // delete fuzzy variable and relative rules ...
			}
			else
				cout<<endl<<"Warning: "<<varname<<" is not a valid variable"<<endl;
			strcpy(token,"");
			continue;
		}

// ----------------------------------------------------------------------------------------------------------
// MEMDUMP statement
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"MEMDUMP")==0)			// show Variables in memory
		{
			infile>>varname;
			number_of_tokens++;
			if (strcmp(varname,"ALL")==0)
				Vars->show();
			else
			{
				Vars->real_name(varname);
				if ((cv=Rules->get_cv_list()->find_cv(varname))!=NULL) 
					cv->show();
				else
				if (Vars->in_list(varname))
					Vars->show(varname);
				else
					cout<<endl<<"Warning: "<<varname<<" is not a valid variable"<<endl;
			}
			strcpy(token,"");
			continue; 
		}

// ----------------------------------------------------------------------------------------------------------
// IMPORT statement
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"IMPORT")==0)	// IMPORT value statement reached
		{
			infile >> token;		// fetch name of what has to be imported
			number_of_tokens++;
			infile>>varname;		// get name of output file
			number_of_tokens++;		  
			if (strcmp(token,"RULESET")==0) 	// read RuleSet from file
			{
				if (Rules!=NULL)
				{
					delete Rules;
					Rules = new RuleSet();
				}
				ifstream *rulefile = new ifstream;
				rulefile->open(varname);
				if (rulefile->fail())
				{
					cout<<endl<<"Error: Unknown RULESET-file "<<varname<<"!"<<endl;
					break;
				}
				Rules->read(rulefile);
				rulefile->close();
				delete rulefile;			// possible memory leak
			}
			if (strcmp(token,"PARAMETERSET")==0) 	// read ParameterSet from file
			{
				ifstream *rulefile = new ifstream;
				rulefile->open(varname);
				if (rulefile->fail())
				{
					cout<<endl<<"Error: Unknown PARAMETERSET-file "<<varname<<"!"<<endl;
					break;
				}
				ps = new ParameterSet();
				ps->read(rulefile);
				Rules->get_ps_list()->add(ps);		// add ParameterSet to RuleSet
				rulefile->close();
				delete rulefile;			// possible memory leak
			}
			if (strcmp(token,"PSLIST")==0) 		// read PSList from file
			{
				ifstream *rulefile = new ifstream;
				rulefile->open(varname);
				if (rulefile->fail())
				{
					cout<<endl<<"Error: Unknown PSLIST-file "<<varname<<"!"<<endl;
					break;
				}
				Rules->get_ps_list()->read(rulefile);	// add PSList to RuleSet
				rulefile->close();
				delete rulefile;			// possible memory leak
			}
			if (strcmp(token,"VARIABLES")==0) 
			{
				Vars->read(varname); 	// read data from file and insert in Vars
				Vars->substitute_all();	// substitute abstract variables with real ones
			}
		
			strcpy(token,"");
			continue;
		}

// ----------------------------------------------------------------------------------------------------------
// EXPORT statement
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"EXPORT")==0)			// EXPORT value statement reached
		{
			infile >> token;					// fetch name of what has to be exported
			number_of_tokens++;
			infile >> varname;					// get name of output file
			number_of_tokens++;
			if (strcmp(token,"RULESET")==0) 			// write RuleSet to outfile
			{
				ofstream *rulefile = new ofstream;
				rulefile->open(varname);
				Rules->write(rulefile);
				rulefile->close();
				delete rulefile;				// possible memory leak
			}
			if (strcmp(token,"EVALTREE")==0) 			// write TEvalTree to outfile
			{
				Evals->write(varname);
			}
			if (strcmp(token,"PSLIST")==0) 		// write PSList to outfile
	        {
				ofstream *rulefile = new ofstream;
				rulefile->open(varname);
				Rules->get_ps_list()->write(rulefile);
				rulefile->close();
				delete rulefile;				// possible memory leak
			}
			if (strcmp(token,"PARAMETERSET")==0)		// write ParameterSet to outfile
			{
				infile>>token;
				number_of_tokens++;
	 	
				ofstream *rulefile = new ofstream;
				rulefile->open(token);
				int n = Rules->get_ps_list()->get_number();	// get number of ParameterSets in Rules
   				for(i=0;i< n;i++)					// fetch ParameterSet from Rules
     	        if (strcmp(varname,(Rules->get_ps_list()->find(i))->get_ps_name())==0)
      	            Rules->get_ps_list()->find(i)->write(rulefile);
				rulefile->close();
				delete rulefile;				// possible memory leak
            }
			if (strcmp(token,"VARIABLES")==0) Vars->write(varname);	// write Vars to file
			strcpy(token,"");
			continue;
		}

// ----------------------------------------------------------------------------------------------------------
// PARAMETERSET
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"PARAMETERSET")==0)// PARAMETERSET statement reached
		{
			infile >> token; number_of_tokens++;	// skip NAME
			infile >> ps_name;			// get name of ParameterSet		 
			number_of_tokens++;
			infile >> token; number_of_tokens++;	// skip LOWER
			infile >> lower;			// get lower end of ParameterSet
			number_of_tokens++;
			infile >> token; number_of_tokens++;	// skip UPPER
			infile >> upper;			// get upper end of ParameterSet
			number_of_tokens++;
			infile >> token; number_of_tokens++;	// skip MEDIAN
			infile>> median;			// get median of ParameterSet
			number_of_tokens++;
			infile >> token; number_of_tokens++;	// skip DISTRIBUTION
			infile >> distribution;		// get distribution [0,1] of ParameterSet
			number_of_tokens++;
			infile >> token; number_of_tokens++;	// skip DILATATION
			infile >> dilatation;		// get dilatation [0,1] of ParameterSet
			number_of_tokens++;
			infile >> token; number_of_tokens++;	// skip IMPORTANCE
			infile >> importance;		// get importance [0,1] of ParameterSet
			number_of_tokens++;
			infile >> token; number_of_tokens++;	// skip TERMS
			infile >> token;			// get first linguistic term or END statement
			number_of_tokens++;
			nl = new NList();			// create list for linguistic terms
			i = 0;				// i counts the number of linguistic terms
			while(strcmp(token,"END")!=0)	// no further linguistic terms
			{
				nl->add(token);			// add term to list of linguistic terms
				i++;				// increment number of terms
				infile>>token;
				number_of_tokens++;
			}
			if (i==0) 
			{
				cout<<endl<<"Error: no linguistic terms defined for ParameterSet "<<ps_name<<"!"<<endl;
				break;
			}
			else
			{
				cl = NULL;
				cl = new CList(i);		// create color list for linguistic terms 
				ps = new ParameterSet(ps_name,lower,upper,median,distribution,dilatation,importance,cl,nl);
				Rules->get_ps_list()->add(ps);
				delete ps;				// possible memory leak
				delete cl;				// possible memory leak
				delete nl;				// possible memory leak
			}
			strcpy(token,"");
			continue;
		}


// ----------------------------------------------------------------------------------------------------------
// FUZZYVAR
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"FUZZYVAR")==0)	// FUZZYVAR statement reached
		{
			infile >> cv_name;			// get name of CombinedVariable
			number_of_tokens++;
			Vars->real_name(cv_name);		// indexed name as e.g. FuzzyVar[i], i==10, -> FuzzyVar[10]
			infile >> ps_name;			// get name of ParameterSet for this CombinedVariable
			number_of_tokens++;
			cv = NULL;
			int n = Rules->get_ps_list()->get_number();	// get number of ParameterSets in Rules
			for(i=0;i< n;i++)			// fetch ParameterSet from Rules
			{
				if (strcmp(ps_name,(Rules->get_ps_list()->find(i))->get_ps_name())==0)
				{
					cv = new CombinedVariable(Rules->get_ps_list()->find(i),cv_name);
					Rules->get_cv_list()->add(cv);
					delete cv;					// possible memory leak
             	    if (!Vars->in_list(cv_name))   // insert Variable named like CombinedVariable in Varlist
     					Vars->insert(cv_name,ps_name,0);	// NB: store ps_name in formula when insert in Varlist
 					else
     				{
      					Vars->set_formula(cv_name,ps_name);	// NB: store ps_name in formula when insert in Varlist
      					Vars->set_comp(cv_name,TRUE);
						Vars->set_value(cv_name,0);
     				}
					break;
				}
			}
			if (cv==NULL)	// appropriate ParameterSet not defined in RuleSet
			{
				cout<<endl<<"Error: unknown PARAMETERSET "<<ps_name<<" for "<<cv_name<<"!"<<endl;
				break;
			}
			strcpy(token,"");
			continue;
		}

// ----------------------------------------------------------------------------------------------------------
// SET
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"SET")==0)	// SET value statement reached
		{
			char ps_name[MAX_NAME_LEN];	// name of ParameterSet
			infile >> cv_name;		// get name of CombinedVariable
			number_of_tokens++;
			infile >> token;    		// skip =
			number_of_tokens++;
		
			if ((strcmp(token,"=")!=0)&&(strcmp(token,":")==0))
			{
				cout<<endl<<"Error: '= or :' missing in SET "<<varname<<" =|: ";
				cout<<token<<" at token #"<<number_of_tokens<<endl;
				break;
			}
 
			infile>>formula;  		// read formula
			number_of_tokens++;
	
			Vars->real_name(cv_name);	// indexed name as e.g. Var[i], i==10, -> Var[10]
			if (strcmp(token,"=")==0)	 
				Vars->real_name(formula);	// indexed names in formula as e.g. Var[i]+12, i==10, -> Var[10]+12
			else
				strcpy(formula,Vars->substitute(Vars->get_formula(cv_name))); // SET var : SUM{...} -> substitute open indices
 	
			if (!Vars->in_list(cv_name))   // insert Variable named like CombinedVariable in Varlist
			{
				cout<<endl<<"Error: unknown FUZZYVAR "<<cv_name<<"!"<<endl;
				break;
			}
			else
			{
				strcpy(ps_name,Vars->get_formula(cv_name));	// store name of ParameterSet for this CombinedVariable
				Vars->set_formula(cv_name,formula);		// insert new expression in formula field
				Vars->set_comp(cv_name,TRUE);
			}
	 
			if (Rules->get_cv_list()->find_cv(formula)!=NULL)	// is right side of "SET FUZZYVAR cv_name = formula" a FUZZYVAR
			{
				// Santa 20.07.1996 Rules->remove_cv_and_its_approaches(Rules->get_cv_list()->find_cv(cv_name));
				Vars->remove(cv_name);				// remove cv from RuleSet and Vars
				cv = new CombinedVariable();		// creat new CombinedVariable
				cv->copy(Rules->get_cv_list()->find_cv(formula)); // copy CombinedVariable of right side
				cv->set_cv_name(cv_name);			// set cv_name
				Rules->get_cv_list()->add(cv);		// add new CombinedVariable to RuleSet
				delete cv;							// possible memory leak
				Vars->insert(cv_name,Vars->get_formula(formula),cv->get_crisp_value()); // insert in Vars
				Vars->set_comp(cv_name,TRUE);
				strcpy(token,"");
				continue;					// leave the SET statement part
			}



			// right side is a regular expression
			Tree = new TExptree(NODE,"","","",0.0,NULL,NULL); 	// build up evaluation tree
			Tree->insert_expr(formula,0);  			// insert expression
			f_value=Tree->calculate(Vars,SMART);			// evaluate expression
			Tree=Tree->clean_up(Tree);				// remove evaluation tree
	 
			Vars->set_value(cv_name,f_value);			// set varname to calculated value of formula
			Vars->set_comp(cv_name,TRUE);				// computed flag = TRUE
			Vars->set_formula(cv_name,ps_name);			// set formula on name of ParameterSet
	 
			cv = Rules->get_cv_list()->find_cv(cv_name);		// fetch CombinedVariable from Rules
			if (cv!=NULL)
				cv->set_value(f_value);				// set value of Combined Variable
			else
			{
				cout<<endl<<"Error: unknown FUZZYVAR "<<cv_name<<"!"<<endl;
				break;
			}
			strcpy(token,"");
			continue;
		}

// ----------------------------------------------------------------------------------------------------------
// FUZZYCONCAT
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"FUZZYCONCAT")==0)	// FUZZYCONCAT statement reached
		{
			concat_operator conc_op;		// Type of concat operator
			char constraint[MAX_NAME_LEN];		// name of CombinedVariable for FUZZYCONCAT
			char name1[MAX_NAME_LEN];		// name of first CombinedVariable
			char name2[MAX_NAME_LEN];		// name of second CombinedVariable
			char ps_name1[MAX_NAME_LEN];		// name of ParameterSet for first CombinedVariable
			char ps_name2[MAX_NAME_LEN];		// name of ParameterSet for second CombinedVariable
			char ps_name3[MAX_NAME_LEN];		// name of ParameterSet for CombinedVariable for FUZZYCONCAT
			char c_operator[MAX_NAME_LEN];		// name of concat operator (AND, OR)
			char in1[MAX_NAME_LEN];		// name of first input linguistic term
			char in2[MAX_NAME_LEN];		// name of second input linguistic term
			char out[MAX_NAME_LEN];		// name of output linguistic term
			CombinedVariable *cv_1;		// first input CombinedVariable 
			CombinedVariable *cv_2;		// second input CombinedVariable	
			CombinedVariable *cv_out;		// output CombinedVariable
			ifstream tablefile;			// stream for table
			unsigned int found;			// flag for success of search in table

			if (strcmp(tablename,"empty")==0)
			{
				cout<<endl<<"Error: no TABLE defined for generation of rules!"<<endl;
				break;
			}
 
			infile >> constraint;		// get name of constraint
			number_of_tokens++;
			infile >>token;		// skip IS
			number_of_tokens++;
			infile >> name1;		// get name of first constraint
			number_of_tokens++;
			infile >> c_operator;		// get operator
			number_of_tokens++;
			infile >> name2;		// get name of second constraint/expression
			number_of_tokens++;
			if ((strcmp(c_operator,"AND")!=0)&&(strcmp(c_operator,"OR")!=0))
			{
				cout<<endl<<"Error: fuzzy operator [AND|OR] expected in FUZZYCONCAT statement!"<<endl;
				break;
			}
			else
			{
				Vars->real_name(name1);		// build real name of first CombinedVariable
				Vars->real_name(name2);		// build real name of second CombinedVariable
				Vars->real_name(constraint);	// build real name of output CombinedVariable
				if (!Vars->in_list(name1))		// does CombinedVariable exist?
				{
					cout<<endl<<"Error: unknown FUZZY VARIABLE "<<name1<<"!"<<endl;
					break;
				}
				if (!Vars->in_list(name2))		// does CombinedVariable exist?
				{
					cout<<endl<<"Error: unknown FUZZY VARIABLE "<<name2<<"!"<<endl;
					break;
				}
				if (!Vars->in_list(constraint))	// does CombinedVariable exist?
				{
					cout<<endl<<"Error: unknown FUZZY VARIABLE "<<constraint<<"!"<<endl;
					break;
				}

				strcpy(ps_name1,Vars->get_formula(name1));		// store name of Parameterset for first CombinedVariable
				strcpy(ps_name2,Vars->get_formula(name2));		// store name of Parameterset for second CombinedVariable
				strcpy(ps_name3,Vars->get_formula(constraint)); 	// store name of Parameterset for output CombinedVariable

				// delete rules for cv 
				// if ((cv_out=Rules->get_cv_list()->find_cv(constraint))!=NULL)
				//    Rules->remove(cv_out); 

				found = FALSE;					// initialize flag for search in table
				tablefile.open(tablename,ios::in);			// open table
				while((!found)&&(tablefile.eof()==0))
				{
					tablefile >> token;
					if (strcmp(token,"CONCAT")==0)		// search for CONCAT block
					{
 						tablefile>>token;
						if (strcmp(token,ps_name1)==0)	// where in1 is of right type for first CombinedVariable
						{
							tablefile>>token;
							if (strcmp(token,ps_name2)==0) 	// where in2 is of right type for second CombinedVariable
							{
								tablefile>>token;
								if (strcmp(token,ps_name3)==0)  	// where out is of right type for output CombinedVariable
								{   
									found = TRUE; 			// right position in table
		 							cv_1 = Rules->get_cv_list()->find_cv(name1);	// get first CombinedVariable
			 						cv_2 = Rules->get_cv_list()->find_cv(name2);	// get second CombinedVariable
			 						cv_out = Rules->get_cv_list()->find_cv(constraint);	// get output Combined Variable
								}
							}
						}
					}
				}
				if (!found)	// no entry for this type of concatenation
				{
					cout<<endl<<"Error: no entry in TABLE for FUZZYCONCAT "<<constraint<<" "<<name1<<" "<<c_operator<<" "<<name2<<"!"<<endl;
					break;
				}
				if (cv_1==NULL)
				{
					cout<<endl<<"Error: FUZZY VARIABLE "<<name1<<" not in RuleSet for FUZZYCONCAT "<<constraint<<"!"<<endl;
					break;
				}
				if (cv_2==NULL)
				{
					cout<<endl<<"Error: FUZZY VARIABLE "<<name2<<" not in RuleSet for FUZZYCONCAT "<<constraint<<"!"<<endl;
					break;
				}
				if (cv_out==NULL)
				{
					cout<<endl<<"Error: FUZZY VARIABLE "<<constraint<<" no longer in RuleSet for FUZZYCONCAT "<<constraint<<"!"<<endl;
					break;
				}
		
				// Add all rules to ruleset
				tablefile>>token;		// skip AND/OR - TAG
				while((tablefile.eof()==0)&&(strcmp(token,c_operator)!=0)) tablefile>>token; // RAGGL 
				if (strcmp(c_operator,"OR")==0) 				// set right operator for concatenation
					conc_op = OR;
				else
					conc_op = AND;
				while(tablefile.eof()==0)
				{
					tablefile>>in1;							// get first input linguistic term
					if ((strcmp(in1,"CONCAT")==0)||(strcmp(in1,"COMPARE")==0)||	// end of block in table reached?
						(strcmp(in1,"AND")==0)||(strcmp(in1,"OR")==0)||
						(strcmp(in1,"=")==0)||(strcmp(in1,"#")==0)||
						(strcmp(in1,"<")==0)||(strcmp(in1,">")==0)||(strcmp(in1,"")==0))
					{
						tablefile.close();						// close table
						break;
					}
					else
					{
						tablefile>>in2;							// get second input linguistic term
						tablefile>>out;							// get output linguistic term
						if((strcmp(in1,"")!=0)&&(strcmp(in2,"")!=0)&&(strcmp(out,"")!=0))
							Rules->add_ComplexRule(cv_1, in1, cv_2, in2, cv_out, out,  conc_op, 1); // add rules to RuleSet
					}
				}
				tablefile.close();								// close table
			} // end of ?(tablename not empty)	  
			strcpy(token,"");
			continue;  
		}



// ----------------------------------------------------------------------------------------------------------
// FUZZYCOMPARE
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"FUZZYCOMPARE")==0)	// FUZZYCOMPARE statement reached
		{
			char constraint[MAX_NAME_LEN];
			char name[MAX_NAME_LEN];
			char formula[MAX_NAME_LEN];
			char ps_name1[MAX_NAME_LEN];
			char ps_name2[MAX_NAME_LEN];
			char c_operator[MAX_NAME_LEN];
			char in[MAX_NAME_LEN];
			char out[MAX_NAME_LEN];
			CombinedVariable *cv;
			CombinedVariable *cv_new;
			char new_cv_name[MAX_NAME_LEN];
			CombinedVariable *cv_out;
			ifstream tablefile;
			unsigned int found;
			float f_comp_value;
			float f_cv_value; 
			int rules_there;

			if (strcmp(tablename,"empty")==0)
			{
				cout<<endl<<"Error: no TABLE defined for generation of rules!"<<endl;
				break;
			}
 
			infile >> constraint;		// get name of constraint
			number_of_tokens++;
			infile >>token;		// skip IS
			number_of_tokens++;
			infile >> name;		// get name of variable
			number_of_tokens++;
			infile >> c_operator;		// get compare operator
			number_of_tokens++;
			infile >> formula;		// get name of expression
			number_of_tokens++;
			if ((strcmp(c_operator,"=")!=0)&&(strcmp(c_operator,"#")!=0)&&
				(strcmp(c_operator,"<")!=0)&&(strcmp(c_operator,">")!=0))
			{
				cout<<endl<<"Error: compare operator [=|#|<|>] expected in FUZZYCOMPARE statement!"<<endl;
				break;
			}
			else
			{
				Vars->real_name(constraint);			// build real name of CombinedVariable for FUZZYCOMPARE
				// delete rules for cv 
				// if ((cv_out=Rules->get_cv_list()->find_cv(constraint))!=NULL)
				//   Rules->remove(cv_out);
		
				Vars->real_name(name);				// build real name of CombinedVariable
				Vars->real_name(formula);				// build real expression of compare value
				if (!Vars->in_list(name))
				{
					cout<<endl<<"Error: unknown (FUZZY) VARIABLE "<<name<<"!"<<endl;
					break;
				}
	     
				if (!Vars->in_list(constraint))
				{
					cout<<endl<<"Error: unknown FUZZY VARIABLE "<<constraint<<"!"<<endl;
					break;
				}
		
				Tree = new TExptree(NODE,"","","",0.0,NULL,NULL); 	// build up evaluation tree
				Tree->insert_expr(formula,0);  			// insert expression
				f_comp_value = Tree->calculate(Vars,SMART);	// evaluate expression
				Tree->clean_up(Tree);			// remove evaluation tree
	
				strcpy(ps_name1,Vars->get_formula(name));		// store name of ParameterSet of CombinedVariable that is compared 
				strcpy(ps_name2,Vars->get_formula(constraint));   	// store name of ParameterSet of CombinedVariable for FUZZYCOMPARE    
				found = FALSE;					// right Position in Table not found
				tablefile.open(tablename,ios::in);			// open Table file for concat and compare 
				while((!found)&&(tablefile.eof()==0))
				{
					tablefile >> token;
					if (strcmp(token,"COMPARE")==0)		// search for COMPARE block
					{
 						tablefile>>token;
						if (strcmp(token,ps_name1)==0)	// where in is of right type for CombinedVariable
						{
							tablefile>>token;
							if (strcmp(token,ps_name2)==0) 	// where out is of right type for CombinedVariable for FUZZYCOMPARE
							{
								found = TRUE; 					// right position in table
								cv = Rules->get_cv_list()->find_cv(name);	// get CombinedVariable from RuleSet
 								if (cv==NULL) break;
		  						strcpy(new_cv_name,constraint);
								strcat(new_cv_name,":");
								strcat(new_cv_name,cv->get_cv_name());		// build name for new CombinedVariable (for compare)
								strcat(new_cv_name,c_operator);		// add c_operator (<,>,=,#) to name
								rules_there = TRUE;				// assume that rules do already exist
								if ((cv_new = Rules->get_cv_list()->find_cv(new_cv_name))==NULL) // does new cv already exist?
								{
	 		 						cv_new = new CombinedVariable();	// if not create new CombinedVariable
			 						cv_new->copy(cv);			// with data from CombinedVariable that is compared
			 						cv_new->set_cv_name(new_cv_name);	// insert name of new CombinedVariable 
									Rules->get_cv_list()->add(cv_new);	// add new CombinedVariable to RuleSet
									rules_there = FALSE;			// rules do not exist
								}
								cv_out = Rules->get_cv_list()->find_cv(constraint);	// get CombinedVariable for FUZZYCOMPARE
							}
						}
					}
				}
				if (!found)	// no entry in TABLE for this type of FUZZYCOMPARE
				{
					cout<<endl<<"Error: no entry in TABLE for FUZZYCOMPARE "<<constraint<<" "<<name<<" "<<c_operator<<" "<<formula<<"!"<<endl;
					break;
				}

				if (cv==NULL)
				{
					cout<<endl<<"Error: unknown (FUZZY) VARIABLE "<<name<<"!"<<endl;
					break;
				}
				if (cv_out==NULL)
				{
					cout<<endl<<"Error: unknown (FUZZY) VARIABLE "<<constraint<<"!"<<endl;
					break;
				}

				ParameterSetBase *ppsb = cv->get_PSB();
				median = ppsb->get_median();				// get median of ParameterSet for CombinedVariable
				free( ppsb );
		        f_cv_value = cv->get_crisp_value();				// get crisp value of CombinedVariable
				if ((f_cv_value>0)&&(f_comp_value>0))				// get new value for new CombinedVariable (compare)
					f_value = median + f_cv_value - f_comp_value;
				else
					if ((f_cv_value<0)&&(f_comp_value<0))
						f_value = median - f_cv_value + f_comp_value;
					else
						f_value = median + f_cv_value - f_comp_value;
				Rules->get_cv_list()->find_cv(new_cv_name)->set_value(f_value);	// set new value for CombinedVariable (compare)
	
				// Add all rules to ruleset using table
           
				tablefile>>token;		// skip <,>,=,# tag
				while((tablefile.eof()==0)&&(strcmp(token,c_operator)!=0)) tablefile>>token; // RAGGL
     			while((tablefile.eof()==0)&&(!rules_there))
				{
					tablefile>>in;								// get input linguistic term
					if ((strcmp(in,"CONCAT")==0)||(strcmp(in,"COMPARE")==0)||	// end of block in table reached?
						(strcmp(in,"AND")==0)||(strcmp(in,"OR")==0)||
						(strcmp(in,"=")==0)||(strcmp(in,"#")==0)||
						(strcmp(in,"<")==0)||(strcmp(in,">")==0)||(strcmp(in,"")==0))
					{
						tablefile.close();						// close table
						break;
					}
					else
					{
						tablefile>>out;							// get output linguistc term
						if ((strcmp(in,"")!=0)&&(strcmp(out,"")!=0))
						Rules->add_BasicRule(cv_new , in, cv_out, out, 1);		// add basic rule to RuleSet
					}
				}
				tablefile.close();								// close table
			} // end of ?(tablename not empty)	  
			strcpy(token,"");
			continue;   
		}


// ----------------------------------------------------------------------------------------------------------
// EVAL
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"EVAL")==0)	// EVAL statement reached
		{
			infile >> cv_name;			// get name of cv 
			number_of_tokens++;
			Vars->real_name(cv_name);		// get real name of cv, e.g., FuzzyVar[i], i = 3, -> FuzzyVar[3] 
			cv = Rules->get_cv_list()->find_cv(cv_name);
			if (cv!=NULL)
			{
				f_value = Rules->evaluate(cv);
				Vars->set_value(cv_name,f_value);
				Vars->set_comp(cv_name,TRUE);
			}
			else
			{
				cout<<endl<<"Error: unknown FUZZYVAR "<<cv_name<<"!"<<endl;
				break;
			}
			strcpy(token,"");
			continue;
		}
	
// ----------------------------------------------------------------------------------------------------------
// AGGREGATE
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"AGGREGATE")==0)	// AGGREGATE statement reached
		{
			infile >> varname;			// get branch
			number_of_tokens++;
			infile >> token;			// get operator
			number_of_tokens++;
			infile >> cv_name;			// get name of cv 
			number_of_tokens++;
			infile >> formula;			// get comment string
			number_of_tokens++;

			Vars->real_name(varname);		// get real name of branch	
			Vars->real_name(formula);
			if (Vars->in_list(formula)) strcpy(formula,Vars->get_formula(formula));
			Vars->real_name(cv_name);		// get real name of cv, e.g., FuzzyVar[i], i = 3, -> FuzzyVar[3] 
			cv = Rules->get_cv_list()->find_cv(cv_name);
			if (cv!=NULL)
				if (Evals!=NULL)
				{
					Evals->insert(cv,varname,formula,token);
					Vars->set_comp(cv_name,TRUE);
				}
				else
				{
					cout<<endl<<"Error: Your first have to create a EVALTREE before inserting "<<cv_name<<"!"<<endl;
					break;
				}
			else
			{
				cout<<endl<<"Error: unknown FUZZYVAR "<<cv_name<<"!"<<endl;
				break;
			}
			sprintf(token,"%f",Evals->get_score());
			Vars->set_formula(Evals->get_name(), token);   
			Vars->set_value(Evals->get_name(),Evals->get_score());	// update new overall evaluation in Vars
			Vars->set_comp(Evals->get_name(),TRUE);
			strcpy(token,"");
			continue;
		}

// ----------------------------------------------------------------------------------------------------------
// EVALTREE
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"EVALTREE")==0)	// EVALTREE statement reached
		{
			TVarlist *branches = new TVarlist();
	
			infile >> ps_name;			// get name of parameterset 
			number_of_tokens++;
			infile >> varname;			// get name of tree 
			number_of_tokens++;
			Vars->insert(varname,"OVERALL_EVALUATION",0); // insert overall node to Vars
			infile >>token;			// get names and aggregate operators of single branches
			number_of_tokens++;
			while (strcmp(token,"END")!=0)
			{
				infile>>formula;
				number_of_tokens++;
				branches->insert(token,formula,0);
				infile >>token;			// get next token
				number_of_tokens++;
		    
			}
			ParameterSet *ps_for_eval = NULL;
			int n = Rules->get_ps_list()->get_number();	// get number of ParameterSets in Rules
			for(i=0;i< n;i++)			       // fetch ParameterSet from Rules
			{
				if (strcmp(ps_name,(Rules->get_ps_list()->find(i))->get_ps_name())==0)
				ps_for_eval = Rules->get_ps_list()->find(i);
			}
			if (ps_for_eval == NULL)
			{
				cout<<endl<<"Error: Unknown ParameterSet "<<token<<"!"<<endl;
				break;
			}
			Evals = new TEvalTree(varname,ps_for_eval,Rules,branches);
			Evals->insert_help_rules(tablename);
			branches->clean_up();
			//RAGGL delete branches;
			strcpy(token,"");
			continue;
		}

// ----------------------------------------------------------------------------------------------------------
// TABLE
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"TABLE")==0)	// TABLE statement reached
		{
			infile >> tablename;			// store name of table used to generate rules
			ifstream tablefile;
			// RAGGL
			//tablefile.open(tablename, ios::nocreate);
			tablefile.open(tablename);
			// RAGGL
			if (tablefile.fail())
			{
				cout<<endl<<"Error: Unknown TABLE "<<tablename<<"!"<<endl;
				break;
			}
			tablefile.close();
			strcpy(token,"");
			continue;
		}

// ----------------------------------------------------------------------------------------------------------
// OPERATORSET
// ----------------------------------------------------------------------------------------------------------
		if (strcmp(token,"OPERATORSET")==0)	// OPERATORSET statement reached
		{
			fuzzy_operator fuzzy_op;
			defuzzification_operator dfuzzy_op;
			infile>>token;				// skip IMPORTANCE
			number_of_tokens++;
			infile>>token;				// get importance operator
			number_of_tokens++;
			fuzzy_op = no_fop;
			if (strcmp(token,"MINIMUM")==0) fuzzy_op = minimum;
			if (strcmp(token,"MAXIMUM")==0) fuzzy_op = maximum;
			if (strcmp(token,"FUZZY_AND")==0) fuzzy_op = fuzzy_and;
			if (strcmp(token,"FUZZY_OR")==0) fuzzy_op = fuzzy_or;
			if (strcmp(token,"ASS_COMP_AND")==0) fuzzy_op = ass_comp_and;
			if (strcmp(token,"ASS_COMP_OR")==0) fuzzy_op = ass_comp_or;
			Rules->get_operator_set()->set_importance_operator(fuzzy_op);

			infile>>token;				// skip AND
			number_of_tokens++;
			infile>>token;				// get AND operator
			number_of_tokens++;
			fuzzy_op = no_fop;
			if (strcmp(token,"MINIMUM")==0) fuzzy_op = minimum;
			if (strcmp(token,"MAXIMUM")==0) fuzzy_op = maximum;
			if (strcmp(token,"FUZZY_AND")==0) fuzzy_op = fuzzy_and;
			if (strcmp(token,"FUZZY_OR")==0) fuzzy_op = fuzzy_or;
			if (strcmp(token,"ASS_COMP_AND")==0) fuzzy_op = ass_comp_and;
			if (strcmp(token,"ASS_COMP_OR")==0) fuzzy_op = ass_comp_or;
			Rules->get_operator_set()->set_and_operator(fuzzy_op);

			infile>>token;				// skip OR
			number_of_tokens++;
			infile>>token;				// get OR operator
			number_of_tokens++;
			fuzzy_op = no_fop;
			if (strcmp(token,"MINIMUM")==0) fuzzy_op = minimum;
			if (strcmp(token,"MAXIMUM")==0) fuzzy_op = maximum;
			if (strcmp(token,"FUZZY_AND")==0) fuzzy_op = fuzzy_and;
			if (strcmp(token,"FUZZY_OR")==0) fuzzy_op = fuzzy_or;
			if (strcmp(token,"ASS_COMP_AND")==0) fuzzy_op = ass_comp_and;
			if (strcmp(token,"ASS_COMP_OR")==0) fuzzy_op = ass_comp_or;
			Rules->get_operator_set()->set_or_operator(fuzzy_op);

			infile>>token;				// skip CERTAINTY
			number_of_tokens++;
			infile>>token;				// get CERTAINTY operator
			number_of_tokens++;
			fuzzy_op = no_fop;
			if (strcmp(token,"MINIMUM")==0) fuzzy_op = minimum;
			if (strcmp(token,"MAXIMUM")==0) fuzzy_op = maximum;
			if (strcmp(token,"FUZZY_AND")==0) fuzzy_op = fuzzy_and;
			if (strcmp(token,"FUZZY_OR")==0) fuzzy_op = fuzzy_or;
			if (strcmp(token,"ASS_COMP_AND")==0) fuzzy_op = ass_comp_and;
			if (strcmp(token,"ASS_COMP_OR")==0) fuzzy_op = ass_comp_or;
			Rules->get_operator_set()->set_certainty_operator(fuzzy_op);

			infile>>token;				// skip INFERENCE
			number_of_tokens++;
			infile>>token;				// get INFERENCE operator
			number_of_tokens++;
			fuzzy_op = no_fop;
			if (strcmp(token,"MINIMUM")==0) fuzzy_op = minimum;
			if (strcmp(token,"MAXIMUM")==0) fuzzy_op = maximum;
			if (strcmp(token,"FUZZY_AND")==0) fuzzy_op = fuzzy_and;
			if (strcmp(token,"FUZZY_OR")==0) fuzzy_op = fuzzy_or;
			if (strcmp(token,"ASS_COMP_AND")==0) fuzzy_op = ass_comp_and;
			if (strcmp(token,"ASS_COMP_OR")==0) fuzzy_op = ass_comp_or;
			Rules->get_operator_set()->set_inference_operator(fuzzy_op);

			infile>>token;				// skip ACCUMULATE
			number_of_tokens++;
			infile>>token;				// get ACCUMULATE operator
			number_of_tokens++;
			fuzzy_op = no_fop;
			if (strcmp(token,"MINIMUM")==0) fuzzy_op = minimum;
			if (strcmp(token,"MAXIMUM")==0) fuzzy_op = maximum;
			if (strcmp(token,"FUZZY_AND")==0) fuzzy_op = fuzzy_and;
			if (strcmp(token,"FUZZY_OR")==0) fuzzy_op = fuzzy_or;
			if (strcmp(token,"ASS_COMP_AND")==0) fuzzy_op = ass_comp_and;
			if (strcmp(token,"ASS_COMP_OR")==0) fuzzy_op = ass_comp_or;
			Rules->get_operator_set()->set_accumulate_operator(fuzzy_op);

			infile>>token;				// skip DEFUZZIFICATION
			number_of_tokens++;
			infile>>token;				// get DEFUZZIFICATION operator
			number_of_tokens++;
			dfuzzy_op = no_dfop;
			if (strcmp(token,"COG")==0) dfuzzy_op = cog;
			Rules->get_operator_set()->set_defuzzy_operator(dfuzzy_op);

			infile>>token;				// skip GAMMA
			number_of_tokens++;
			infile>>token;				// get GAMMA operator
			number_of_tokens++;
			Rules->get_operator_set()->set_gamma(atof(token));

			infile>>token;				// skip ACCURACY
			number_of_tokens++;
			infile>>token;				// get ACCURACY operator
			number_of_tokens++;
			Rules->get_operator_set()->set_accuracy(atoi(token));

			infile>>token;				// skip END
			number_of_tokens++;

			strcpy(token,"");
			continue;
		}

		/*
		if (strcmp(token,"")!=0)
			{
			 cout<<endl<<"Error: Sytax error at token #"<<number_of_tokens<<": "<<token<<" ";
			 i = 10;
			 while((infile.eof()==0)&&(i>6))
				 {
				  infile>>token;
				  cout<<token<<" ";
				  i--;
				 }
			 cout<<"... "<<endl;
			 break;
			}
		*/
	}

	if (Controls!=NULL) Controls->clean_up();
	if (Vars!=NULL) Vars->clean_up();
	// RAGGL
	// if (Evals!=NULL) delete Evals;
	// RAGGL
	if (Rules!=NULL) delete Rules;
	// RAGGL
	return Evals;
	// RAGGL
}
