/************************************************************************
 * This library is free software; you can redistribute it and/or	*
 * modify it under the terms of the GNU Library General Public		*
 * License as published by the Free Software Foundation; either		*
 * version 2 of the License, or (at your option) any later version.	*
 * 									*
 * This library is distributed in the hope that it will be useful,	*
 * but WITHOUT ANY WARRANTY; without even the implied warranty of	*
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU	*
 * Library General Public License for more details.			*
 * 									*
 * You should have received a copy of the GNU Library General Public	*
 * License along with this library; if not, write to the		*
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,		*
 * Boston, MA  02111-1307, USA.						*
 * 									*
 * The GNU Library General Public License file is included in		*
 * Online-Documentation							*
 *									*
 * GNU's Homepage - http://www.gnu.ai.mit.edu/home.html			*
 *									*
 * Questions & comments referring to *FLIP++: please contact us:	*
 *									*
 * Snail-Mail:								*
 *	Dr. Wolfgang SLANY						*
 *	Database and Expert System Group (E184-2)     			*
 *	Institute of Information Systems               			*
 *	Vienna University of Technology					*
 *	Paniglgasse 16, A-1040 Vienna, AUSTRIA				*
 * Telephone:								*
 *	+43-1-58801-6123     (area-code 1 maybe changed to 201 in 1998) *
 * FAX:									*
 *	+43-1-5055304	     (area-code 1 maybe changed to 201 in 1998) *
 * eMail:								*
 *	sflip@dbai.tuwien.ac.at						*
 * WWW:									*
 *	http://www.dbai.tuwien.ac.at/proj/StarFLIP/			*
 *                                                                      *
 * $Id: optimize.cpp,v 1.3 1997/06/24 17:27:32 reichel Exp $ *
 ***********************************************************************/
// ----------------------------------------------------------------------------------------------------------
//	What   : Domain Knowledge Representation with *FLIP++
//	Project: DomFLIP++
//	Author : Andreas Raggl
//	Start  : 10.12.1996
//	Last   : 14.01.1997
// 	File   : optimize.cpp
// ----------------------------------------------------------------------------------------------------------

#include "optimize.h"

// global functions

ostream& operator<<		( ostream& os, COptimize& c )
{
	c.Write( os );
	return os;
};

istream& operator>>		( istream& is, COptimize& c )
{
	c.Read( is );
	return is;
};

// class definition of COptimize

COptimize::COptimize	( CShiftSchedulingDomain *c ) : CBase( "COptimize", 1 )
{
	cDomain = c;
	file1.open("c.dat", ios::out|ios::trunc);
	file2.open("o.dat", ios::out|ios::trunc);
	iTries = 0;
};

COptimize::~COptimize	()
{
        file1.close();
        file2.close();
};

void COptimize::RepairDeep	()
{
	int iDepth = 1, iMaxDepth = 5;
	int tries = 0;

	cDomain->Evaluate();
	fBestEvaluation = cDomain->GetEvaluation();
	cBestSchedule = cDomain->GetSchedule();
	while ( iDepth < iMaxDepth )
	{
		bool bResult = RepairTry( iDepth );
		if ( bResult )
		{
			fBestEvaluation = cDomain->GetEvaluation();
			cBestSchedule = cDomain->GetSchedule();
			iDepth = 1;
		}
		else
			iDepth++;
		if (tries++ > 40) break;
		cout << "Evaluation: " << fBestEvaluation << endl;
                cout << "**************************************************" << endl;
                char filename[20];
                sprintf(filename, "o%d", tries);
                cDomain->Write(filename);
	}
};

bool COptimize::RepairTry	( int iDepth )
{
	CSchedule cPreviousPlan = cDomain->GetSchedule();
	int iViolatedGroup, iViolatedDay;

	if ( iDepth == 0 )
		if ( cDomain->GetEvaluation() > fBestEvaluation )
			return true;
		else
			return false;
	else
	{
		int i,j;
		for ( i = 0; i < (int)cDomain->GetRepairPositions().size(); i++ )
		{
			for ( j = 0; j < (int)cDomain->GetRepairPositions()[i].size(); j++ )
			{
				iViolatedGroup = cDomain->GetRepairPositions()[i][j].group;
				iViolatedDay = cDomain->GetRepairPositions()[i][j].day;
				int iStart = iViolatedDay;
				int iStart2 = iViolatedDay;
				int k = 0;
				while ( cDomain->Repair( iViolatedGroup, iViolatedDay, iStart, iStart2 ) )
				{
					k++;
					cDomain->Evaluate();
					bool bResult = RepairTry( iDepth - 1 );
					cout << "viol: " << i << " start: " << j << " repairtry: " << k << " eval: " << cDomain->GetEvaluation() << endl;
					if ( bResult ) 
						return true;
					else
						cDomain->SetSchedule( cPreviousPlan );
				}
			} // for
		}; // for 
		return false;
	} // else
};

void COptimize::RepairRandom	()
{
	int tries = 0;
	int iDepth = 1;

	cDomain->Evaluate();
	fBestEvaluation = cDomain->GetEvaluation();
	cBestSchedule = cDomain->GetSchedule();
	while ( 1 )
	{
		RepairTry2( iDepth );
		if (tries++ > 40) break;
		cout << "Evaluation: " << fBestEvaluation << endl;
                cout << "**************************************************" << endl;
                char filename[20];
                sprintf(filename, "o%d", tries);
                cDomain->Write(filename);
	}
};

void COptimize::RepairTry2	( int iDepth )
{
	int iMaxViolations = 1;
	int iMaxRepairStarts = 1;
	CSchedule cPreviousSchedule = cDomain->GetSchedule();
	int iViolatedGroup, iViolatedDay;

	if ( iDepth == 0 )
		return;
	else
	{
		int i, j;
		for ( i = 0; i < iMaxViolations; i++ )
		{
			for ( j = 0; j < iMaxRepairStarts; j++ )
			{
				iViolatedGroup = cDomain->GetRepairPositions()[i][j].group;
				iViolatedDay = cDomain->GetRepairPositions()[i][j].day;
				srand( (unsigned) time( NULL ) );
				int iStart = rand() % ( cDomain->GetLengthOfCycle() * 7 );
				int iStart2 = rand() % ( cDomain->GetLengthOfCycle() * 7 );
				int k = 0;
				float fxBestEvaluation = -1;
				while ( cDomain->Repair( iViolatedGroup, iViolatedDay, iStart, iStart2 ) )
				{
					k++;
					cDomain->Evaluate();
				        file1 << iTries++ << space << cDomain->GetEvaluation() << endl;
				        if ( cDomain->GetEvaluation() > fBestEvaluation )
					  {
					   fBestEvaluation = cDomain->GetEvaluation();
				           file2 << (iTries - 1) << space << fBestEvaluation << endl;
					  }
					RepairTry2( iDepth - 1 );
					cout << "viol: " << i << " start: " << j << " repairtry: " << k << " eval: " << cDomain->GetEvaluation() << endl;
					if ( cDomain->GetEvaluation() > fxBestEvaluation )
					{
						fxBestEvaluation = cDomain->GetEvaluation();
						cBestSchedule = cDomain->GetSchedule();
					 cout << "Iteration " << k << " better. " << endl;
					}
					if ( k == 10) break;
					iStart = rand() % ( cDomain->GetLengthOfCycle() * 7 );
					iStart2 = rand() % ( cDomain->GetLengthOfCycle() * 7 );
					cDomain->SetSchedule( cPreviousSchedule );
				}
			        if ( k < 10 ) 
{ cout << iViolatedGroup << space << iViolatedDay << space << iStart << space << iStart2 << endl;
 cDomain->Write("error.dat"); }
			} // for
		}; // for 
	} // else
};

void COptimize::Read	( istream& is )
{
	CBase::Read( is );
};

void COptimize::Write	( ostream& os ) const
{
	CBase::Write( os );
};
